--! \file

require "strict"

require('lj_include'){
    'stdio',
    'string',
    'unistd',
    'fcntl',
    'sys/types',
    'signal',
    'linux/time',
    'sys/time',
    'sys/resource',
}

local __ = require "__"
local log = require "log"
local class = require "class"
local bit = require "bit"
local bits = require "bits"
local cjson = require "cjson.safe"
local crypto = require "crypto"
local dbus = require "dbus"
local ev = require "ev"
local ffi = require "ffi"
local http_parser = require "http_parser"
local lpeg = require "lpeg"
local membuf = require "membuf"
local scgi_server = require "scgi_server"
local u = require "unix"
local setup_spec = require "setup_spec"
local sie_parser = require "sie_parser"
local sqlite3 = require "lsqlite3"
local sqlite = require "sqlite_ffi"
local timestamps = require "timestamps"
local us = require "underscore"
local uuid = require "uuid"
local network_db = require "network_db"
local lfs = require "lfs"
local setup_migrator = require 'setupdb_migrate'
local system_config = require "system_config"
local fs_info = require 'fs_info'
local sys_notify = require 'sys_notify'
local status_lib = require 'status_lib'
local sie_file_mgr = require 'sie_file_mgr'
local armada_mmio = require 'armada_mmio'
local pn = require 'push-notifications'

local bus

local sl = status_lib:new('system_status')
local sys_status = {
    test = {running=false, types={}},
    error = {error=false},
}
sl:set_status(sys_status)

local storage_path = system_config.storage.path or '/storage'

local sql_exec

local C = ffi.C

local bit_band = bit.band
local bit_lshift = bit.lshift
local bit_rshift = bit.rshift
local ffi_string = ffi.string

local ev_loop = ev.Loop.default

local setup_db_file = '/var/lib/armada/setup.db'

local local_authorization_key = "locally_authorized"
local start_is_locked = true

local pending_tests = {}
local pending_test_watchdog_timer = nil

-- GUI helper function to notify it when download payload has been sent
local function downloadCompleteNotification(session_key)
    local sess_key = session_key or 'none'
    local sn = sys_notify:new('DOWNLOAD Complete')
    sn:notify(sess_key, true)
end

-- Wrapper for executing OS shell commands
local function shellf(fmt, ...)
    local cmd = string.format(fmt, ...)
    local ok, term, status = os.execute(cmd)
    if term == nil then
        -- 5.1 semantics
        ok = bit.band(ok, 0xFF) == 0
    end
    return ok
end

-- Function used to generate a hostname based on target type and MAC address of
-- the specified interface.
local function make_default_hostname(iface)
    local f = io.popen('ip link show dev ' .. (iface or 'eth0') ..
        ' | grep ether')
    if f then
        for line in f:lines() do
            local mac = line:match('link/ether (.-) ')
            if mac then
                f:close()
                local v1, v2, v3 = mac:match('^.-:.-:.-:(.-):(.-):(.-)$')
                local unit = ''
                if system_config.target == 'armada' then
                    unit = (system_config.device_type or 'cx23r') .. '-'
                else
                    unit = 'ex23r-'
                end
                local hname = string.format('%s%s%s%s', unit, v1, v2, v3)
                return hname
            end
        end
        f:close()
    end
end
local hostname = make_default_hostname()

local function dbg(fmt, ...)
    log:log(log.DEBUG, string.format(fmt, ...))
end

local uri_prefixes = {'/comet/', '/api/'}
local uri_unprefix_matches = {}
for _, uri_prefix in ipairs(uri_prefixes) do
    table.insert(uri_unprefix_matches, '^' .. uri_prefix .. '(.*)$')
end

local handlers = {}
local always_authorized_handlers = {
    sign_in = true,
    sign_out = true,
}
local requests = {}
local always_authorized_requests = {}

local function epoch_timestamp()
    local sec, nsec = u.clock_gettime()
    return sec + 1e-9 * nsec
end

local function iso_timestamp()
    return timestamps.epoch_to_iso(epoch_timestamp())
end

local function child_watcher(server, loop, child, revents)
    local status = child:getstatus()
    local pid = child:getrpid()
    if pid and pid ~= 0 then
        local kid = server.kids[pid]
        if kid.fn then
            kid.fn(server, loop, child, revents)
            server.kids[pid] = nil
        end
    end
end

--! A Comet server.
--? class api_server scgi_server
local m = class(scgi_server)
--? alias m api_server

function m:init(settings)
    m.super().init(self, settings)

    for k, v in pairs(handlers) do
        for _, uri_prefix in ipairs(uri_prefixes) do
            requests[uri_prefix .. k] = v
        end
    end
    for k, v in pairs(always_authorized_handlers) do
        for _, uri_prefix in ipairs(uri_prefixes) do
            always_authorized_requests[uri_prefix .. k] = v
        end
    end

    self.kids = {} -- indexed by child pid
    ev.Child.new(function(l, c, r) child_watcher(self, l, c, r) end,
        0, false):start(ev_loop)

    ev.Signal.new(function() -- Ignore SIGPIPE -- KLUDGE move to server.lua
    end, u.SIGPIPE):start(ev_loop)

    local function die()
        self.ev_loop:unloop()
    end

    for _, sig in ipairs{u.SIGINT, u.SIGTERM} do
        ev.Signal.new(die, sig):start(ev_loop)
    end

    self.comet_clients = {}
    self.accumulators = {}
    self.fake_comet_setup = {}
    self.running_tests_kludge = {}

    bus = dbus.bus_get_private('system')
    if bus then
        bus:hook_libev_default_loop()
        self.master = assert(bus:auto_proxy_wait('com.somat.armada.Master',
                                                '/com/somat/armada/Master'))
        self.network = assert(bus:auto_proxy_wait('com.somat.armada.Network',
                                                 '/com/somat/armada/Network'))
        self.logger = assert(bus:auto_proxy_wait('com.somat.armada.Logger',
                                                 '/com/somat/armada/Logger'))
        self.timing = assert(bus:auto_proxy_wait('com.somat.armada.Timing',
                                                 '/com/somat/armada/Timing'))
--[[
        -- Get the inside network name
        local rrs = self.network:QueryLocalRR('?x23r*', 'A', '*', 'SOA')
        if not rrs or #rrs < 1 then
            log:log(log.ERR, "Unable to find inside network 'A' record")
            return
        end
        local host = rrs[1][1]

        -- If an alias file exists on this machine,
        -- set the alias to whatever it is in that file
        local fh = io.open('/var/lib/armada/alias.txt', 'r')
        if fh then
            local alias = fh:read('*l')
            fh:close()
            if alias then
                local rrs = self.network:QueryLocalRR(alias, 'ALIAS', host, '*')
                if rrs and #rrs > 0 then
                    self.network:RemoveRR(alias, 'ALIAS', host)
                end
                self.network:AddRR(alias, 'ALIAS', host)
            end
        end
]]
        self.network.ChangedRRs:bind(
            function(removed, added)
                self:changed_rr(removed, added)
            end)

        -- Event signaled when a test is spawned by master
        self.master.TestSpawned:bind(function(id, pid)
                local p = tonumber(pid) or 0
                if p > 0 then
                    self.running_tests_kludge[id] = p
                end
            end)
    end

    -- During boot there can be a temporary inability to open the setup
    -- database.  So...we attempt to do up to 5 retries 1 second apart to get
    -- it open.  If the database can't be opened in 5 retries then we error
    -- out.
    local ok
    for retry = 1, 5 do
        ok, self.setup_db = pcall(sqlite.new, sqlite, setup_db_file)
        if ok then break end
        C.sleep(1)
        self.setup_db = nil
    end
    if not self.setup_db then
        log:log(log.ERR, "Unable to open setup database after 5 attempts")
        os.exit(-4)
    end

    sie_file_mgr.setup_db = self.setup_db
    sie_file_mgr.clean_sie()
    --sie_file_mgr.check_sie_files()

    self.throttled_live_data = {}
    self.comet_proxies_by_name = {}
    self.comet_proxies_by_socket = {}
end

local function read_file(fname)
    local fh = io.open(fname)
    if not fh then
        return nil
    end
    local content = fh:read('*a')
    fh:close()
    return content
end

local send_response -- forward decl

local function coro_post_mortem(server, loop, child, revents)
    local pid = child:getrpid()
    local status = child:getstatus()
    local kid = server.kids[pid]
    local r = {
        ok = status.exit_status == 0,
        pid = pid,
        exited = status.exited,
        exit_status = status.exit_status,
        stopped = status.stopped,
        signaled = status.signaled,
        term_signal = status.term_signal,
    }
    local ok, value = coroutine.resume(kid.w.coro, r)
    if coroutine.status(kid.w.coro) == 'dead' then
        -- KLUDGE KLUDGE KLUDGE correct???
        if not ok then
            pcall(function() server.setup_db:exec("ROLLBACK;") end)
            send_response(kid.w, 500, "Server Error", "text/plain", "%s\n",
                value)
        end
    end
end

function m:fork(loop, w, tag, cfg, cmd, ...)
    local kid = {}
    kid.w = w
    kid.pid = C.fork()
    if kid.pid == 0 then
        loop:fork()
        -- We are the child

        -- Child needs it's own log entries
        log:close()
        log:open('api-server-worker', log.USER)
        log:log(log.NOTICE, "'" .. tag .. "' Worker starting")

        -- The child needs it's own DBUS connect so responses to DBUS requests
        -- made by the child don't end up being sent to the parent
        local nbus = dbus.bus_get_private('system')
        if not nbus then
            log:error("bus_get_private() failed")
            kid.ok = false
            kid.exec_failed = true
            kid.dbus_failed = true
            return kid
        end
        bus:close()
--        bus:unref()
        bus = nbus
        nbus:hook_libev_default_loop()
        w.server.master =
            nbus:auto_proxy('com.somat.armada.Master',
                            '/com/somat/armada/Master')
        w.server.network =
            nbus:auto_proxy('com.somat.armada.Network',
                            '/com/somat/armada/Network')

        local dev_null_fd = u.open('/dev/null', u.O_RDONLY) -- KLUDGE
        C.dup2(dev_null_fd, 0)
        C.dup2(dev_null_fd, 1)
        C.dup2(dev_null_fd, 2)
        C.close(dev_null_fd)

        if type(cfg) == 'number' then
            cfg = { cpu_limit = cfg > 0 and cfg or nil }
        end
        assert(type(cfg) == 'table')
        if cfg.cpu_limit then
            local lim = cfg.cpu_limit
            if type(lim) == 'number' then
                lim = {lim, lim + 1}
            end
            assert(type(lim) == 'table')
            local rlim = ffi.new('struct rlimit', lim)
            C.setrlimit(C.RLIMIT_CPU, rlim)
        end
        if type(cmd) == 'function' then
            local ok = cmd()
            os.exit(ok or 0)
        end
        local in_args = type(cmd) == 'table' and cmd or {cmd, ...}
        local args = {}
        for i, v in ipairs(in_args) do
            args[i] = tostring(v)
        end
        local cargs = ffi.new('char *[?]', #args + 1)
        local t = {} -- Keep stuff from being GC'd
        for i, arg in ipairs(args) do
            t[i] = ffi.new('char[?]', #arg + 1, arg)
            cargs[i - 1] = t[i]
        end
        C.execvp(cargs[0], cargs)
        log:error("execvp() failed") -- KLUDGE more descriptive
        kid.ok = false
        kid.exec_failed = true
        return kid
    elseif kid == -1 then
        log:error("fork() failed") -- KLUDGE more descriptive
        kid.ok = false
        kid.fork_failed = true
        return kid
    end
    kid.fn = coro_post_mortem
    self.kids[kid.pid] = kid
    return coroutine.yield(kid.w.coro)
end

-- The 'is_replay' argument is added because this function is (ab)used by
-- the Comet stream initialization code to "replay" the current conditions
-- into the Comet stream. I.e., it will not only be called when RRs change,
-- but also called again for all RRs anytime a new GUI client connects.
-- When called for a replay, is_replay is set to true, so that the
-- processing code knows not to take actions that should only be taken
-- when the RR is REALLY being removed or added to the system (such as
-- updating sys_status).
function m:changed_rr(removed, added, is_replay)
    self.changed = {}

    for i,rr in ipairs(removed or {}) do
        self:rr_change("remove", rr, is_replay)
    end
    for i,rr in ipairs(added or {}) do
        self:rr_change("add", rr, is_replay)
    end

    self:emit_accumulated_changes()
end

-- Function used to make sure the yellow LED gets turned off when there are
-- no tests running.
function m:tests_monitor()
    if not self.tests_monitor_timer then
        self.net_db = self.net_db or network_db:new()
        self.tests_monitor_timer = ev.Timer.new(function(l, t, e)
                if not self.net_db:tests_running() then
                    sys_status.test.running=false
                    sl:set_status(sys_status)
                    self.running_tests = 0
                    t:stop(l)
                end
            end, 1, 1)
    end
    self.tests_monitor_timer:start(ev_loop)
end

local function start_pending_test_watchdog()
    if not pending_test_watchdog_timer then
        pending_test_watchdog_timer = ev.Timer.new(function(l, t, e)
            local pending_test_id_map = {}
            for _, pending_test in ipairs(pending_tests) do
                pending_test_id_map[pending_test.id] = pending_test
            end
            -- Get a list of all pending-test PID files, and check each
            -- one. If there is no corresponding entry in pending_tests,
            -- remove the PID file. Otherwise, if the PID file contains a
            -- valid PID, check for the existence of that process; if it
            -- doesn't exist, clear the pending_tests record.
            local pid_file_list = io.popen('ls /var/run/pending-test-* 2>/dev/null', 'r')
            if not pid_file_list then return end
            local pids = {}
            for pid_filename in pid_file_list:lines() do
                local pending_test_id = pid_filename:match('pending%-test%-(.*)%.pid$')
                if pending_test_id_map[pending_test_id] then
                    local f = io.open(pid_filename, 'r')
                    if f then
                        local pid = tonumber(f:read('*a'):match('(-?%d+)'))
                        if pid then pids[pending_test_id] = pid end
                        f:close()
                    end
                else
                    os.execute('rm ' .. pid_filename)
                end
            end
            pid_file_list:close()
            for pending_test_id, pid in pairs(pids) do
                if pid > 0 then
                    local rv = C.kill(ffi.new('pid_t', tonumber(pid)), 0)
                    if rv ~= 0 then
                        for i, pending_test in ipairs(pending_tests) do
                            if pending_test.id == pending_test_id then
                                table.remove(pending_tests, i)
                                break
                            end
                        end
                    end
                end
            end
            local i = 1
            local n = #pending_tests
            while i <= n do
                if not pids[pending_tests[i].id] then
                    table.remove(pending_tests, i)
                    n = n - 1
                else
                    i = i + 1
                end
            end
        end, 1, 1)
        pending_test_watchdog_timer:start(ev_loop)
    end
end

local function reload_nginx_config()
    -- Tell NGINX to re-read it's configuration files.
    local ch
    if system_config.device_type and system_config.device_type == 'edaqxr' then
        ch = io.popen('ps -eo pid,args | grep nginx | grep master | grep -v grep')
    else
        ch = io.popen('ps -o pid,args | grep nginx | grep master | grep -v grep')
    end
    if ch then
        for line in ch:lines() do
            local pid = line:match('(%d+) nginx: master')
            if pid then
                C.kill(ffi.new('pid_t', tonumber(pid)), C.SIGHUP)
                break
            end
        end
        ch:close()
    end
end

-- Function used to immediately kill all locally running tests
local function killall_tests()
    local plist
    if system_config.device_type and system_config.device_type == 'edaqxr' then
        plist = io.popen('ps -eo pid,group | grep acq')
    else
        plist = io.popen('ps -o pid,group | grep acq')
    end
    if plist then
        local pid_list = {}
        for p in plist:lines() do
            local pid = p:match('(%d+) ')
            if pid then
                pid_list[#pid_list + 1] = pid
            end
        end
        plist:close()
        if #pid_list > 0 then
            for _, pid in ipairs(pid_list) do
                C.kill(ffi.new('pid_t', tonumber(pid)), C.SIGINT)
            end
        end
    end
end

local start_delay

function m:delay_start()
    if not self.start_delay_timer then
        self.start_delay_timer = ev.Timer.new(function(l, t, e)
                t:stop(l)
                start_delay = false
            end, 5, 5)
    end
    start_delay = true
    if not self.start_delay_timer:is_active(ev_loop) then
        self.start_delay_timer:start(ev_loop)
    end
end

function m:reconfigure_nginx(unit, ip)
    if ip then
        local fh =
            io.open('/tmp/nginx/conf.d/' .. unit .. '_upstream.conf', 'w')
        if not fh then return end
        fh:write(string.format("    upstream stream_%s {\n", unit))
        fh:write(string.format("        server     %s:8080;\n", ip))
        fh:write("    }\n")
        fh:write(string.format("    upstream stream_control_%s {\n", unit))
        fh:write(string.format("        server     %s:8081;\n", ip))
        fh:write("    }\n")
        fh:close()
        os.execute('/usr/bin/killall -HUP nginx')
    else
        os.remove('/tmp/nginx/conf.d/' .. unit .. '_upstream.conf')
        os.execute('/usr/bin/killall -HUP nginx')
    end
end

-- The 'is_replay' argument is added because this function is (ab)used by
-- the Comet stream initialization code to "replay" the current conditions
-- into the Comet stream. I.e., it will not only be called when RRs change,
-- but also called again for all RRs anytime a new GUI client connects.
-- When called for a replay, is_replay is set to true, so that the
-- processing code knows not to take actions that should only be taken
-- when the RR is REALLY being removed or added to the system (such as
-- updating sys_status).
local mx_pis = {}
local xr_pis = {}

function m:rr_change(action, rr, is_replay)
--print("rr_change(" .. action .. ", " .. cjson.encode(rr))
    if rr[1] == "_armada-connectors" then
        if rr[3]:find('mx%d+b%-%x+%..+') then
            -- Only send local sourced connectors for MX devices
            -- WARNING: The GUI will perform the following lines. They are commented out here
            --          so the GUI can keep track of which proxies are running and which are not
            --if not rr[3]:find(hostname, 1, true) then return end
            --rr[3] = rr[3]:match('^(.*mx%d+b%-%x+)%..+$')
        end
        local cmd = "#" .. action .. "_connector"
        self:dispatch_to_clients{{cmd, rr[3]}}
    elseif rr[1] == "_armada-hardware" then
        local cmd = "#" .. action .. "_hardware"
        self:dispatch_to_clients{{cmd, rr[3]}}
    elseif rr[1] == "_reset-error" then
        if action == 'remove' then
            sys_status.error.error = false
            sl:set_status({error={error=false,reset=true}})
        end
    elseif rr[2] == "ALIAS" then
        if action == 'add' then
            local cmd = "#" .. action .. "_alias"
            local b = {}
            if rr[3]:find('^mx%d+b%-%x+%..+$') then
                -- Only send local sourced alias' for MX devices
                if not rr[3]:find(hostname, 1, true) then return end
                rr[3] = rr[3]:match('^(mx%d+b%-%x+)%..+$')
            end
            b[rr[1]] = rr[3]
            self:dispatch_to_clients{{cmd, b}}
        end
    elseif rr[2] == "TXT" and rr[1]:find("^_system%-notification") then
        local m = cjson.decode(rr[3])
        if action == 'add' then
            if m then
                if m.type and m.type == 'error' then
                    if not sys_status.error.error or m.severity == 'normal' then
                        local cmd = "#" .. action .. "_notification"
                        self:dispatch_to_clients{{cmd, rr[3]}}
                    end
                    if m.severity ~= 'normal' then
                        sys_status.error.error = true
                        sl:set_status(sys_status)
                    end
                    return
                end
                if m.type and m.type == '_notice' and
                        m.msg:find(__'Data acquisition has begun') and
                        not is_replay then
                    sys_status = sl:get_status()
                    sys_status.test = sys_status.test or {}
                    sys_status.test.running = true
                    sys_status.test.types = sys_status.test.types or {}
                    sys_status.error = sys_status.error or {}
                    sys_status.error.error = sys_status.error.error or false
                    sl:set_status(sys_status)
                end
                local cmd = "#" .. action .. "_notification"
                self:dispatch_to_clients{{cmd, rr[3]}}
            end
        elseif m and m.reboot then
            if system_config.max_test_restarts > 0 then
                -- If reboot was specified when the notification is removed, do
                -- the reboot after a short delay
                ev.Timer.new(function(l, t, e)
                        log:log(log.ERR, "System reboot due to error")
                        os.execute('/usr/bin/reboot error')
                    end, 2, 0):start(ev_loop)
            else
--[[
                -- If system error resets are overridden, just kill the test(s)
--[[
                log:log(log.ERR, "Stopping test due to critical error\n" ..
                    "Test will be not be restarted on next reboot")
]]
                killall_tests()
            end
        elseif action == 'remove' and m and m.announce_remove then
            local cmd = "#" .. action .. "_notification"
            self:dispatch_to_clients{{cmd, rr[3]}}
        end
    elseif rr[2] == "TXT" and rr[3]:match("^available_channels:") then
        local cmd = "#" .. action .. "_channel"
        local js = rr[3]:gsub("^.-%:", "")
        local chs = cjson.decode(js)
--print("rr_change(AVAILABLE_CHANNELS): " .. cjson.encode(chs))
        local proxy_chs = {} -- KLUDGE
        for _, ch in ipairs(chs) do

            proxy_chs[#proxy_chs + 1] = ch.connector
            if action == "remove" then
                self.throttled_live_data[ch.connector] = nil
            end

            -- If this is an MX device channel, extract the network device
            -- from the "connector" and make a new field called
            -- "network_node".
            if ch.connector and ch.connector:find('^.+mx%d+b%-%x+%..+$') then
                local c, n = ch.connector:match('^(.+mx%d+b%-%x+)%.(.+)$')
                if c and n then
                    ch.connector = c
                    ch.network_node = n
                end
            end

            self:dispatch_to_clients{{cmd, {ch}}} -- KLUDGE
        end
        if action == "add" and #self.comet_clients > 0 then
            if not self.comet_proxies_by_name[rr[1]] then
                local host, port, ip = self.network:ResolveService(rr[1])
                if host then
                    self:connect_proxy(rr[1], ip, port, proxy_chs)
                end
            end
        end
    elseif rr[2] == "TXT" and rr[3]:match("^platform_info:") then
        local cmd = '#' .. action .. '_platform_info'
        local js = rr[3]:gsub("^.-%: ", "")
        local pi = cjson.decode(js)
        if pi and pi.device_id and pi.device_id:find('^mx%d+b%-%x+') then
            -- Only send platform info for proxies this node is running
            if not rr[1]:find(hostname, 1, true) then return end
            if action == 'add' then
                mx_pis[pi.device_id] = pi
            else
                mx_pis[pi.device_id] = nil
            end
        end
        if pi and pi.device_id and
            (pi.device_id:find('^cx23r') or pi.device_id:find('^edaqxr')) then
            local ip
            if action == 'add' then
                ip = pi.zcip_address
                if not ip then
                    log:log(log.NOTICE,
                        "XR device with no reported ZCIP address")
                end
                xr_pis[pi.device_id] = pi
            else
                xr_pis[pi.device_id] = nil
            end
            self:reconfigure_nginx(pi.device_id, ip)
        end
        self:dispatch_to_clients{{cmd, pi}}
    elseif rr[2] == "TXT" and rr[3]:match("^connector_status:") then
        if action == 'add' then
            local cmd = "#connector_status"
            local js = rr[3]:gsub("^.-%:", "")
            local cstatus = cjson.decode(js)
            self:dispatch_to_clients{{cmd, {cstatus}}}
        end
    elseif rr[2] == "TXT" and rr[3]:match("^teds_status:") then
        if not rr[3]:find(hostname, 1, true) then return end
        if action == 'add' then
            local cmd = "#teds_status"
            local js = rr[3]:gsub("^.-%:", "")
            js = string.gsub(js, '(mx%d+b%-%x+)%..-"', '%1"')
            local tstatus = cjson.decode(js)
            self:dispatch_to_clients{{cmd, {tstatus}}}
        end
    elseif rr[2] == "TXT" and rr[1]:match("._armada%-tests$") then
        self:accumulate_changes("tests", action, rr)
        -- Update a lookup table showing what type(s) of test(s) are
        -- currently running (sie, digital_display, etc.); this is set
        -- in sys_status for access by powerd.lua, which sets the yellow
        -- LED state based on what kinds of tests are currently running.
        if not is_replay then
            local k, v = rr[3]:match("^(.-):%s*(.*)$")
            if k and k == 'setup-file' then
                local spec = cjson.decode(v)
                local summary
                for _, stanza in ipairs(spec) do
                    if stanza.id == 'summary' then
                        summary = stanza
                        break
                    end
                end
                local purpose = 'sie'
                if summary and summary.digital_display_type then
                    purpose = summary.digital_display_type
                elseif summary and summary.test_run_mode and
                    summary.test_run_mode == 'remote_control' then
                    purpose = 'remote'
                elseif summary and summary.purpose then
                    purpose = summary.purpose
                end
                if action == 'add' then
                    if purpose ~= 'remote' then --remote led status is set in the test engine
                        sys_status.test.types[purpose] =
                            (sys_status.test.types[purpose] or 0) + 1
                        sl:set_status(sys_status)
                    end
                elseif action == 'remove' then
                    if purpose == 'sie' then
                        self:delay_start()
                    end
                    if purpose == 'remote' then
                        sys_status.test.types.remote = nil
                        sys_status.test.running = false
                    elseif sys_status and sys_status.test and
                        sys_status.test.types and
                        sys_status.test.types[purpose] then
                        sys_status.test.types[purpose] =
                            (sys_status.test.types[purpose] or 1) - 1
                    end
                    sl:set_status(sys_status)
                end
            end
        end
    elseif rr[2] == "SRV" and rr[1]:match("._armada%-tests$") then
        self:accumulate_changes("tests", action, rr)
        local spec = cjson.decode(rr[3])
        if action == 'add' then
            self.running_tests = self.running_tests or 0
            self.running_tests = self.running_tests + 1
            self:tests_monitor()
            local test_id = rr[1]:match('^(.*)%._armada%-tests$')
            for i, pending_test in ipairs(pending_tests) do
                if pending_test.id == test_id then
                    table.remove(pending_tests, i)
                    break
                end
            end
            local pid_filename = '/var/run/pending-test-' .. test_id .. '.pid'
            os.execute('rm ' .. pid_filename .. ' 2>/dev/null')
        elseif action == 'remove' and self.running_tests and not is_replay then
            local hash = rr[1]:match('^(.+)%._armada%-tests')
--[[
            if hash then
                os.execute('rm /tmp/nginx/conf.d/' .. hash .. '* 2>/dev/null')
                reload_nginx_config()
            end
]]
            self.running_tests = self.running_tests - 1
            if self.running_tests <= 0 then
                if self.tests_monitor_timer then
                    self.tests_monitor_timer:stop(ev_loop)
                    self.tests_monitor_timer:clear_pending(ev_loop)
                end
                self.running_tests = 0
                sys_status.test.running = false
                sl:set_status(sys_status)
            end
        end
    elseif rr[2] == "TXT" and rr[1]:match("%._armada%-files$") then
        self:accumulate_changes("file", action, rr)
    elseif rr[2] == "TXT" and
        (rr[1]:match("._cx23r%-storage$") or
            rr[1]:match("._edaqxr%-storage$")) then
        self:accumulate_changes("storage", action, rr)
    elseif rr[2] == "TXT" and rr[1]:match("interactive%-trigger") then
        if action == 'add' then
            self:accumulate_changes("inter_trig", action, rr)
        end
--[[
    elseif rr[2] == 'A' and (rr[1]:find('edaqxr') or rr[1]:find('cx23r')) and
        rr[1] ~= hostname then
        local ip
        if action == 'add' then
            ip = '169.254.46.114'
        end
        self:reconfigure_nginx(rr[1], ip)
]]
    end
end

function m:dispatch_to_clients(message)
    for i = self.start_comet_client or 1, #self.comet_clients do
        self:write_comet(self.comet_clients[i], message)
    end
end

function m:accumulate_changes(tag, action, rr)
    self.accumulators[tag] = self.accumulators[tag] or {}
    local acc = self.accumulators[tag]
    local id = rr[1]
    local k, v

    if rr[2] == "TXT" then
        k, v = rr[3]:match("^(.-):%s*(.*)$")
        if not k then
            return
        end
    elseif rr[2] == "SRV" then
        local port, host = rr[3]:match("^%s*%d+%s+%d+%s+(%d+)%s+(%S+)")
        k, v = "srv_port", port -- KLUDGE for now, should have host
    end
    self.changed[tag] = self.changed[tag] or {}
    self.changed[tag][id] = true
    if action == "add" then
        acc[id] = acc[id] or {}
        acc[id][k] = v
    else
        if acc[id] then
            acc[id][k] = nil
        end
    end
end

function m:emit_accumulated_changes()
    for tag, changes in pairs(self.changed) do
        local acc = self.accumulators[tag]
        local update, remove = "#update_" .. tag, "#remove_" .. tag
        for f, _ in pairs(changes) do
            local rec = acc[f]
            if rec and next(rec) then
                rec.id = f -- KLUDGE
                self:dispatch_to_clients{{update, rec}}
                rec.id = nil -- KLUDGE
            else
                self:dispatch_to_clients{{remove, f}}
                acc[f] = nil
            end
        end
    end
end

local comet_parse_table
ffi.cdef[[
    enum {
        COMET_PARSE_ERROR = 0xFF,
        COMET_PARSE_DONE = 0xFE,
        COMET_STATE_MASK = 0x1F,
        COMET_OP_MASK = 0xE0,
        COMET_OP_SHIFT = 5,
        COMET_MARK_PREV = 1,
        COMET_MARK_THIS = 2,
        COMET_MARK_NEXT = 3,
        COMET_MARK_BACKSLASH = 4,
        COMET_MAX_MARKS = 1024,
    };
]]
do
    local IDLE = 0
    local OBJECT_START = 1
    local KEY = 2
    local KEY_BACKSLASH = 3
    local KEY_DONE = 4
    local COLON = 5
    local VALUE_STRING = 6
    local VALUE_STRING_BACKSLASH = 7
    local VALUE_NUMBER = 8
    local VALUE_DONE = 9
    local COMMA = 10
    local MAX_STATE = 11

    local table_size = 256 * MAX_STATE
    comet_parse_table = ffi.new('uint8_t[?]', table_size, C.COMET_PARSE_ERROR)

    local function trans(from_state, to_state, flags, ...)
        local a = {...}
        local f = bit.bor(to_state, flags)
        if #a == 2 and type(a[1]) == 'number' and type(a[2]) == 'number' then
            for c = a[1], a[2] do
                comet_parse_table[from_state * 256 + c] = f
            end
        elseif #a == 1 and type(a[1]) == 'table' then
            for _, c in ipairs(a[1]) do
                comet_parse_table[from_state * 256 + c] = f
            end
        else
            error('Bad trans() call')
        end
    end

    local whitespace = {0x0A, 0x0D, 0x20}
    local digits = {0x30, 0x31, 0x32, 0x33, 0x34, 0x35, 0x36, 0x37, 0x38, 0x39}
    local plus_minus = {0x2B, 0x2D}
    local dot = {0x2E}
    local anycase_e = {0x45, 0x65}
    local double_quote = {0x22}
    local colon = {0x3A}
    local comma = {0x2C}
    local backslash = {0x5C}
    local left_curly = {0x7B}
    local right_curly = {0x7D}
    local end_packet = {0x04}

    local mark_prev = bit.lshift(C.COMET_MARK_PREV, C.COMET_OP_SHIFT)
    local mark_this = bit.lshift(C.COMET_MARK_THIS, C.COMET_OP_SHIFT)
    local mark_next = bit.lshift(C.COMET_MARK_NEXT, C.COMET_OP_SHIFT)
    local mark_backslash = bit.lshift(C.COMET_MARK_BACKSLASH, C.COMET_OP_SHIFT)

    trans(IDLE, IDLE, 0, whitespace)
    trans(IDLE, OBJECT_START, 0, left_curly)
    trans(IDLE, IDLE, C.COMET_PARSE_DONE, end_packet)

    trans(OBJECT_START, OBJECT_START, 0, whitespace)
    trans(OBJECT_START, KEY, mark_next, double_quote)
    trans(OBJECT_START, IDLE, 0, right_curly)

    trans(KEY, KEY, 0, 0x00, 0xFF)
    trans(KEY, KEY_BACKSLASH, mark_backslash, backslash)
    trans(KEY, KEY_DONE, mark_prev, double_quote)

    trans(KEY_BACKSLASH, KEY, 0, 0x00, 0xFF)

    trans(KEY_DONE, KEY_DONE, 0, whitespace)
    trans(KEY_DONE, COLON, 0, colon)

    trans(COLON, COLON, 0, whitespace)
    trans(COLON, VALUE_STRING, mark_next, double_quote)
    trans(COLON, VALUE_NUMBER, mark_this, digits)
    trans(COLON, VALUE_NUMBER, mark_this, plus_minus)

    trans(VALUE_STRING, VALUE_STRING, 0, 0x00, 0xFF)
    trans(VALUE_STRING, VALUE_STRING_BACKSLASH, mark_backslash, backslash)
    trans(VALUE_STRING, VALUE_DONE, mark_prev, double_quote)

    trans(VALUE_STRING_BACKSLASH, VALUE_STRING, 0, 0x00, 0xFF)

    -- KLUDGE too-lenient number parsing!!!
    trans(VALUE_NUMBER, VALUE_NUMBER, 0, digits)
    trans(VALUE_NUMBER, VALUE_NUMBER, 0, dot)
    trans(VALUE_NUMBER, VALUE_NUMBER, 0, anycase_e)
    trans(VALUE_NUMBER, VALUE_NUMBER, 0, plus_minus) -- E-013, e+12 ...
    trans(VALUE_NUMBER, VALUE_DONE, mark_prev, whitespace)
    trans(VALUE_NUMBER, COMMA, mark_prev, comma)
    trans(VALUE_NUMBER, IDLE, mark_prev, right_curly)

    trans(VALUE_DONE, VALUE_DONE, 0, whitespace)
    trans(VALUE_DONE, COMMA, 0, comma)
    trans(VALUE_DONE, IDLE, 0, right_curly)

    trans(COMMA, COMMA, 0, whitespace)
    trans(COMMA, KEY, mark_next, double_quote)
end

local marks = ffi.new('int[?]', C.COMET_MAX_MARKS)

local function comet_parse(buf, len, out)
    local mk_i = 0
    local state = 0
    local consumed = 0
    local backslash = false
    for i = 0, len - 1 do
        state = comet_parse_table[buf[i] +
            bit_lshift(bit_band(C.COMET_STATE_MASK, state), 8)]
        local flags = bit_band(C.COMET_OP_MASK, state)
        if flags ~= 0 then
            if state < C.COMET_PARSE_DONE then -- KLUDGE
                local op = bit_rshift(flags, C.COMET_OP_SHIFT)
                if op == C.COMET_MARK_BACKSLASH then
                    backslash = true
                else
                    marks[mk_i] = i + op - 2
                    if backslash then
                        marks[mk_i] = -marks[mk_i]
                        backslash = false
                    end
                    mk_i = mk_i + 1
                end
            elseif state == C.COMET_PARSE_DONE then
                assert(bit_band(mk_i, 3) == 0)
                for j = 0, mk_i - 1, 4 do
                    local s1, e1 = marks[j], marks[j + 1]
                    local s2, e2 = marks[j + 2], marks[j + 3]
                    if e1 > 0 and e2 > 0 then
                        local key = ffi_string(buf + s1, e1 - s1 + 1)
                        out[key] = ffi_string(buf + s2, e2 - s2 + 1)
                    else
                        e1 = math.abs(e1)
                        e2 = math.abs(e2)
                        -- backslash special case
                        -- We know both are strings, so include leading
                        -- and trailing double quotes for cjson.decode()
                        local key = ffi_string(buf + s1 - 1, e1 - s1 + 3)
                        local val = ffi_string(buf + s2 - 1, e2 - s2 + 3)
                        key = cjson.decode(key)
                        out[key] = cjson.decode(val)
                    end
                end
                mk_i = 0
                state = 0
                consumed = i + 1
            elseif state == C.COMET_PARSE_ERROR then
                return consumed, false
            end
        end
    end
    return consumed, true
end

function m:comet_proxy_reader(loop, w, pxy)
    local buf = pxy.buf
    buf:assure_available()
    local got = C.read(pxy.socket, buf:tail(), buf:available())
    if got <= 0 then
        -- KLUDGE error here?
        self.close(loop, w)
        self.comet_proxies_by_socket[pxy.socket] = nil
        self.comet_proxies_by_name[pxy.name] = nil
        return
    end
    buf.size = buf.size + got

    if not pxy.got_headers then
        local http, end_pos = http_parser.response:match(tostring(buf))
        if not http then
            return
        end
        buf:remove(0, end_pos - 1)
        pxy.got_headers = true
    end

    local len = #buf
    local consumed, ok = comet_parse(buf.buf, len, self.throttled_live_data)
    if not ok then
        -- KLUDGE handle errors somehow here...
    end
    if consumed == len then
        buf:clear()
    else
        buf:remove(0, consumed)
    end
end

function m:connect_proxy(name, proxy_ip, proxy_port, channels)
    local pxy = {
        name = name,
        buf = membuf:new(8192),
        got_headers = false,
    }
    local proxy_w = self:connect(proxy_ip, proxy_port,
        function(loop, w) self:comet_proxy_reader(loop, w, pxy) end)
    if not proxy_w then
        -- KLUDGE emit some kind of better error message here
        log:error(__"Couldn't connect to proxy server")
        return nil
    end
    pxy.watcher = proxy_w
    pxy.socket = proxy_w.socket
    self.comet_proxies_by_socket[pxy.socket] = pxy
    self.comet_proxies_by_name[name] = pxy
    local json = cjson.encode(channels) .. "\n"
    local req = "GET /comet_stream HTTP/1.0\r\n"
        .. "Content-type: application/json\r\n"
        .. "Content-length: " .. #json .. "\r\n"
        .. "\r\n"
        .. json
    C.write(pxy.socket, req, #req)
end

function m:live_updater(loop, timer)
    if self.fake_live_data then
        self.fake_live_data(self.throttled_live_data,
            epoch_timestamp() - self.base_time)
    end

    -- WTF?  Basically in a nutshell, only dispatch stuff to clients when there
    -- is actually something to send, otherwise, don't
    for _, _ in pairs(self.throttled_live_data) do
        local json = cjson.encode({{'#live_data', self.throttled_live_data}})
        json = string.gsub(json, '(@mx%d+b%-%x+)%..-"', '%1"')
        self.throttled_live_data = {}
        self:dispatch_to_clients(json)
        break
    end
end

local function send_response_headers(w, code, http_msg, content_type, fmt, ...)
    local extras = #w.out_headers > 0 and table.concat(w.out_headers) or ''
    local msg =
        "HTTP/1.1 " .. code .. " " .. http_msg .. "\n"
        .. "Cache-Control: no-cache, no-store, must-revalidate\n"
        .. "Pragma: no-cache\n"
        .. "Expires: -1\n"
        .. "Content-Type: " .. content_type .. "\n"
        .. extras
        .. "\n"
        .. (#{...} == 0 and fmt or string.format(fmt, ...))
    while true do
        local amt, estr, eno = u.write(w.socket, msg)
        if not amt then
            if eno ~= u.EAGAIN then return end
        elseif amt == msg:len() then return
        else msg = string.sub(msg, amt + 1)
        end
    end
end

function send_response(w, code, http_msg, content_type, fmt, ...)
    send_response_headers(w, code, http_msg, content_type, fmt, ...)
    w.server.close(w.server.ev_loop, w)
end

local function redirect(w, code, http_msg, location)
    u.write(w.socket,
        "HTTP/1.1 " .. code .. " " .. http_msg .. "\n"
        .. "Location: " .. location .. "\n"
        .. "\n"
    )
    w.server.close(w.server.ev_loop, w)
end

function m:start_comet_stream(loop, w, begin_func)
    local op = w.headers.REQUEST_METHOD
    if op == 'POST' and w.content then
        self.fake_comet_setup[w.session_key] = assert(cjson.decode(w.content))
        send_response(w, 200, "OK", "text/plain", "")
        return
    else
        w.posted = self.fake_comet_setup[w.session_key] or {}
    end
    if w.posted.fake_live_data then
        self.fake_live_data =
            (loadstring('return ' .. w.posted.fake_live_data))()
    end

    if #self.comet_clients == 0 then
        local rate = tonumber(w.query.update_rate or w.posted.update_rate)
        self.live_update_rate = rate or 5
        self.live_w = function(t_loop, t_timer)
            self:live_updater(t_loop, t_timer)
        end
        local interval = 1.0 / self.live_update_rate
        self.live_data_timer = ev.Timer.new(self.live_w, interval, interval)
        self.live_data_timer:start(ev_loop)
    end

    self[begin_func](self, w)
    self.comet_clients[#self.comet_clients + 1] = w
    self.base_time = self.base_time or epoch_timestamp()

    -- Propagate current conditions to our Comet client
    -- KLUDGE horrifically inefficient
    if self.network then
        self.start_comet_client = #self.comet_clients
        self:changed_rr({}, self.network:QueryRR('*', '*', '*', '*'), true)
        self:dispatch_to_clients(cjson.encode({{'#block_end'}}))
        self.start_comet_client = nil
    end

    if w.posted.fake_preamble then
        self:write_comet(w, w.posted.fake_preamble)
    end

    w.cleanup = function() self:cleanup_comet_stream(loop, w) end
end

function m:cleanup_comet_stream(loop, w)
    for i = 1, #self.comet_clients do
        if self.comet_clients[i] == w then
            table.remove(self.comet_clients, i)
            break
        end
    end
    if #self.comet_clients == 0 then
        for name, pxy in pairs(self.comet_proxies_by_name) do
            self.close(loop, pxy.watcher)
        end
        self.comet_proxies_by_name = {}
        self.comet_proxies_by_socket = {}
        self.live_data_timer:stop(loop)
        self.live_data_timer = nil
    end
end

handlers['stream.comet'] = function(loop, w)
    w.server:start_comet_stream(loop, w, 'begin_comet')
end

handlers['stream.eventsource'] = function(loop, w)
    w.server:start_comet_stream(loop, w, 'begin_eventsource')
end

local function out_sorted(out, t, fmt, stringify)
    if type(t) ~= 'table' then
        return
    end
    stringify = stringify or tostring
    local heads = {}
    for k, _ in pairs(t) do
        heads[#heads + 1] = k
    end
    table.sort(heads)
    for _, head in ipairs(heads) do
        out[#out + 1] = string.format(fmt, tostring(head), stringify(t[head]))
    end
end

function handlers.echo(loop, w, frag)
    local out = {}
    out[#out + 1] = string.format("frag=%s\n", frag)
    out_sorted(out, w.headers, "header %s => %s\n")
    out[#out + 1] = "\n"
    out[#out + 1] = string.format("accept_language: %s\n",
        cjson.encode(w.accept_language))
    out_sorted(out, w.content_type, "content_type %s => %s\n")
    out_sorted(out, w.query, "query %s => %s\n")
    out_sorted(out, w.cookies, "cookie %s => %s\n")
    out[#out + 1] = 'CONTENT_VALUES:\n'
    out_sorted(out, w.content_values, "content_values %s => %s\n")
    out[#out + 1] = 'MULTIPART:\n'
    out_sorted(out, w.multipart, "multipart %s => %s\n", cjson.encode)
    out[#out + 1] = 'END MULTIPART:\n'
    out[#out + 1] = "\n\n==== DEBUG ====\n\n"
    if w.debug then
        for _, v in ipairs(w.debug) do
            out[#out + 1] = v
        end
    end
    out[#out + 1] = "\n\n==== CONTENT ====\n\n"
    local content = string.gsub(w.content, "([\r\n])", function(c)
        return string.format("%%%02X%s", string.byte(c), c == "\n" and c or "")
    end)
    out[#out + 1] = content
    send_response(w, 200, "OK", "text/plain", table.concat(out))
end

function handlers.delete_file(loop, w, f)
    local files = {}
    if w.headers.REQUEST_METHOD == 'POST' and w.content and
        not f:find('^/.*%.sie$') then
        files = cjson.decode(w.content) or {}
    else
        files[#files + 1] = f:match('^/(.*)$')
    end

    local removed_files, nonremoved_files = sie_file_mgr.delete_sie_files(files)
    local response = {
        result = (#nonremoved_files == 0 and #removed_files > 0) and
            'success' or 'fail',
        removed_files = removed_files,
        nonremoved_files = nonremoved_files,
    }
    send_response(w, 200, "OK", "application/json", cjson.encode(response))
end

local function html_escape(str)
    return str:gsub('[&"<]', {
                        ['&'] = '&amp;',
                        ['"'] = '&quot;',
                        ['<'] = '&lt;'
                    })
end

function handlers.get_log(loop, w, frag)
    local entries = { '<th>Time</th><th>Component</th>' ..
                      '<th>Level</th><th>Message</th>'}
    local db = sqlite3.open('/var/lib/armada/log.db')
    for r in db:rows([[
        SELECT time, component, level, message FROM log
               ORDER BY rowid DESC LIMIT 50
    ]]) do
        table.insert(entries, string.format(
            '<tr><td>%s</td><td>%s</td><td>%s</td><td>%s</td></tr>',
            html_escape(r[1]), html_escape(r[2]),
            html_escape(r[3]), html_escape(r[4])))
    end
    db:close()
    send_response(w, 200, "OK", "text/html", "<table>%s</table>",
                  table.concat(entries))
end

local function db_error(loop, w, status, message, result)
    result = result or {ok=false}
    send_response(w, status, message, "application/json", cjson.encode(result))
end

local function sql_raw(db, sql)
    local ok, result = pcall(function() return db:exec(sql) end)
    if ok then
        return {ok=true}, ok
    else
        return {ok=false, error_code=result}, ok
    end
end

sql_exec = function(db, table_name, sql, args)
    local query = string.format(sql, table_name):gsub('^%s*(.-)%s$', '%1')
    local ok, result = pcall(function() return db:exec(query, args) end)
    if ok then
        return result, ok
        -- KLUDGE broadcast results to listeners based on table names here?
    else
        return {ok=false, error=result}, ok
    end
end

local function sql_do(db, tbl_name, sql)
    local query = sql:format(tbl_name):gsub('^%s*(.-)%s$', '%1')
    local ok, result = pcall(function() return db:exec(query) end)
    if ok then
        return {ok=true}, ok
    else
        return {ok=false, error_code=result}, ok
    end
end

function handlers.raw_db(loop, w, frag)
    local op = w.headers.REQUEST_METHOD
    local ok, result, ignore = false, nil, nil

    local db_name, doc = frag:match("^/([%w%_%-]+)/([%w%_%-]+)$")
    if not db_name then
        db_name = frag:match("^/([%w%_%-]+)/?$")
        doc = ""
    end
    local tbl_name = "rdb_" .. db_name

    if not db_name then
        return db_error(loop, w, 404, "No such database", result)
    end

    local db = w.server.setup_db
    ignore, ok = sql_raw(db, "BEGIN TRANSACTION;")

    local sec, nsec = u.clock_gettime()
    local stamp = string.format("(%022.0f)", sec * 1e9 + nsec)

    if doc ~= "" then -- Document level
        if op == "GET" then
            if doc:match("^_") then
                if doc == "_all_docs" then
                    local r
                    r, ok = sql_exec(db, tbl_name, [[
                        SELECT content FROM "%s";
                    ]])
                    if ok then
                        result = {}
                        for _, v in ipairs(r) do
                            result[#result + 1] = v.content
                        end
                        result = "[" .. table.concat(result, ",\n") .. "]"
                    else
                        result = {ok=false, error='No such database'}
                    end
                end
            else
                local got, ok = sql_exec(db, tbl_name, [[
                    SELECT * FROM "%s" WHERE id = ?;
                ]], {doc})
                if ok and #got == 1 and got[1].content then
                    result = got[1].content
                else
                    result = {ok=false, error="No such document"}
                end
            end
        elseif op == "PUT" then
            result, ok = sql_exec(db, tbl_name, [[
                UPDATE "%s" SET content = ?, updated = ? WHERE id = ?;
            ]], {w.content, stamp, doc})
        elseif op == "POST" then
            result, ok = sql_exec(db, tbl_name, [[
                INSERT INTO "%s" (id, content, updated) VALUES (?, ?, ?);
            ]], {doc, w.content, stamp})
        elseif op == "DELETE" then
            result, ok = sql_exec(db, tbl_name, [[
                DELETE FROM "%s" WHERE id = ?;
            ]], {doc})
        end
    else -- Database level
        if op == "GET" and db_name == "_all_dbs" then
            result, ok = sql_exec(db, "sqlite_master", [[
                SELECT substr(name, 5) AS name
                FROM "%s"
                WHERE type = 'table' AND name LIKE 'rdb*_%%' escape '*';
            ]])
            if ok then
                for i, v in pairs(result) do
                    result[i] = v.name
                end
            end
        elseif op == "PUT" then
            result, ok = sql_do(db, tbl_name, [[
                CREATE TABLE "%s" (
                    id TEXT NOT NULL,
                    content TEXT NOT NULL,
                    updated INTEGER NOT NULL DEFAULT 0
                );
            ]])
        elseif op == "DELETE" then
            result, ok = sql_do(db, tbl_name, [[
                DROP TABLE "%s";
            ]])
        end
    end

    ignore, ok = sql_raw(db, ok and "COMMIT;" or "ROLLBACK;")

    if type(result) == 'table' then
        result = cjson.encode(result)
    end

    send_response(w, ok and 200 or 500, ok and "OK" or "ERROR",
        "application/json", result)
end

local function json_concat(r)
    local result = {}
    for _, v in ipairs(r) do
        result[#result + 1] = v.content
    end
    result = "[" .. table.concat(result, ",\n") .. "]"
    return result
end

local function deep_equal(a, b)
    if a == b then
        return true
    end
    if type(a) ~= type(b) then
        return false
    end
    if type(a) == 'table' then
        for k, v in pairs(a) do
            if not deep_equal(v, b[k]) then
                return false
            end
        end
        for k, v in pairs(b) do
            if not a[k] then
                return false
            end
        end
        return true
    end
    return false
end

local function table_delta(old, new)
    local del, add = {}, {}
    for k, v in pairs(old) do
        if not deep_equal(v, new[k]) then
            del[k] = v
            add[k] = (new[k] == nil or new[k] == cjson.null)
                and cjson.null or v
        end
    end
    for k, v in pairs(new) do
        if not old[k] then
            del[k] = cjson.null
            add[k] = v
        end
    end
    return del, add
end

function handlers.tracked_db(loop, w, frag)
    local op = w.headers.REQUEST_METHOD
    local ok, result, ignore = false, nil, nil

    local _1, schema, db_name, doc
        = frag:match("^/(([%w%_%-%/]+)/)([%w%_%-]+)/([%w%_%-%.%@%~%%]*)$")

    if not schema then
        doc = ""
        _1, schema, db_name
            = frag:match("^/(([%w%_%-%/]+)/)([%w%_%-]*)$")
    end

    if not schema then
        return db_error(loop, w, 404, "Malformed path", result)
    end

    schema = schema:gsub("%/+$", "")
    doc = w.server.url_unescape(doc)

    local tbl_name = string.format("tsd_%s", schema)
    local history_tbl_name = string.format("tsh_%s", schema)

    if not db_name then
        return db_error(loop, w, 404, "No such database", result)
    end

    local db = w.server.setup_db
    ignore, ok = sql_raw(db, "BEGIN TRANSACTION;")

    local stamp = iso_timestamp()
    local description = w.query.description
    local track_history = false
    local comet_updates = nil

    if doc ~= "" then -- Document level
        if op == "GET" then
            if doc:match("^_") then
                if doc == "_all_docs" then
                    local r
                    r, ok = sql_exec(db, tbl_name, [[
                        SELECT content
                        FROM "%s"
                        WHERE db_name = ? AND id != '' AND deleted = '';
                    ]], {db_name})
                    if ok and #r > 0 then
                        if w.api_internal.export then
                            local setup = {}
                            for i = 1, #r do
                                setup[i] = cjson.decode(r[i].content)
                            end
                            local ss = setup_spec:new(setup)
                            local ok, err = pcall(function() result = 
                                ss:export("user_order") end)
                            if not ok then
                                --ok is true because we want a 200 return code,
                                --since error is probably not network related
                                result = {ok=true, error_message=err}
                            end
                            downloadCompleteNotification(w.session_key)
                        else
                            local o = {'['}
                            for i = 1, #r do
                                if i > 1 then
                                    o[#o + 1] = ',\n'
                                end
                                o[#o + 1] = r[i].content
                            end
                            o[#o + 1] = ']\n'
                            result = table.concat(o)
                        end
                    else
                        result = {ok=false, error='No such database'}
                    end
                elseif doc == "_historical" then
                    local r
                    local at = w.query.at or iso_timestamp()
                    r, ok = sql_exec(db, tbl_name, [[
                        SELECT content
                        FROM "%%s"
                        WHERE db_name = ? AND id != ''
                            AND added <= ?
                            AND (deleted = '' OR deleted > ?);
                    ]], {db_name, at, at})
                    if ok then
                        result = json_concat(r)
                    else
                        result = {ok=false, error='No such database'}
                    end
                elseif doc == "_most_recent" then
                    local r
                    local extras = ''
                    if w.query.limit then
                        extras = 'LIMIT ' .. w.query.limit
                    end
                    local match_id = w.query.match_id or ''
                    r, ok = sql_exec(db, tbl_name, string.format([[
                        SELECT content, added
                        FROM "%%s"
                        WHERE db_name = ? AND id != '' AND deleted = ''
                        ORDER BY added DESC
                        %s;
                    ]], extras), {w.session.user_id}) -- KLUDGE ???
                    if ok then
                        result = {}
                        for _, v in ipairs(r) do
                            local e = cjson.decode(v.content)
                            e._timestamp = v.added
                            result[#result + 1] = e
                        end
                    else
                        result = {ok=false, error='No such database'}
                    end
                end
            else
                local got, ok = sql_exec(db, tbl_name, [[
                    SELECT *
                    FROM "%s"
                    WHERE db_name = ? AND id = ? AND deleted = '';
                ]], {db_name, doc})
                if ok and #got == 1 and got[1].content then
                    result = got[1].content
                else
                    result = {ok=false, error="No such document"}
                end
            end
        elseif op == "POST" and doc == "_delta" then
            local payload = cjson.decode(w.content)
            local d_old, d_new = payload.old, payload.new
            if w.api_internal.import then
                description = description or 'Import setup'
            end
            description = description or payload.description
            local err = nil
            if type(d_old) ~= 'table' or type(d_new) ~= 'table' then
                err = "Missing or non-{} old or new field"
            end
            if not err then
                -- Check that old and new have exactly the same set of keys
                for k, v in pairs(d_old) do
                    if not d_new[k] then
                        err = k .. " in old not in new" -- KLUDGE I18N
                        break
                    end
                end
                if not err then
                    for k, v in pairs(d_new) do
                        if not d_old[k] then
                            err = k .. " in new not in old" -- KLUDGE I18N
                            break
                        end
                    end
                end
            end
            if not err then
                -- First we check to make sure that the old state matches
                -- the current state of the database. Abort if not.
                -- KLUDGE add coalesce capability:
                --  if coalesce_flag_or_condition
                --      UPDATE %s SET content = ?
                --      WHERE db_name = ? AND id = ? AND deleted = ''
                local get_cur = db:prepare(string.format([[
                    SELECT content
                    FROM "%s"
                    WHERE db_name = ? AND id = ? AND deleted = '';
                ]], tbl_name))
                local do_delete = db:prepare(string.format([[
                    UPDATE "%s"
                    SET deleted = ?
                    WHERE db_name = ? AND id = ? AND deleted = '';
                ]], tbl_name))
                local do_insert = db:prepare(string.format([[
                    INSERT INTO "%s" (db_name, id, content, added)
                    VALUES (?, ?, ?, ?);
                ]], tbl_name))
                local cu = {}
                local changed = false
                for id, old in pairs(d_old) do
                    local new = d_new[id]
                    local deletable = nil
                    local insertable = nil
                    local got = get_cur:exec{db_name, id}
                    if type(got) ~= 'table' or #got > 1 then
                        err = string.format("Conflict at %s", id)
                        break
                    end
                    if old == cjson.null then
                        if #got == 0 then
                            insertable = new
                            cu[id] = new
                        else
                            err = string.format(
                                "Conflict at %s, insert on existing", id)
                            break
                        end
                    elseif new == cjson.null then
                        if #got > 0 then
                            deletable = id
                            cu[id] = cjson.null
                        else
                            err = string.format(
                                "Conflict at %s, delete on nonexisting", id)
                            break
                        end
                    else
                        local cur
                        if #got == 1 then
                            cur = cjson.decode(got[1].content)
                        elseif #got == 0 then
                            cur = {} -- KLUDGE WRONG BROKEN (???)
                        else
                            err = string.format(
                                "Internal error at %s, multiple existing", id)
                            break
                        end
                        -- We only match keys in delta, not in current
                        -- content, as it's a delta
                        for k, v in pairs(old) do
                            if not us.is_equal(cur[k], v) then
                                -- The following makes nil and "" equivalent
                                -- v will not be nil by definition
                                if not (cur[k] == nil
                                        and (v == "" or v == cjson.null)) then
                                    err = string.format(
                                        "Conflict at %s/%s [%s <=> %s]",
                                        id, k, tostring(cur[k]), tostring(v))
                                    break
                                end
                            end
                            if new[k] == cjson.null then
                                cur[k] = nil
                            else
                                cur[k] = new[k]
                            end
                        end
                        deletable = id
                        insertable = cur
                    end
                    if deletable then
                        do_delete:exec{stamp, db_name, deletable}
                        changed = true
                    end
                    if insertable then
                        insertable = cjson.encode(insertable)
                        do_insert:exec{db_name, id, insertable, stamp}
                        changed = true
                    end
                end
                if not err and changed then
                    comet_updates = {{'#tracked_db_delta', {
                        schema = schema, db = db_name,
                        source_session_key = w.session_key,
                        description = description,
                        old = d_old, new = d_new,
                    }}}
                end
                get_cur:finalize()
                do_delete:finalize()
                do_insert:finalize()
                result = {ok = not err, error = err}
            end
        elseif op == "POST" or op == "PUT" or op == "DELETE" then
            local d_old, d_new
            local got, ok = sql_exec(db, tbl_name, [[
                SELECT content
                FROM "%s"
                WHERE db_name = ? AND id = ? AND deleted = '';
            ]], {db_name, doc})
            if ok then
                if type(got) ~= 'table' or #got > 1 then
                    err = string.format("Internal error at %s", id)
                    ok = false
                else
                    d_old, d_new = {}, {}
                    d_new[doc] = cjson.null -- DELETE default case
                    d_old[doc] = (#got == 0)
                        and cjson.null
                        or cjson.decode(got[1].content)
                end
            end
            if ok then
                -- Here, PUT equivalent to POST
                result, ok = sql_exec(db, tbl_name, [[
                    UPDATE "%s"
                    SET deleted = ?
                    WHERE db_name = ? AND id = ? AND deleted = '';
                ]], {stamp, db_name, doc})
            end
            if ok and (op == "PUT" or op == "POST") then
                result, ok = sql_exec(db, tbl_name, [[
                    INSERT INTO "%s" (db_name, id, content, added)
                    VALUES (?, ?, ?, ?);
                ]], {db_name, doc, w.content, stamp})
                d_new[doc] = cjson.decode(w.content)
            end
            if ok then
                comet_updates = {{'#tracked_db_delta', {
                    schema = schema, db = db_name,
                    source_session_key = w.session_key,
                    description = description,
                    old = d_old, new = d_new,
                }}}

                -- If success AND this was a change to to any of the
                -- email information, signal all listeners that a
                -- change has occurred.
                if ok and op == "POST" and
                    (description == 'mail_server_change' or
                     description == 'mail_notification_change') then
                    w.server.master:SignalSystem(description, w.content)
                end
            end
            track_history = ok
        end
    else -- Database level
        if op == "GET" then
            if db_name == "_all_dbs" then
                local r
                local match_id = w.query.match_id or ''
                r, ok = sql_exec(db, tbl_name, [[
                    SELECT db_name, content
                    FROM "%s"
                    WHERE id = ? AND deleted = '';
                ]], {match_id})
                if ok then
                    result = {}
                    for _, v in ipairs(r) do
                        result[v.db_name] = cjson.decode(v.content)
                    end
                end
            elseif db_name == "_most_recent" then
                local r
                local extras = ''
                if w.query.limit then
                    extras = 'LIMIT ' .. w.query.limit
                end
                local match_id = w.query.match_id or ''
                r, ok = sql_exec(db, tbl_name, string.format([[
                    SELECT content, added
                    FROM "%%s"
                    WHERE deleted = ''
                    ORDER BY added DESC
                    %s;
                ]], extras), {})
                if ok then
                    result = {}
                    for _, v in ipairs(r) do
                        local e = cjson.decode(v.content)
                        r.timestamp = v.added
                        result[#result + 1] = e
                    end
                else
                    result = {ok=false, error='No such database'}
                end
            end
        elseif op == "PUT" then
            description = description
                or string.format("Create database '%s'", db_name)
            result, ok = sql_do(db, tbl_name, [[
                CREATE TABLE "%s" (
                    db_name TEXT NOT NULL,
                    id TEXT NOT NULL,
                    content TEXT NOT NULL,
                    added TEXT NOT NULL,
                    deleted TEXT NOT NULL DEFAULT ''
                );
            ]])
            if ok then
                result, ok = sql_do(db, '', string.format([[
                    CREATE INDEX "%s_idx"
                    ON "%s" (db_name, id, deleted);
                ]], tbl_name, tbl_name))
            end
            if ok then
                result, ok = sql_do(db, history_tbl_name, [[
                    CREATE TABLE "%s" (
                        db_name TEXT NOT NULL,
                        stamp TEXT NOT NULL,
                        author TEXT NOT NULL DEFAULT '',
                        description TEXT NOT NULL DEFAULT ''
                    );
                ]])
            end
            if ok then
                result, ok = sql_do(db, '', string.format([[
                    CREATE INDEX "%s_idx"
                    ON "%s" (db_name, stamp);
                ]], history_tbl_name, history_tbl_name))
            end
            if ok then
                result, ok = sql_exec(db, tbl_name, [[
                    INSERT INTO "%s" (db_name, id, content, added)
                    VALUES (?, ?, ?, ?);
                ]], {db_name, "", w.content, stamp})
            end
            if ok then
                comet_updates = {
                    {'#tracked_db_create', {schema=schema, db=db_name}}
                }
            end
            track_history = ok
        elseif op == "DELETE" then
            local db_names
            if w.content and cjson.decode(w.content) then
                db_names = cjson.decode(w.content)
            else
                db_names = {db_name}
            end
            ok = true
            db_names = db_names or {}
            local _deleted, _nondeleted = {}, {}
            for _, db_name in ipairs(db_names) do
                -- get records that will be deleted
                -- will be sent over comet stream
                local deleted, lok = sql_exec(db, tbl_name, [[
                    SELECT * FROM "%s"
                    WHERE db_name = ? AND deleted = '';
                ]], {db_name})

                if not lok or #deleted == 0 then
                    _nondeleted[#_nondeleted + 1] = db_name
                else
                    -- This allows undo. Maybe add a 'force' query flag to
                    -- actually delete rows?
                    local res, lok = sql_exec(db, tbl_name, [[
                        UPDATE "%s" SET deleted = ?
                        WHERE db_name = ? AND deleted = '';
                    ]], {stamp, db_name})
                    if lok then
                        _deleted[#_deleted + 1] = db_name
                        comet_updates = {
                            {
                                '#tracked_db_delete',
                                {
                                    schema = schema,
                                    db = db_name,
                                    deleted = deleted
                                }
                            }
                        }
                        w.server:dispatch_to_clients(comet_updates)
                    end
                end
            end
            result = {
                requested = db_names,
                deleted = _deleted,
                nondeleted = _nondeleted,
            }
            comet_updates = nil
        end
    end

    if track_history then
        local author = w.query.author -- KLUDGE user_id, or ??? Backdoor???
        if not author and w.session then
            author = w.session.user_id
        end
        result, ok = sql_exec(db, history_tbl_name, [[
            INSERT INTO "%s" (db_name, stamp, author, description)
            VALUES (?, ?, ?, ?);
        ]], {db_name, stamp, author or "", description or ""})
        if ok then
            result = {ok=true}
        end
    end
    ignore, ok = sql_raw(db, ok and "COMMIT;" or "ROLLBACK;")

    if ok and comet_updates then
        w.server:dispatch_to_clients(comet_updates) -- KLUDGE not selective
    end

    if w.api_internal.import then
        return {ok = ok, result = result}
    end

    if type(result) == 'table' then
        result = cjson.encode(result)
    end

    if not result or type(result) ~= 'string' then
        result = cjson.encode{
            ok = false,
            msg = 'The request was not processed',
            url = frag,
        }
        ok = false
    end

    send_response(w, ok and 200 or 500, ok and "OK" or "ERROR",
        "application/json", result)
end

local function _sign_in(loop, w, frag)
    local user_name, password = w.query.user_name, w.query.password
    local status = nil -- NOT, Repeat, NOT 401 Unauthorized

    if type(user_name) ~= "string" or type("password") ~= "string" then
        error('Missing user_name or password')
    end
    if user_name:len() == 0 then
        error('Empty user_name')
    end

    local db = w.server.setup_db
    local got = sql_exec(db, "tsd_users", [[
        SELECT * FROM "%s" WHERE db_name = 'users' AND id = ? AND deleted = '';
    ]], {w.query.user_name or ""})

    if #got < 1 then
        error('No user entry for user_name')
    elseif #got > 1 then
        error('Too many user entries for user_name')
    elseif not got[1].content then
        error('No user record for user_name')
    end

    local user = cjson.decode(got[1].content)
    if type(user) ~= "table" or type(user.password_sha) ~= "string" then
        error('Bad user entry for user_name')
    end

    local salt = user.salt or '' -- KLUDGE
    local hash = bits.tohex(crypto.SHA1(password .. salt))
    if hash ~= user.password_sha then
        if user_name == 'admin' and password == 'password' then
            error('login via default admin password failure')
        end
        error('Incorrect password')
    end

    if not w.authorization_key then
        w.authorization_key = uuid.generate()
    end

    user.password_sha, user.salt = nil, nil -- Clear out private stuff
    local session_key = uuid.generate()
    w.server.session_db:update_session{
        session_key = session_key,
        authorization_key = w.authorization_key,
        timestamp = epoch_timestamp(),
        user_id = user.id,
        state = {user=user, hostname=hostname}, -- KLUDGE
    }

    local result = {
        ok = true,
        session_key = session_key,
        authorization_key = w.authorization_key,
        url = "/console.html", -- KLUDGE make url reflect user history
    }

    local cookie = "\nSet-Cookie: authorization_key=" .. w.authorization_key
--        .. "; Secure"

    send_response(w, 200, "OK", "application/json" .. cookie,
        cjson.encode(result))
end

function handlers.hash(loop, w, frag)
    local s = string.sub(frag, 2)
    local ret = bits.tohex(crypto.SHA1(s))
    send_response(w, 200, "OK", "application/json", cjson.encode({s, ret}))
end

function handlers.sign_in(loop, w, frag)
    -- This is a special case to make sure that internal information
    -- doesn't leak outside during the authorization process
    local ok, result = pcall(function() return _sign_in(loop, w, frag) end)
    if not ok then
        if not result:find('login via default admin password failure') then
            log(log.WARNING, __'/api/sign_in error: "{0}"\nTraceback: {1}\n',
                result, debug.traceback())
        end
        result = {ok=false, error="Forbidden"}
        send_response(w, 403, "Forbidden", "application/json",
            cjson.encode(result))
    end
end

function handlers.sign_out(loop, w, frag)
    local result = false
    if w.query.session_key then
        w.server.session_db:end_session(w.query.session_key)
        result = true;
    end
    send_response(w, 200, "OK", "application/json", cjson.encode({result=result}))
end

function handlers.update_settings(loop, w, frag)
    local op = w.headers.REQUEST_METHOD
    local up = assert(cjson.decode(w.content))
    for socket, client in pairs(w.server.comet_clients) do
        if client.session_key == w.session_key then
            for _, update in ipairs(up.updates) do
                client.server:write_comet(client, {{'#console_log', update}})
                local connector = update.connector:gsub("^.-@", "")
                local host, port, ip
                    = w.server.network:ResolveService(connector)
                if host then
                    local cb_w = w.server:connect(ip, port,
                        function(loop, w)
                            local got = u.read(w.socket, 8192)
                            if not got or got == "" then
                                w.server.close(loop, w)
                                return
                            end
                        end
                    )
                    local json = cjson.encode(update) .. "\n"
                    local req = "POST /update_settings/ HTTP/1.0\r\n"
                        .. "Content-type: application/json\r\n"
                        .. "Content-length: " .. #json .. "\r\n"
                        .. "\r\n"
                        .. json
                    u.write(cb_w.socket, req)
                end
            end
        end
    end
    send_response(w, 200, "OK", "application/json", cjson.encode({ok=1}))
end

function handlers.session_data(loop, w, frag)
    local session = w.server.session_db:get_session(w.session_key) -- KLUDGE ? nil case?
    send_response(w, 200, "OK", "application/json",
        cjson.encode(session.state))
end

-- Notice: This function will only return connected sessions. Stale sessions will no be included
function handlers.dump_sessions(loop, w, frag)
    local connected_sessions = {}
    -- get all sessions
    local sessions = w.server.session_db:get_all_sessions()
    for i = w.server.start_comet_client or 1, #w.server.comet_clients do
        local session_key = w.server.comet_clients[i].session_key
        -- include only the ones with connected sessions
        for j = 1, #sessions do
            local session = sessions[j]
            if session_key == session.session_key then
                table.insert(connected_sessions, session)
            end
        end
    end
    send_response(w, 200, "OK", "application/json",
        cjson.encode(connected_sessions))
end

local api_url_parser
do
    local C, Cc, Cg, Ct, P = lpeg.C, lpeg.Cc, lpeg.Cg, lpeg.Ct, lpeg.P
    local V = lpeg.V
    local uuid_digit = lpeg.R('09', 'af', 'AF')

    local function method(s)
        return Cg(P(s), 'method') * Cg(Cc(true), string.lower(s)) * P':/'
    end
    local function prefix(s)
        return Cg(C(s), 'prefix') * P'/'
    end
    local function action(s, action)
        if action then
            return P(s) * Cg(Cc(action), 'action')
        else
            return Cg(P(s), 'action')
        end
    end
    local function d(n)
        return P(uuid_digit^n - uuid_digit^(n+1))
    end

    api_url_parser = P{'api',
        base = P'api/',
        delete = method('DELETE'),
        get = method('GET'),
        post = method('POST'),
        put = method('PUT'),
        fake_uuid = Cg(P(P(1) - P'/')^1, 'id'),
        uuid = Cg(P(
            d(8) * '-' * d(4) * '-' * d(4) * '-' * d(4) * '-' * d(12)
          ) / string.lower, 'id'),
        api = V'session_api' + V'sie_api' + V'system_api' + V'test_api',
        session_prefix = V'base' * prefix('session'),
        session_api = Ct(
            V'get' * V'session_prefix' * action('state')
          + V'get' * V'base' * Cg(C'session', 'prefix')
                * Cg(Cc'session', 'action')
        ) * -P(1),
        sie_prefix = V'base' * prefix('sie'),
        sie_id = V'sie_prefix' * V'fake_uuid',
        sie_api = Ct(
            V'get' * V'sie_prefix' * V'uuid' * ( 
                action('.setup', 'export')
              + action('.setup', 'extract')
            )
          + V'post' * V'sie_prefix' * V'uuid' * (
                action('.setup', 'extract')
            )
          + V'get' * V'sie_id' * P'/'
              * Cg(C((P(1) - '.setup')^1 * '.setup'), 'filename')
              * (
                action('', 'export')
              + action('', 'extract')
            )
          + V'post' * V'sie_id' * P'/'
              * Cg(C((P(1) - '.setup')^1 * '.setup'), 'filename')
              * (
                action('', 'extract')
            )
          + V'get' * V'sie_id' * '/' * (
                action('export')
              + action('extract')
            )
              * ('/' * Cg(P(1)^0, 'filename'))^-1
          + V'post' * V'sie_id' * '/' * (
                action('extract')
            )
              * ('/' * Cg(P(1)^0, 'filename'))^-1
        ) * -P(1),
        system_prefix = V'base' * prefix('system'),
        system_api = Ct(
            V'get' * V'system_prefix' * (
                action('version')
              + action('_dev_config')
              + action('_init_sie')
              + action('get_oscillator_offset')
              + action('get_datetime')
              + action('log')
              + action('get_errors')
              + action('available_channels')
              + action('export_log')
              + action('unlock_start')
              + action('cancel_restart')
              + action('get_cal_cert')
              + action('get_ca')
              + action('setupdb_status')
              + action('delete_databases')
              + action('customer_service_package')
              + action('status')
              + action('platform_info')
              + action('update_firmware')
              + action('led_control')
              + action('do_fpga_recovery')
              + action('reboot')
              + action('push_notification')
            )
          + V'post' * V'system_prefix' * (
                action('update_firmware')
              + action('alias')
              + action('set_datetime')
              + action('acknowledge_notify')
              + action('reboot')
              + action('control')
              + action('set_cal_cert')
              + action('set_inter_trig')
              + action('echo_local_file')
              + action('configure_channel')
            )
          + V'delete' * V'system_prefix' * (
                action('log')
              + action('sdbx')
            )
        ) * -P(1),
        test_prefix = V'base' * prefix('test'),
        test_id = V'test_prefix' * V'fake_uuid',
        test_api = Ct(
            V'get' * V'test_prefix' * V'uuid' * action('.sie', 'data')
          + V'get' * V'test_id' * P'/'
              * Cg(C((P(1) - '.sie')^1 * '.sie'), 'filename')
              * action('', 'data')
          + V'get' * V'test_prefix' * action('data_file_list')
          + V'get' * V'test_prefix' * action('rename')
          + V'get' * V'test_prefix' * V'uuid' * action('.setup', 'export')
          + V'get' * V'test_id' * P'/'
              * Cg(C((P(1) - '.setup')^1 * '.setup'), 'filename')
              * action('', 'export')
          + V'get' * V'test_id' * '/' * action('export')
              * ('/' * Cg(P(1)^0, 'filename'))^-1
          + V'get' * V'test_id' * P'/' * action('data')
              * ('/' * Cg(P(1)^0, 'filename'))^-1
          + V'put' * V'test_id' * action('', 'import')
          + V'put' * V'test_prefix' * action('import')
          + V'post' * V'test_prefix' * action('create')
          + V'post' * V'test_prefix' * action('start')
          + V'post' * V'test_prefix' * action('initialize')
          + V'post' * V'test_prefix' * action('restart')
          + V'post' * V'test_id' * P'/' * action(
              P 'import'
              + 'modify'
              + 'replace'
              + 'start'
              + 'initialize'
              + 'restart'
              + 'stop'
              + 'uninitialize'
              + 'init_start'
              + 'enable'
              + 'suspend'
              + 'stream'
              + 'truncate_data'
              -- + 'add' + 'remove' -- KLUDGE think modify/replace sufficient
            )
          + V'post' * V'test_prefix' * action('import')
          + V'delete' * V'test_id'
        ) * -P(1),
    }
end

local api = {
    session = {},
    sie = {},
    system = {},
    test = {},
}

local function multipart_file(w, r)
    if r.post and w.multipart then
        for i = 1, #w.multipart do
            local mp = w.multipart[i]
            if mp.tmp_filename then
                return mp
            end
        end
    end
end

function api.session.session(loop, w, r)
    if w.session then
        send_response(w, 200, "OK", "application/json", cjson.encode(w.session))
    else
        local err = {__"No session with key {0}", w.session_key}
        send_response(w, 500, "ERROR", "application/json", cjson.encode(err))
    end
end

function api.session.state(loop, w, r)
    if w.session and w.session.state then
        send_response(w, 200, "OK", "application/json",
            cjson.encode(w.session.state))
    else
        local err = {__"No session with key {0}", w.session_key}
        send_response(w, 500, "ERROR", "application/json", cjson.encode(err))
    end
end

function api.sie.export(loop, w, r)
    local f = w.server:fork(loop, w, 'SIE Export', {}, function()
            local disp = r.id
            if disp and not disp:find('%.sie$') then
                disp = disp .. '.sie'
            end
            local sie_name = storage_path .. '/' .. disp
            local pod = sie_file_mgr.get_sie_path_on_disk(disp:gsub('%.sie$', ''))
            if pod then
                sie_name = pod
            end
            local sp = sie_parser:new(sie_name)
            local setup = sp:get_setup()
            table.insert(w.out_headers,
                'Content-Disposition: attachment; filename="' ..
                r.filename .. '"\n'
            )
            send_response(w, 200, "OK", "application/json", setup)
            downloadCompleteNotification(w.session_key)
        end)
    if f.ok then
        w.server.close(w.server.ev_loop, w)
    else
        send_response(w, 200, "OK", "application/json", cjson.encode({
                result = 'error',
                msg = 'Server error:  fork failed',
            }))
            
    end
end

-- Function used to send files
local function upload_file(w, fh)
    local v
    local nl = ev.Loop.new()
    ev.Timer.new(function(l, t, e)
            if not v then
                v = fh:read(50000)
                if not v then
                    t:stop(l)
                    l:unloop()
                    return
                end
            end
            local amt, estr, err = u.write(w.socket, v, v:len())
            if not amt then
                if err ~= u.EAGAIN then
                    t:stop(l)
                    l:unloop()
                    log:log(log.ERR, "Error sending file")
                    return
                end
            elseif amt == v:len() then
                v = nil
            else
                v = string.sub(v, amt + 1)
            end
        end, .01, .01):start(nl)
    nl:loop()
    fh:close()
end

-- Function that actually does the work for SIE extractors
local function _sie_extract_common(loop, w, r, cname_func, archive)

    -- Get SIE file to extract data from
    local disp = r.id
    if disp and not disp:find('%.sie$') then
        disp = disp .. '.sie'
    end
    local sie_name = storage_path .. '/' .. disp
    local pod = sie_file_mgr.get_sie_path_on_disk(disp:gsub('%.sie$', ''))
    if pod then
        sie_name = pod
    end

    local method = r.method:lower()

    local cnames
    if method == 'post' then
        -- Get channel list from JSON content
        cnames = cjson.decode(w.content)
        if not cnames then
            send_response(w, 200, "OK", "application/json", cjson.encode({
                    result = 'error',
                    msg = 'Request content required',
                }))
            return
        end
        cnames = cnames.channels or {}
    else
        cnames = {w.query.channel}
    end

    -- If no channels were given, then use the provided channel search function
    -- to find them.
    if #cnames < 1 then
        local sp = sie_parser:new(sie_name)
        local setup = sp:get_source_setup()
        local cjs = cjson.decode(setup)
        if not cjs then
            send_response(w, 200, "OK", "application/json",
                cjson.encode({
                    result = 'error',
                    msg = 'Unable to extract SXR from SIE file',
                }))
            return
        end
        cnames = cname_func(cjs)

        -- If there are no message data channels, return
        if not cnames or type(cnames) ~= 'table' or #cnames < 1 then
            send_response(w, 200, "OK", "application/json",
                cjson.encode({
                    result = 'success',
                    msg = 'No channel data available',
                }))
            return
        end
    end

    local extract_path = storage_path .. '/extractor/' .. uuid.generate()
    os.execute(string.format('rm -rf %s;mkdir -p %s',
        extract_path, extract_path))

    local c = {}
    if method == 'post' then
        c = cjson.decode(w.content)
    elseif w.query.start_time and w.query.duration then
        c = {
            range = {
                start = w.query.start_time,
                duration = w.query.duration,
            },
        }
    end
    if c.range then
        os.execute(string.format('cd %s;sie-message-extract -r%s:%s %s "%s"',
            extract_path, c.range.start, c.range.duration, sie_name,
            table.concat(cnames, '" "')))
    else
        os.execute(string.format('cd %s;sie-message-extract %s "%s"',
            extract_path, sie_name, table.concat(cnames, '" "')))
    end

    local filename = w.query.filename or
        (archive and 'extracted-data.tgz' or 'extracted-data.txt')

    local header_lines = {}
    table.insert(header_lines, "HTTP/1.1 200 OK")
    table.insert(header_lines, "Cache-Control: no-cache")
    table.insert(header_lines, "Pragma: no-cache")
    table.insert(header_lines, string.format(
        'Content-Disposition: attachment; filename="%s"', filename))

    if not archive then

        -- Calculate total size of download so browser can show progress
        local total_size = 0
        local file_count = 0
        for name in lfs.dir(extract_path) do
            if name:find('%.txt$') then
                file_count = file_count + 1
                local fh = io.open(extract_path .. '/' .. name, 'r')
                if fh then
                    local file_size = fh:seek("end")
                    total_size = total_size + file_size
                    fh:close()
                end
            end
        end

        -- Make sure at least 1 file was extracted
        if file_count == 0 then
            send_response(w, 200, "OK", "application/json",
                cjson.encode({
                    result = 'error',
                    msg = 'No channels could be extracted',
                }))
            return
        end

        -- Set socket to blocking to guarantee we write all data even if the
        -- client can't keep up. This code runs in a fork so blocking is OK.
        u.setfl(w.socket, 0) -- clear O_NONBLOCK

        table.insert(header_lines, "Content-Type: application/octet-stream")
        table.insert(header_lines, "Content-Length: " .. total_size)
        u.write(w.socket, table.concat(header_lines, '\r\n') .. '\r\n\r\n')

        for name in lfs.dir(extract_path) do
            if name:find('%.txt$') then
                local fh =
                    io.open(extract_path .. '/' .. name, 'r')
                if fh then
                    local file_size = fh:seek("end")
                    fh:seek("set", 0);
                    local remaining = file_size
                    while remaining > 0 do
                        local amt = 16 * 1024
                        if amt > remaining then amt = remaining end
                        local data = fh:read(amt)
                        u.write(w.socket, data, data:len())
                        remaining = remaining - data:len()
                    end
                    fh:close()
                end
            end
        end
    else
        table.insert(header_lines, "Content-Type: application/x-gzip")
        local fh = io.popen('cd ' .. extract_path .. ';tar -czf - ' .. archive)
        if fh then
            u.write(w.socket, table.concat(header_lines, '\r\n') .. '\r\n\r\n')

            upload_file(w, fh)
        else
            send_response(w, 200, "OK", "application/json", cjson.encode({
                    result = 'error',
                    msg = 'Unable to open archive file',
                }))
        end
    end
    os.execute(string.format('rm -rf %s 2>/dev/null', extract_path))
    downloadCompleteNotification(w.session_key)
end

-- Function used to extract raw CAN message data from an SIE file
local function _sie_extract_raw_can(loop, w, r)
    local f = w.server:fork(loop, w, 'SIE Extract Raw CAN', {}, function()
            -- Use the common extractor code
            return _sie_extract_common(loop, w, r,
                function(cjs)
                    local cn = {}
                    for _, stanza in ipairs(cjs) do
                        if stanza.connector and stanza.data_type and
                            stanza.name and
                            stanza.data_type == 'message_can' then
                            cn[#cn + 1] = stanza.name
                        end
                    end
                    return cn
                end)
        end)
    if f.ok then
        w.server.close(w.server.ev_loop, w)
    else
        send_response(w, 200, "OK", "application/json", cjson.encode({
                result = 'error',
                msg = 'Server error:  fork failed',
            }))
    end
end

-- Function used to extract text message data from and SIE file
local function _sie_extract_text(loop, w, r)
    local f = w.server:fork(loop, w, 'SIE Extract TEXT', {}, function()
            -- Use the common extractor code
            return _sie_extract_common(loop, w, r,
                function(cjs)
                    local cn = {}
                        for _, stanza in ipairs(cjs) do
                            if stanza.connector and stanza.data_type and
                                stanza.name and
                                stanza.message_format and
                                stanza.message_format == 'text/plain' and
                                stanza.data_type == 'message_raw' then
                                cn[#cn + 1] = stanza.name
                            end
                        end
                    return cn
                end)
        end)
    if f.ok then
        w.server.close(w.server.ev_loop, w)
    else
        send_response(w, 200, "OK", "application/json", cjson.encode({
                result = 'error',
                msg = 'Server error:  fork failed',
            }))
    end
end

-- Function used to extract NMEA message data from and SIE file
local function _sie_extract_nmea(loop, w, r)
    local f = w.server:fork(loop, w, 'SIE Extract NMEA', {}, function()
            -- Use the common extractor code
            return _sie_extract_common(loop, w, r,
                function(cjs)
                    local cn = {}
                        for _, stanza in ipairs(cjs) do
                            if stanza.connector and stanza.data_type and
                                stanza.name and
                                stanza.message_format and
                                stanza.message_format == 'text/nmea' and
                                stanza.data_type == 'message_raw' then
                                cn[#cn + 1] = stanza.name
                            end
                        end
                    return cn
                end)
        end)
    if f.ok then
        w.server.close(w.server.ev_loop, w)
    else
        send_response(w, 200, "OK", "application/json", cjson.encode({
                result = 'error',
                msg = 'Server error:  fork failed',
            }))
    end
end

-- Function used to extract video frame data from video channels in an SIE file
local function _sie_extract_video(loop, w, r)
    local f = w.server:fork(loop, w, 'SIE Extract Video', {}, function()
            -- Use the common extractor code
            return _sie_extract_common(loop, w, r,
                function(cjs)
                    local cn = {}
                        for _, stanza in ipairs(cjs) do
                            if stanza.connector and stanza.data_type and
                                stanza.name and
                                stanza.message_format and
                                stanza.message_format == 'image/jpeg' and
                                stanza.data_type == 'message_raw' then
                                cn[#cn + 1] = stanza.name
                            end
                        end
                    return cn
                end, "*/*.jpg")
        end)
    if f.ok then
        w.server.close(w.server.ev_loop, w)
    else
        send_response(w, 200, "OK", "application/json", cjson.encode({
                result = 'error',
                msg = 'Server error:  fork failed',
            }))
    end
end

function api.sie.hardware_info(loop, w, r)
    local f = w.server:fork(loop, w, 'SIE Extract Hardware Info', {}, function()
            local disp = r.id
            if disp and not disp:find('%.sie$') then
                disp = disp .. '.sie'
            end
            local sie_name = storage_path .. '/' .. disp
            local pod = sie_file_mgr.get_sie_path_on_disk(disp:gsub('%.sie$', ''))
            if pod then
                sie_name = pod
            end
            local sp = sie_parser:new(sie_name)
            sp:open()
            local ok, v = pcall(sp._get_file_tags, sp, "somat:system_hardware")
            sp:close()
            if not ok then
                send_response(w, 200, "OK", "application/json", cjson.encode({
                        result = 'error',
                        msg = string.format(
                            'Unable to retrieve system hardware ' ..
                            'info from SIE file:  %s', v),
                    }))
                return
            end
            local filename = w.query.filename or 'extracted-data.sxri'
            table.insert(w.out_headers,
                'Content-Disposition: attachment; filename="' ..
                filename .. '"\n'
            )
            send_response(w, 200, "OK", "application/json", v)
        end)
    if f.ok then
        w.server.close(w.server.ev_loop, w)
    else
        send_response(w, 200, "OK", "application/json", cjson.encode({
                result = 'error',
                msg = 'Server error:  fork failed',
            }))
    end
end

-- Main function used to process /api/sie/XXXXX/extract requests
function api.sie.extract(loop, w, r)
    local method = r.method:lower()
    local what
    if method == 'post' then
        local content = cjson.decode(w.content)
        if not content then
            send_response(w, 200, "OK", "application/json", cjson.encode({
                    result = 'error',
                    msg = "JSON content required",
                }))
        end
        what = content.what   -- What we want extracted
    else
        what = w.query.what
    end
    if not what then
        send_response(w, 200, "OK", "application/json", cjson.encode({
                result = 'error',
                msg = "What is to be extracted must be specified",
            }))
    end
    if api.sie[what] then
        return api.sie[what](loop, w, r)
    end
    if what:upper() == 'CAN' then
        return _sie_extract_raw_can(loop, w, r)
    elseif what:upper() == 'NMEA' then
        return _sie_extract_nmea(loop, w, r)
    elseif what:upper() == 'VIDEO' then
        return _sie_extract_video(loop, w, r)
    elseif what:upper() == 'TEXT' then
        return _sie_extract_text(loop, w, r)
    end
    send_response(w, 200, "OK", "application/json", cjson.encode({
            result = 'error',
            msg = string.format("Don't know how to extract %s", what),
        }))
end

function api.sie.get_source_setup(loop, w, r)
    local f = w.server:fork(loop, w, 'SIE Get Source Setup', {}, function()
            local disp = r.id
            if disp and not disp:find('%.sie$') then
                disp = disp .. '.sie'
            end
            local sie_name = storage_path .. '/' .. disp
            local pod = sie_file_mgr.get_sie_path_on_disk(disp:gsub('%.sie$', ''))
            if pod then
                sie_name = pod
            end
            local sp = sie_parser:new(sie_name)
            local setup = sp:get_source_setup()
            table.insert(w.out_headers,
                'Content-Disposition: attachment; filename="' ..
                r.filename .. '"\n'
            )
            send_response(w, 200, "OK", "application/json", setup)
        end)
    if f.ok then
        w.server.close(w.server.ev_loop, w)
    else
        send_response(w, 200, "OK", "application/json", cjson.encode({
                result = 'error',
                msg = 'Server error:  fork failed',
            }))
            
    end
end

function api.sie.get_source_summary(loop, w, r)
    local f = w.server:fork(loop, w, 'SIE Get Source Summary', {}, function()
            local disp = r.id
            if disp and not disp:find('%.sie$') then
                disp = disp .. '.sie'
            end
            local sie_name = storage_path .. '/' .. disp
            local pod = sie_file_mgr.get_sie_path_on_disk(disp:gsub('%.sie$', ''))
            if pod then
                sie_name = pod
            end
            local sp = sie_parser:new(sie_name)
            local summary = sp:get_source_summary()
            table.insert(w.out_headers,
                'Content-Disposition: attachment; filename="' ..
                r.filename .. '"\n'
            )
            send_response(w, 200, "OK", "application/json", summary)
        end)
    if f.ok then
        w.server.close(w.server.ev_loop, w)
    else
        send_response(w, 200, "OK", "application/json", cjson.encode({
                result = 'error',
                msg = 'Server error:  fork failed',
            }))
            
    end
end

function api.sie.get_tags(loop, w, r)
    local f = w.server:fork(loop, w, 'SIE Get Tags', {}, function()
            local disp = r.id
            if disp and not disp:find('%.sie$') then
                disp = disp .. '.sie'
            end
            local sie_name = storage_path .. '/' .. disp
            local pod = sie_file_mgr.get_sie_path_on_disk(disp:gsub('%.sie$', ''))
            if pod then
                sie_name = pod
            end
            local sp = sie_parser:new(sie_name)
            send_response(w, 200, "OK", "application/json", sp:get_tags())
        end)
    if f.ok then
        w.server.close(w.server.ev_loop, w)
    else
        send_response(w, 200, "OK", "application/json", cjson.encode({
                result = 'error',
                msg = 'Server error:  fork failed',
            }))
            
    end
end

function api.sie.get_file_tags(loop, w, r)
    local f = w.server:fork(loop, w, 'SIE Get File Tags', {}, function()
            local disp = r.id
            if disp and not disp:find('%.sie$') then
                disp = disp .. '.sie'
            end
            local sie_name = storage_path .. '/' .. disp
            local pod = sie_file_mgr.get_sie_path_on_disk(disp:gsub('%.sie$', ''))
            if pod then
                sie_name = pod
            end
            local sp = sie_parser:new(sie_name)
            send_response(w, 200, "OK", "application/json", sp:get_file_tags())
        end)
    if f.ok then
        w.server.close(w.server.ev_loop, w)
    else
        send_response(w, 200, "OK", "application/json", cjson.encode({
                result = 'error',
                msg = 'Server error:  fork failed',
            }))
            
    end
end

function api.sie.get_test_tags(loop, w, r)
    local f = w.server:fork(loop, w, 'SIE Get Test Tags', {}, function()
            local disp = r.id
            if disp and not disp:find('%.sie$') then
                disp = disp .. '.sie'
            end
            local sie_name = storage_path .. '/' .. disp
            local pod = sie_file_mgr.get_sie_path_on_disk(disp:gsub('%.sie$', ''))
            if pod then
                sie_name = pod
            end
            local sp = sie_parser:new(sie_name)
            send_response(w, 200, "OK", "application/json", sp:get_test_tags())
        end)
    if f.ok then
        w.server.close(w.server.ev_loop, w)
    else
        send_response(w, 200, "OK", "application/json", cjson.encode({
                result = 'error',
                msg = 'Server error:  fork failed',
            }))
            
    end
end

function api.sie.get_channel_tags(loop, w, r)
    local f = w.server:fork(loop, w, 'SIE Get Channel Tags', {}, function()
            local disp = r.id
            if disp and not disp:find('%.sie$') then
                disp = disp .. '.sie'
            end
            local sie_name = storage_path .. '/' .. disp
            local pod = sie_file_mgr.get_sie_path_on_disk(disp:gsub('%.sie$', ''))
            if pod then
                sie_name = pod
            end
            local sp = sie_parser:new(sie_name)
            send_response(w, 200, "OK", "application/json", sp:get_channel_tags())
        end)
    if f.ok then
        w.server.close(w.server.ev_loop, w)
    else
        send_response(w, 200, "OK", "application/json", cjson.encode({
                result = 'error',
                msg = 'Server error:  fork failed',
            }))
            
    end
end

function api.sie.get_xml_metadata(loop, w, r)
    local f = w.server:fork(loop, w, 'SIE Get XML Metadata', {}, function()
            local disp = r.id
            if disp and not disp:find('%.sie$') then
                disp = disp .. '.sie'
            end
            local sie_name = storage_path .. '/' .. disp
            local pod = sie_file_mgr.get_sie_path_on_disk(disp:gsub('%.sie$', ''))
            if pod then
                sie_name = pod
            end
            local sp = sie_parser:new(sie_name)
            send_response(w, 200, "OK", "application/xml",
                sp:get_tag_by_name("sie:xml_metadata"))
        end)
    if f.ok then
        w.server.close(w.server.ev_loop, w)
    else
        send_response(w, 200, "OK", "application/json", cjson.encode({
                result = 'error',
                msg = 'Server error:  fork failed',
            }))
            
    end
end

function api.sie.get_system_version(loop, w, r)
    local f = w.server:fork(loop, w, 'SIE Get System Version', {}, function()
            local disp = r.id
            if disp and not disp:find('%.sie$') then
                disp = disp .. '.sie'
            end
            local sie_name = storage_path .. '/' .. disp
            local pod = sie_file_mgr.get_sie_path_on_disk(disp:gsub('%.sie$', ''))
            if pod then
                sie_name = pod
            end
            local sp = sie_parser:new(sie_name)
            send_response(w, 200, "OK", "application/json", sp:get_system_version())
        end)
    if f.ok then
        w.server.close(w.server.ev_loop, w)
    else
        send_response(w, 200, "OK", "application/json", cjson.encode({
                result = 'error',
                msg = 'Server error:  fork failed',
            }))
            
    end
end

function api.system.sdbx(loop, w, r)
    local method = r.method:lower()
    if method ~= 'delete' then
        send_response(w, 200, "OK", "application/json",
            cjson.encode({
                msg = string.format("%s method not supported", method:upper()),
                result = 'error',
            }))
        return
    end
    local q = w.query
    local file = q.file_name
    if not file then
        send_response(w, 200, "OK", "application/json",
            cjson.encode({
                msg = 'Parameter "file_name" required',
                result = 'error',
            }))
        return
    end
    os.execute(string.format("rm -f '/storage/sdbx/%s' 2>/dev/null", file))
    send_response(w, 200, "OK", "application/json",
        cjson.encode({
            msg = string.format("File '%s' removed", file),
            result = 'success',
        }))
end

function api.system.log(loop, w, r)
    local q = w.query
    local count = q.count or 10
    local offset = q.offset or 0
    local lower = q.start_time or '0000-01-01'
    local upper = q.end_time or '9999-12-31'
    local order = q.order or 'newest'
    local level = q.level or '%%'
    local method = r.method:lower()
    local clear_led = true

    if q.clear_led then
        clear_led = q.clear_led == 'true'
    end

    -- Delete can not be done in a fork
    if method == 'delete' then
        local ret = w.server.logger:Purge(lower, upper)
        if ret == sqlite3.OK then
            send_response(w, 200, "OK", "application/json",
                cjson.encode({{purged=true}}))
        else
            send_response(w, 200, "OK", "application/json",
                cjson.encode({{purged=false, error = "Unable to purge log"}}))
        end
        return
    end

    local f = w.server:fork(loop, w, 'System Log', 0, function()
        -- KLUDGE maybe try using HTTP Range or Content-Range with unit "dates"

        assert(order == 'newest' or order == 'oldest',
            "'order' must be 'newest' or 'oldest'")

        -- KLUDGE use database.open() for versioning
        local db = sqlite3.open('/var/lib/armada/log.db')

        if order == 'oldest' and not q.start_time then
            local rows
            for r in db:rows[[
                SELECT count(*) FROM log;
            ]] do
                rows = tonumber(r[1])
            end
            offset = rows - count - offset
        end

        send_response_headers(w, 200, "OK", "application/json", '[')

        local req = string.format([[
            SELECT time, component, level, message
            FROM log WHERE time >= ? and time <= ?
            AND level LIKE ?
            AND ((component NOT LIKE '%%kern'
            AND component NOT LIKE '%%auth%%'
            AND (level LIKE 'warn%%' OR level LIKE 'err%%'
                OR level LIKE 'crit%%'))
            OR message LIKE 'test-engine%%'
            OR message LIKE 'run-test%%'
            OR message LIKE 'restart-%%'
            OR component == 'powerd')
            ORDER BY rowid %s LIMIT ? OFFSET ?;
        ]], (order == 'newest' and 'DESC' or 'ASC'))
        if system_config.get_all_log_entries then
            req = string.format([[
                SELECT time, component, level, message
                FROM log
                WHERE time >= ? and time <= ?
                AND level LIKE ?
                ORDER BY rowid %s LIMIT ? OFFSET ?;
            ]], (order == 'newest' and 'DESC' or 'ASC'))
        end
        local sth = db:prepare(req)
        local i = 0
        sth:bind_values(lower, upper, level, count, offset)
        for r in sth:rows() do
            local t = {
                date = r[1],
                facility = r[2],
                priority = r[3],
                message = r[4],
            }
            u.write(w.socket, (i == 0 and '' or ',\n') .. cjson.encode(t))
            i = i + 1
        end

        db:close()
        u.write(w.socket, ']\n')
    end)
    if f.ok then
        w.server.close(w.server.ev_loop, w)
        if clear_led then
            sys_status.error.error = false
            sl:set_status({error={error=false,reset=true}})
            w.server.network:AddRR('_reset-error', 'TXT', 'Reset Error')
            ev.Timer.new(function(l, t, e)
                    w.server.network:RemoveRR(
                        '_reset-error', 'TXT', 'Reset Error')
                end, .25, 0):start(ev.Loop.default)
        end
    else
        send_response(w, 500, "ERROR", "application/json", cjson.encode(f))
    end
end


function api.system.get_errors(loop, w, r)

    local count = w.query.count or 5

    local f = w.server:fork(loop, w, 'System Errors', 10, function()

        local db = sqlite3.open('/var/lib/armada/log.db')

        send_response_headers(w, 200, "OK", "application/json", '[')

        local sql = [[
            SELECT time, component, level, message
            FROM log WHERE ((component NOT LIKE '%%kern'
            AND component NOT LIKE '%%auth%%'
            AND (level LIKE 'warn%%' OR level LIKE 'err%%'
                OR level LIKE 'crit%%'))
            OR message LIKE 'test-engine%%'
            OR message LIKE 'run-test%%'
            OR message LIKE 'restart-%%'
            OR component == 'powerd')
            ORDER BY rowid DESC
        ]]

        local sth = db:prepare(sql)

        local error_count = 0

        for r in sth:rows() do

            local t = {
                date = r[1],
                facility = r[2],
                priority = r[3],
                message = r[4],
            }

            if t.priority == 'err' then
                u.write(w.socket, (error_count == 0 and '' or ',\n') .. cjson.encode(t))
                error_count = error_count + 1
                if count > 0 and error_count >= count then break end
            else
                if error_count > 0 then break end
            end

        end

        db:close()
        u.write(w.socket, ']\n')

    end)

    if f.ok then
        w.server.close(w.server.ev_loop, w)
    else
        send_response(w, 500, "ERROR", "application/json", cjson.encode(f))
    end

end

local function print_table(w, t, indent)
    local indent = indent or 0
    for k, v in pairs(t) do
        if type(v) == 'table' then
            print_table(w, v, indent + 4)
        else
            u.write(w.socket, string.format("%s%s: %s\n",
                string.rep('  ', indent), k, tostring(v)))
        end
    end
end

function api.system.export_log(loop, w, r)
    local f = w.server:fork(loop, w, 'System Export Log', 0, function()
        local q = w.query
        local count = q.count or -1  -- unlimited unless specified
        local lower = q.start_time or '0000-01-01'
        local upper = q.end_time or '9999-12-31'
        local order = q.order or 'newest'
        local filename = q.filename or 'exported_log'
        filename = tostring(filename)

        assert(order == 'newest' or order == 'oldest',
            "'order' must be 'newest' or 'oldest'")

        -- KLUDGE use database.open() for versioning
        local db = sqlite3.open('/var/lib/armada/log.db')

        local header_lines = {}
        table.insert(header_lines, "HTTP/1.1 200 OK\n")
        table.insert(header_lines, "Cache-Control: no-cache\n")
        table.insert(header_lines, "Pragma: no-cache\n")
        table.insert(header_lines, "Content-Type: text/plain\n")
        table.insert(header_lines, 'Content-Disposition: attachment; ')
        table.insert(header_lines, 'filename="' .. filename .. '"\n\n')
        u.write(w.socket, table.concat(header_lines))

        u.write(w.socket, '--------- System Information\n')
        u.write(w.socket, '----------------------------- ' ..
            system_config.device_type:upper() .. '\n')

        -- Grab the firmware version and prepend it to the log
        local fh = io.open('/etc/version.json', 'r')
        if fh then
            local vinfo = fh:read('*a')
            fh:close()
            local j = cjson.decode(vinfo)
            if vinfo and j then
                u.write(w.socket, string.format("Version: %s\n", j.version))
                u.write(w.socket, string.format("Build: %s\n", j.build))
                if system_config.device_type == 'edaqxr' then
                    u.write(w.socket, "Product: edaq-web\n")
                else
                    u.write(w.socket, string.format("Product: %s\n",
                        j.build_product))
                    u.write(w.socket, string.format("Description: %s\n",
                        j.build_description))
                    u.write(w.socket, string.format("Hash: %s\n", j.source_hash))
                end
                u.write(w.socket, string.format("Device: %s\n", hostname))
            elseif vinfo then
                u.write(w.socket, vinfo .. '\n')
            end
        end

        -- Grab all of the platform info and prepend it to the log
        local rrs =
            w.server.network:QueryRR('*', 'TXT', 'platform_info:*', '*')
        if rrs then
            u.write(w.socket,
                '----------------------------- Connected Hardware\n')
        end
        for _, rr in ipairs(rrs) do
            local cj = cjson.decode(rr[3]:match('^platform_info: (.+)$'))
            if cj and cj.device and cj.device ~= 'CX23R' and
                cj.device ~= 'EDAQXR' then
                print_table(w, cj)
                u.write(w.socket, '\n')
            end
        end

        -- Grab Disk information
        local fh = io.popen('df')
        if fh then
            u.write(w.socket,
                '----------------------------- Disk information\n')
            for line in fh:lines() do
                u.write(w.socket, line .. '\n')
            end
            fh:close()
        end

        -- Grab Number of SIE files
        fh = io.popen('ls -l /storage/*.sie | wc -l')
        if fh then
            u.write(w.socket,
                '----------------------------- Number of SIE files\n')
            for line in fh:lines() do
                u.write(w.socket, line .. '\n')
            end
            fh:close()
        end

        -- Grab Number of setups
        local sxrs = sql_exec(w.server.setup_db, 'tsd_setups',
            [[SELECT db_name FROM "%s" WHERE deleted = "" AND id = "summary"
                AND NOT content LIKE "%%database%%";]])
        if sxrs then
            u.write(w.socket,
                '----------------------------- Number of SXR files\n')
            u.write(w.socket, string.format("%d\n", #sxrs))
        end

        -- Grab Number of clients
        u.write(w.socket,
            '----------------------------- Number of Connected Clients\n')
        u.write(w.socket, string.format("%d\n", #w.server.comet_clients))
        
        u.write(w.socket, '----------------------------- Log content\n')
        local log_header = {}
        table.insert(log_header, "Time                              ")
        table.insert(log_header, "Component           ")
        table.insert(log_header, "Level       ")
        table.insert(log_header, "Message\n")
        u.write(w.socket, table.concat(log_header))

        local sth = db:prepare(string.format([[
            SELECT time, component, level, message
            FROM log
            WHERE time >= ? and time <= ?
            ORDER BY rowid %s LIMIT ?;
        ]], (order == 'newest' and 'DESC' or 'ASC')))
        local i = 0
        sth:bind_values(lower, upper, count)
        for r in sth:rows() do
            local line = {}
            local date_str = string.format("%s     ", r[1])
            table.insert(line, date_str)
            local comp_str = string.sub(r[2], 1, 20)
            table.insert(line, comp_str)
            if #comp_str < 20 then
                local comp_padding = string.rep(" ", 20 - #comp_str)
                table.insert(line, comp_padding)
            end
            local level_str = r[3]
            table.insert(line, level_str)
            if #level_str < 12 then
                local level_padding = string.rep(" ", 12 - #level_str)
                table.insert(line, level_padding)
            end
            table.insert(line, r[4])  -- log message string

            u.write(w.socket, (i == 0 and '' or ',\n') .. table.concat(line))
            i = i + 1
        end

        db:close()
        u.write(w.socket, '\n')
        downloadCompleteNotification(w.session_key)
    end)

    if f.ok then
        w.server.close(w.server.ev_loop, w)
        sys_status.error.error = false
        sl:set_status(sys_status)
    else
        send_response(w, 500, "ERROR", "application/json", cjson.encode(f))
    end
end

local get_fpga_version = require 'fpga_utils'

local sdm = require 'setupdb_migrate'
local expected_setup_version = sdm.get_info().expected_setup_file_version

local fpga_version

-- Internal use only
function api.system._init_sie(loop, w, r)
    sie_file_mgr.check_sie_files()
    send_response(w, 200, "OK", "application/json", cjson.encode({result='ok'}))
end

function api.system.version(loop, w, r)
    -- append the setup version to the response
    local version = {}
    local f = io.open('/etc/version.json')
    if f then
        version = cjson.decode(f:read('*a')) or {}
        f:close()
    end
    f = io.open('/tmp/power-micro-version.json', 'r')
    version.power_micro_version = 'Unknown'
    if f then
        local t = cjson.decode(f:read('*a'))
        if t then
            version.power_micro_version = {
                running =
                    string.format("%d.%d", t.running.major, t.running.minor),
                expected =
                    string.format("%d.%d", t.expected.major, t.expected.minor),
                is_expected =
                    t.running.major == t.expected.major and
                    t.running.minor == t.expected.minor,
            }
        end
        f:close()
    end
    if not fpga_version then
        local running, expected = get_fpga_version()
        if running ~= 'Unknown' then
            fpga_version = {
                running = running,
                is_expected = expected,
            }
        end
    end
    version.fpga_version = fpga_version or {running = 'Unknown'}
    version.setup_version = expected_setup_version or 'Unknown'
    send_response(w, 200, "OK", "application/json", cjson.encode(version))
end

function api.system.setupdb_status(loop, w, r)
    local sdm = require 'setupdb_migrate'
    send_response(w, 200, "OK", "application/json",
        cjson.encode(sdm.get_info()))
end

-- API used to remove databases
function api.system.delete_databases(loop, w, r)
    send_response(w, 200, "OK", "application/json", '')
    ev.Timer.new(function(l, t, e)
            os.execute('rm -f /var/lib/armada/*.db*;sync;sync;/usr/bin/reboot')
        end, .25, 0):start(loop)
end

-- API function used to download GZIP'd tarball of databases
function api.system.customer_service_package(loop, w, r)
    local f = w.server:fork(loop, w, 'System Customer Service Package', {}, function()
            local disp = w.query.filename
            if not disp then disp = 'customer_service_package.tar' end
            if disp and not disp:find('%.tar$') then
                disp = disp .. '.tar'
            end

            local cmds = {
                {
                    heading = 'DMESG CONTENT',
                    cmd = '/bin/dmesg',
                },
                {
                    heading = 'PROCESS STATUS',
                    cmd = '/bin/ps -eaf',
                },
                {
                    heading = 'DISK SPACE',
                    cmd = '/bin/df -kh'
                },
            }

            -- File to dump stuff into
            local my_uuid = uuid.generate()
            local cs_file = 'cs-' .. my_uuid .. '.txt'
            local tar_file = 'csp-' .. my_uuid .. '.tar'

            local fh = io.open('/var/lib/armada/' .. cs_file, 'w')

            -- Run all the commands and append the content to the file
            for _, ci in ipairs(cmds) do
                fh:write(string.rep('-', 80) .. '\n')
                fh:write(string.format('%s (%s)\n', ci.heading, ci.cmd))
                fh:write(string.rep('-', 80) .. '\n')
                local fh1 = io.popen(ci.cmd)
                if fh1 then
                    local v = fh1:read('*a')
                    fh:write(v)
                    fh1:close()
                end
            end
            fh:close()

            -- Create and send the TAR archive
            local csp = storage_path .. '/csp.tar'
            local handle = io.popen("ls /var/lib/armada")
            local result = handle:read("*a")
            handle:close()
            if string.find(result, "recovery") then
                os.execute('cd /var/lib/armada;tar -cf ' .. csp ..
                    ' recovery log.db setup.db ' .. cs_file)
            else
                os.execute('cd /var/lib/armada;tar -cf ' .. csp ..
                    ' log.db setup.db ' .. cs_file)
            end
            table.insert(w.out_headers,
                'Content-Disposition: attachment; filename="' ..disp.. '"\n' ..
                'X-Accel-Redirect: ' .. csp .. '\n')
            send_response(w, 200, "OK", "application/tar", '')
            downloadCompleteNotification(w.session_key)
            C.unlink('/var/lib/armada/' .. cs_file)
            C.unlink(csp)
        end)
    if f.ok then
        downloadCompleteNotification(w.session_key)
        w.server.close(w.server.ev_loop, w)
    else
        f.error = 'Failed'
        send_response(w, 200, "OK", "application/json", cjson.encode(f))
    end
end

function api.system.platform_info(loop, w, r)
    local f = w.server:fork(loop, w, "Platform Info", {}, function()
            local rrs =
                w.server.network:QueryRR('*', 'TXT', 'platform_info:*', '*')
            local pi = {}
            for _, rr in ipairs(rrs) do
                local cj = cjson.decode(rr[3]:match('^platform_info: (.+)$'))
                pi[#pi + 1] = cj
            end
            send_response(w, 200, "OK", "application/json", cjson.encode(pi))
    end)
    if f.ok then
        w.server.close(w.server.ev_loop, w)
    else
        send_response(w, 200, "OK", "application/json", cjson.encode({
                result = 'error',
                msg = 'Server error:  fork failed',
            }))
    end
end

function api.system.status(loop, w, r)
    local status = sl:get_status('all')
    local f = w.server:fork(loop, w, "System Status", {}, function()
            -- append the setup version to the response
            local version = {}
            local f = io.open('/etc/version.json')
            if f then
                version = cjson.decode(f:read('*a')) or {}
                f:close()
            end
            f = io.open('/tmp/power-micro-version.json', 'r')
            version.power_micro_version = 'Unknown'
            if f then
                local t = cjson.decode(f:read('*a'))
                if t then
                    version.power_micro_version = {
                        running = string.format("%d.%d",
                            t.running.major, t.running.minor),
                        expected = string.format("%d.%d",
                            t.expected.major, t.expected.minor),
                        is_expected =
                            t.running.major == t.expected.major and
                            t.running.minor == t.expected.minor,
                    }
                end
                f:close()
            end
            if not fpga_version then
                local running, expected = get_fpga_version()
                if running ~= 'Unknown' then
                    fpga_version = {
                        running = running,
                        is_expected = expected,
                    }
                end
            end
            version.fpga_version = fpga_version or {running = 'Unknown'}
            version.setup_version = expected_setup_version or 'Unknown'
            status.version = version
            send_response(w, 200, "OK", "application/json",
                cjson.encode(status))
    end)
    if f.ok then
        w.server.close(w.server.ev_loop, w)
    else
        send_response(w, 200, "OK", "application/json", cjson.encode({
                result = 'error',
                msg = 'Server error:  fork failed',
            }))
            
    end
end

function api.system._dev_config(loop, w, r)
    local c = system_config
    -- NOTE:  Does not deal with type 'function' in subtables!!
    for k, v in pairs(c) do
        if type(v) == 'function' then
            c[k] = nil
        end
    end
    send_response(w, 200, "OK", "application/json", cjson.encode(c))
end

function api.system.get_datetime(loop, w, r)
    local dt_str = os.date("!*t")
    send_response(w, 200, "OK", "application/json", cjson.encode(dt_str))
end

function api.system.set_datetime(loop, w, r)
    local dt_input = assert(cjson.decode(w.content))

    for k, v in pairs(dt_input) do
        dt_input[k] = tonumber(v)
        if not dt_input[k] then
            send_response(w, 200, "OK", "application/json",
                cjson.encode{
                    result = "error",
                    msg = string.format("Field '%s' is invalid", k),
                })
            return
        end
    end

    local dt_str = string.format("%4d.%02d.%02d-%02d:%02d:%02d",
        dt_input.year, dt_input.month, dt_input.day,
        dt_input.hour, dt_input.min, dt_input.sec)

    local offset = os.time(dt_input) - os.time(os.date('*t'))

    local ok = shellf("date -u -s %s", dt_str)
    if ok then
        -- If configuratiion enables setting of the time stamping hardware time
        -- when the date/time are set, then set it too!
        w.server.timing:AdjustClock(offset, 0, 'APISERVER')

        send_response(w, 200, "OK", "application/json",
                    cjson.encode(os.date("*t")))
    else
        send_response(w, 200, "OK", "application/json", cjson.encode{
            result = 'error',
            msg = string.format('shell command failed using string %s', dt_str)
        })
    end
end

-- Note, can not fork this function because the HTTP response is actually
-- proxied from the channel proxy back to the requesting entity.  For whatever
-- reason if that is attempted via a forked version of this function, the
-- information never gets back to the requester and a 502 error occurs.
--function api.system.control(loop, w, r)
function api.system.control(loop, w, r)
    local cw = w
    local control = assert(cjson.decode(w.content))

    local num_tests = 0
    for _, num in pairs(sys_status.test.types or {}) do
        if tonumber(num) then --filter out the table in which the remote state is in from this table 
            num = tonumber(num)
            num_tests = num_tests + num
        end
    end

    if num_tests > 0 and control.request and
        (control.request == 'reboot' or
         control.request == 'change_drd' or
         control.request == 'change_speed_mode' or
         control.request == 'update_firmware') and
        control.connector and control.connector:lower():find('mx%d+') then

        send_response(w, 200, 'OK', 'application/json',
            cjson.encode{
                result = 'error',
                msg = 'Request is rejected because the test engine is running and the request would result in rebooting the MX device',
            })
        return

--[[
        -- This code is untested but is intended to allow the reboot to happen
        -- if there are non-SIE tests running by stopping those tests first.
        -- Use of this code requires a less selective conditional above.
        local num_non_sie_tests = 0
        for k, v in pairs(sys_status.test.types or {}) do
            if k == 'sie' and v > 0 then
                -- There is an SIE test running, reboot is not allowed at all
                send_response(w, 200, 'OK', 'application/json',
                    cjson.encode{
                        result = 'notice',
                        msg = 'Device reboots not allowed while SIE ' ..
                            'tests are running',
                    })
                return
            else
                num_non_sie_tests = num_non_sie_tests + 1
            end
        end
        if num_non_sie_tests > 0 then
            -- There are tests running that are NOT SIE tests, so stop them
            -- and allow the reboot to happen.
            for _, test_pid in pairs(w.server.running_tests_kludge or {}) do
                C.kill(ffi.new('pid_t', tonumber(test_pid)), u.SIGTERM)
            end
            os.remove(storage_path .. '/.test-to-restart.info')
            os.remove(storage_path .. '/.cyclic-test.info')
        end
]]
    end

    local function broadcast_command(request, content)
        local owner = '_' .. request .. '.' .. uuid.generate()
        w.server.network:ModifyRRs({}, {{owner, 'TXT', content}})
        w.server.network:ModifyRRs({{owner, 'TXT', content}}, {})
    end

    -- For broadcast requests ONLY, support batched multiple requests.
    if w.query.broadcast and w.query.batch then
        for _, data in ipairs(control) do
            broadcast_command(data.request, cjson.encode(data))
        end
        send_response(w, 200, 'OK', 'application/json',
            cjson.encode{
                result = 'success',
                msg = 'Channel zero request broadcast',
            })
        return
    end

    if not control.request then
        send_response(w, 200, 'OK', 'application/json',
            cjson.encode{
                result = 'error',
                msg = 'Control API request is missing "request" parameter',
            })
        return
    elseif not control.params then
        send_response(w, 200, 'OK', 'application/json',
            cjson.encode{
                result = 'error',
                msg = 'Control API request is missing "params" parameter',
            })
        return
    elseif not control.connector then
        send_response(w, 200, 'OK', 'application/json',
            cjson.encode{
                result = 'error',
                msg = 'Control API request is missing "connector" parameter',
            })
        return
    end

    -- If this is an MX device connector and the mx proxy host isn't specified
    -- default it to this unit.
    if control.connector:find('mx%d+b%-%x+$') then
        control.connector = control.connector .. '.' .. hostname
    end

    -- If broadcasting this control request, doit and return success
    if w.query.broadcast then
        broadcast_command(control.request, w.content)
        send_response(w, 200, 'OK', 'application/json',
            cjson.encode{
                result = 'success',
                msg = 'Channel zero request broadcast',
            })
        return
    end

    local connector = control.connector:match('@(.+)$')
    local host, port, ip =
        w.server.network:ResolveService(connector)
    local emsg = string.format("Can not resolve service for '%s'",
        control.connector or 'UNSPECIFIED CONNECTOR')
    if host then
        local cb_w = w.server:connect_nb(ip, port,
            function(loop, w)
                local got = u.read(w.socket, 8192)
                if not got or got == "" then
                    w.server.close(loop, w)
                    cw.server.close(loop, cw)
                else
                    while #got > 0 do
                        local amt, estr, eno = u.write(cw.socket, got)
                        if not amt then
                            if eno ~= u.EAGAIN then
                                w.server.close(loop, w)
                                cw.server.close(loop, cw)
                                return
                            end
                        elseif amt > 0 then
                            got = got:sub(amt + 1)
                        end
                    end
                end
            end,
            function(loop, cbs)
                if not cbs then
                    send_response(w, 200, 'OK', 'application/json',
                        cjson.encode{
                            result = 'error',
                            msg = 'Connect to remote server failed'
                        })
                    return
                end
                local json = cjson.encode(control) .. "\r\n"
                local req = "POST /control/ HTTP/1.0\r\n" ..
                    "Content-type: application/json\r\n" ..
                    "Content-length: " .. #json .. "\r\n" ..
                    "\r\n" ..  json
                while #req > 0 do
                    local written, estr, eno = u.write(cbs.socket, req)
                    if not written then
                        if eno ~= u.EAGAIN then
                            send_response(w, 200, 'OK', 'application/json',
                                cjson.encode{
                                    result = 'error',
                                    msg = 'Unable to forward request: ' .. estr,
                                })
                            return
                        end
                    elseif written > 0 then
                        req = req:sub(written + 1)
                    end
                end
                return
            end)
        if cb_w then return end
        emsg = string.format("NB connect to %s failed", control.connector)
    end
    local f = {
        result = 'error',
        msg = emsg,
    }
    send_response(w, 404, "NOT FOUND", "application/json", cjson.encode(f))
end

function api.system.unlock_start(loop, w, r)
    start_is_locked = nil
    local f = {
        result = 'success',
        msg = 'Test start is unlocked'
    }
    send_response(w, 200, "OK", "application/json", cjson.encode(f))
end

function api.system.cancel_restart(loop, w, r)
    os.execute('killall -SIGUSR1 restart-tests')
    local f = {
        result = 'success',
        msg = 'Restart of interrupted test cancelled',
    }
    send_response(w, 200, "OK", "application/json", cjson.encode(f))
    start_is_locked = nil
end

function api.system.acknowledge_notify(loop, w, r)
    -- Reset the flag to prevents other notifications from being sent
    sys_status.error.error = false

    -- Make sure escape crap is removed!
    w.content = string.gsub(w.content, '\\"', '"')
    w.content = string.gsub(w.content, '\\', '')

    -- Decode the JSON
    local cj = cjson.decode(w.content)
    if not cj then
        local f = {
            result = 'error',
            msg = 'Unable to decode JSON content',
        }
        send_response(w, 200, "OK", "application/json", cjson.encode(f))
    end
    if not cj.id then
        local f = {
            result = 'error',
            msg = 'JSON content does not contain id field',
        }
        send_response(w, 200, "OK", "application/json", cjson.encode(f))
    end
    local owner = '_system-notification-' .. cj.id

    local action = w.query.action
    if action then
        if action == 'cancel_restart' then
            os.execute('killall -SIGUSR1 restart-tests')
            start_is_locked = nil
        end
    end

    -- Remove all notifications found with the given owner
    w.server.network:RemoveLocalRROwner(owner)

    -- Send success
    local f = {
        result = 'success',
        msg = 'OK',
    }
    send_response(w, 200, "OK", "application/json", cjson.encode(f))
end

-- Start the FPGA recovery process
function api.system.do_fpga_recovery(loop, w, r)
    os.execute('touch /var/lib/armada/.do-fpga-recovery')
    local f = {
        result = 'success',
        msg = 'Starting FPGA Recovery process'
    }
    send_response(w, 200, "OK", "application/json", cjson.encode(f))
    ev.Timer.new(function(l, t, e)
            os.execute('sync;sync;/usr/bin/reboot')
        end, .250, 0):start(ev.Loop.default)
end

local function get_remote_password(network, host)
    local rrs = network:QueryRR(host, 'TXT', 'platform_info: *', '*')
    if not rrs or #rrs < 1 then return 'root' end
    rrs[1][3] = rrs[1][3]:match('^platform_info:%s-(.+)$')
    if not rrs[1][3] then return 'root' end
    local t = cjson.decode(rrs[1][3])
    if not t then return 'root' end
    local extra = t.extra
    if extra and (extra == '' or extra == 'alpha' or extra == 'beta') then
        return '_cx mada7'
    end
    return 'root'
end

function api.system.led_control(loop, w, r)
    local unit = w.query.unit
    local led = w.query.led or 'green'
    local state = w.query.state or 'on'
    local leds = {
        red = true,
        green = true,
        yellow = true,
        blue = true
    }
    local states = {
        off = true,
        one_hz = true,
        two_hz = true,
        four_hz = true,
        eight_hz = true,
        on = true
    }
    if not leds[led] or not states[state] then
        local f = {
            result = 'fail',
            msg = 'Invalid parameter value'
        }
        send_response(w, 200, "OK", "application/json", cjson.encode(f))
        return
    end

    local cmd
    if unit then
        local ip, _ = w.server.network:ResolveHostname(unit)
        if not ip then
            local f = {
                result = 'fail',
                msg = string.format("Unable to get IP address for %s", unit)
            }
            send_response(w, 200, "OK", "application/json", cjson.encode(f))
            return
        end
        local pwd = get_remote_password(w.server.network, unit)
        cmd = string.format(
            "DROPBEAR_PASSWORD='%s' /usr/bin/dbclient -y root@%s " ..
            "'led-control %s %s'", pwd, ip, state, led)
    else
        cmd = string.format('/usr/bin/led-control %s %s', state, led)
    end
    os.execute(cmd)
    local f = {
        result = 'success',
        msg = 'LED state set'
    }
    send_response(w, 200, "OK", "application/json", cjson.encode(f))
end

-- Function used to process "alias" requests
function api.system.alias(loop, w, r)
    -- Make sure decodable JSON payload has been provided
    local content = cjson.decode(w.content)
    if not content then
        send_response(w, 200, "OK", "application/json",
            cjson.encode{
                result = 'error',
                error = 'alias:  unable to decode payload as JSON',
            })
        return
    end

    -- Request to remove an alias
    if content.remove then
        local rrs = w.server.network:QueryRR(content.remove, 'ALIAS', '*', '*')
        if #rrs > 0 then
            -- Let all know to remove the alias
            w.server.network:ModifyRRs({},
                {{rrs[1][1], 'REMOVEALIAS', rrs[1][3]}})
            w.server.network:ModifyRRs({{rrs[1][1], 'REMOVEALIAS', rrs[1][3]}},
                {})
            send_response(w, 200, "OK", "application/json",
                cjson.encode({result = 'success',msg = 'alias: alias deleted'}))
        end
        send_response(w, 200, "OK", "application/json",
            cjson.encode{result = 'error', error = 'alias: no such alias'})
        return
    end

    -- Add an alias
    -- Format of JSON payload is
    -- "alias": "name"
    --
    -- Grab the alias and name from the content
    local alias, name
    for k, v in pairs(content) do
        alias = k
        name = v
        break
    end

    -- Make sure the required content exists
    if not name or not alias then
        send_response(w, 200, "OK", "application/json",
            cjson.encode{
                result = 'error',
                error = 'alias:  One or more required parameters not given ' ..
                    'in JSON payload.',
            })
        return
    end

    -- Make sure the name and alias are NOT the same
    if name == alias then
        send_response(w, 200, "OK", "application/json",
            cjson.encode{
                result = 'error',
                error = 'alias:  An alias may not alias itself.  ' ..
                    'The "alias" and "name" must be different.',
            })
        return
    end

    -- Check to make sure the thing we are aliasing exists as an 'A', 'SRV',
    -- or 'ALIAS' record.
    -- NOTE:  We may want to remove this to allow alias' to be set even though
    -- the item the alias points to doesn't exist but may exist in the future.
    local found
    for _, rtype in ipairs{'A', 'SRV', 'ALIAS'} do
        local n = name
        -- If aliasing an MX device look for MX with network node extension
        if n:find('^mx%d+') then
            n = n .. "*"
        end
        local rrs = w.server.network:QueryRR(n, rtype, '*', '*')
        if #rrs > 0 then
            found = true
            break
        end
    end
    if not found then
        send_response(w, 200, "OK", "application/json",
            cjson.encode{
                result = 'error',
                error = 'alias:  No such name to alias to.',
            })
        return
    end

    -- MX devices are allowed to have multiple entries for the same device
    if not name:find('^mx%d+') then
        -- Make sure this alias does not exist already!
        local rrs = w.server.network:QueryRR(alias, 'ALIAS', '*', '*')
        if #rrs > 0 then
            send_response(w, 200, "OK", "application/json",
                cjson.encode{
                    result = 'error',
                    error = 'alias:  Alias already exists.',
                })
            return
        end
    end

    -- Everything is kosher -- add the alias
    w.server.network:ModifyRRs({}, {{alias, 'SETALIAS', name}})

    -- Make sure it has time to propagate to remote network databases before
    -- removal
    ev.Timer.new(function(l, t, e)
            w.server.network:ModifyRRs({{alias, 'SETALIAS', name}}, {})
        end, 1, 0):start(ev.Loop.default)

    send_response(w, 200, "OK", "application/json",
        cjson.encode({result = 'success', msg = 'alias: Alias added',}))
end

function api.system.update_firmware(loop, w, r)
    local method = r.method:lower()
    local unit = w.query.unit
    if method == 'get' then
        if unit and unit:find('dx23r%-......') then
            if not shellf(
                '/usr/bin/dx-updater ' ..
                '/usr/share/dx23r-firmware/__ng-update.tar.gz ' ..
                unit .. ' &') then
                send_response(w, 200, "OK", "application/json",
                    cjson.encode{
                        ok = false,
                        error = "dx updater failed to run"
                    })
            else
                send_response(w, 200, "OK", "application/json",
                    cjson.encode{
                        ok = true,
                        message = "Update in progress",
                    })
            end
            return
        end
        send_response(w, 200, "OK", "application/json",
            cjson.encode{
                ok=false,
                error=string.format("Unknown unit type %s", unit)
            })
        return
    end

    -- KLUDGE allow PUT here as well
    local mp = assert(multipart_file(w, r))
    assert(shellf("/bin/gunzip -t '%s'", mp.tmp_filename))
    assert(shellf("/bin/rm -rf /var/cache/update"))
    if mp.filename:match('\\') then
        mp.filename = mp.filename:match('^.+\\(.+%..+)') or mp.filename
    end
--[[
    assert(shellf("/bin/cp -a '%s' /var/cache/update/%s",
                  mp.tmp_filename, mp.filename or '__ng-update.tar.gz'))
]]
    assert(shellf("/bin/mv '%s' /__ng-update.tar.gz", mp.tmp_filename))
    local device = hostname:match('^(.-)%-%x+$')
    if w.query.update_all and device then
        local net_db = network_db:new()
        local my_ip = net_db:get_module_link_local_ip()
        net_db:close()
        if my_ip then
            -- Make sure the firmware update gets propagated to ALL nodes that
            -- are the same device as me!
            local rrs = w.server.network:QueryRR(
                '_armada-device-info', 'TXT', '*' .. device .. '*', '*')
            for _, rr in ipairs(rrs) do
                local j = cjson.decode(rr[3])
                if j then
                    if j.hostname and j.hostname ~= hostname and
                        j.internal_ll_ip then
                        log:log(log.NOTICE, string.format(
                            "Updating firmware on %s at IP address %s",
                            device:upper(), j.internal_ll_ip))
                        local pwd =
                            get_remote_password(w.server.network, j.hostname)
                        os.execute('DROPBEAR_PASSWORD="' .. pwd .. '" ' ..
                            'dbclient -y root@' ..  j.internal_ll_ip ..
                            ' "nc ' .. my_ip .. ' 1234 >/__ng-update.tar.gz;' ..
                            'sync;sync;/usr/bin/reboot"')
                    end
                end
            end
        end
    end
    ev.Timer.new(function() shellf("/usr/bin/reboot") end, 1, 0):start(loop)
    send_response(w, 200, "OK", "text/plain", '{"ok":true}\n')
end

function api.system.reboot(loop, w, r)
    local unit = w.query.unit
    local method = r.method:lower()
    if method == 'get' and unit then
        local ip, _ = w.server.network:ResolveHostname(unit)
        if not ip then
            local f = {
                result = 'fail',
                msg = string.format("Unable to get IP address for %s", unit)
            }
            send_response(w, 200, "OK", "application/json", cjson.encode(f))
            return
        end
        local pwd = get_remote_password(w.server.network, unit)
        local cmd = string.format(
            "DROPBEAR_PASSWORD='" .. pwd .. "' /usr/bin/dbclient -y root@%s " ..
            "'sync;sync;/usr/bin/reboot'", ip)
        os.execute(cmd)
        local f = {
            result = 'success',
            msg = string.format("%s is being rebooted", unit)
        }
        send_response(w, 200, "OK", "application/json", cjson.encode(f))
        return
    end
    send_response(w, 200, "OK", "application/json", '{"ok":true}\n')
    ev.Timer.new(function() shellf("/usr/bin/reboot") end, 1, 0):start(loop)
end

function api.system.push_notification(loop, w, r)
    -- This has to be run in a fork to not cause problems!
    local f = w.server:fork(loop, w, "PN Support", {}, function()
            local action = w.query.action
            local method = r.method:lower()
            if method == 'get' and action then
                local resp, err
                if action == 'create' then
                    local email = w.query.email
                    local name = w.query.name
                    if not email or not name then
                        send_response(w, 200, "OK", "application/json",
                            cjson.encode{
                                ok = false,
                                error = "'email' and 'name' parameters required"
                            })
                        return
                    end
                    resp, err = pn.newid(email, name)
                elseif action == 'verify' then
                    local chan_id = w.query.id
                    if not chan_id then
                        send_response(w, 200, "OK", "application/json",
                            cjson.encode{
                                ok = false,
                                error = "'id' parameter required"
                            })
                        return
                    end
                    resp, err = pn.verify(chan_id)
                elseif action == 'test' then
                    local chan_id = w.query.id
                    if not chan_id then
                        send_response(w, 200, "OK", "application/json",
                            cjson.encode{
                                ok = false,
                                error = "'id' parameter required"
                            })
                        return
                    end
                    resp, err = pn.send(chan_id, 'Test Subject', 'Test Message')
                end
                send_response(w, 200, "OK", "application/json",
                    cjson.encode{
                        ok = not err,
                        error = err and resp,
                        response = not err and resp,
                    })
                return
            end
            send_response(w, 200, "OK", "application/json", cjson.encode{
                ok = false,
                error = "Unknown method or action",
            })
    end)
    if not f.ok then
        send_response(w, 500, "ERROR", "application/json", cjson.encode(f))
    end
end

local function available_channels(loop, w, r)
    local net_db = network_db:new()
    net_db:begin_transaction()
    local channels = net_db:available_channels()
    net_db:commit()
    net_db:close()
    local ss = setup_spec:new(channels)
    local result = ss:export()
    send_response(w, 200, "OK", "application/json", result)
end

function api.system.available_channels(loop, w, r)
    local f = w.server:fork(loop, w, "System Available Channels", 10,
        function() available_channels(loop, w, r) end)
    if f.ok then
        w.server.close(w.server.ev_loop, w)
    else
        send_response(w, 500, "ERROR", "application/json", cjson.encode(f))
    end
end

local function get_unique_name(db, name)
    if not name then return end
    local unique_name = name
    local last = 0
    local found = 0
    -- Create SQL statement
    local find_setups = db:prepare(string.format([[
        SELECT content
        FROM "%s"
        WHERE id = 'summary' AND deleted = '';
    ]], 'tsd_setups'))
    -- Run SQL statement
    local got = find_setups:exec{}
    if type(got) == 'table' and #got > 0 then
        local t = {}
        -- Does exact 'name' exist?
        for i = 1, #got do
            if got[i].content then
                if got[i].content:match('"test_name":"(.-)",.+') == name then
                    found = found + 1
                end
            end
        end
        -- Does exact 'name_nnn' exist?
        for i = 1, #got do
            if got[i].content then
                local _name = got[i].content:match('"test_name":"(.-)",.+')
                for j = 1, 999 do
                    if _name == string.format("%s_%03d", name, j) then
                        last = j
                    end
                end
            end
        end
        if found > 0 then
            unique_name = string.format("%s_%03d", name, last + 1)
        else
            unique_name = name
        end
    end
    -- Cleanup SQL statement
    find_setups:finalize()
    return unique_name
end

local function get_extension(filename)
    if filename then
        return filename:match('.+%.(.+)$') or filename
    end
end

local function extract_setup_name(filename)
    if filename then
        if filename:match('\\') then
            return filename:match('^.+\\(.+)%..+') or filename
        end
        return filename:match('(.+)%..+')
    end
end

local function delete_preexisting_sdbx( sensor_db_name ,w)
    --lua doesn't like trying to string match with () characters.
    local name = sensor_db_name:gsub( '%(' ,'%%%(')
    name = name:gsub( '%)' ,'%%%)')
    local db = w.server.setup_db
    local sdbx_dbs = sql_exec( db ,'tsd_setups'
            ,[[SELECT db_name,content FROM "%s" WHERE deleted = '' and id = "summary";]])

    local db_name
    for i,v in ipairs( sdbx_dbs) do
        if v.content:match( name) then
            db_name = v.db_name
            break
        end
    end

    if db_name then
        sql_exec( db ,'tsd_setups' ,[[DELETE FROM "%s" WHERE db_name = ?;]]
                ,{db_name})
    end
end    

local function update_sdbx_content( w ,r ,sdbx_file)
    local db = w.server.setup_db
    local id = sql_exec( db ,'tsd_setups' 
                    ,[[SELECT id FROM "%s" WHERE db_name = ?;]]
                    ,{ r.id} )
    if id then
        id = id[1].id
        id = id:match( "%w+%-%w+%-%w+%-%w+%-%w+" )
        local content = table.concat({ w.sdbx_actual ,', "id":"' ,id ,'"}' })

        sql_exec( db ,'tsd_setups' 
            ,[[UPDATE "%s" SET content = ? WHERE id = ?;]]
            ,{ content ,id})

        os.execute(string.format('mv /tmp/%s /storage/sdbx/%s', sdbx_file,
            sdbx_file))
    end
end

function api.test.import(loop, w, r)
    local ftype = w.query.type
    local origin = w.query.origin
    local extended =
        w.query.extended_names and ' enable_extended_signal_names' or ''

    local supported_types = { sxr = true, dbc = true, sdbx = true, xlsx = true, txt = true }

    if not supported_types[ftype] then
        log:error('Invalid query field "type", must be "sxr", "dbc", '..
            '"sdbx", "xlsx" or "txt", got: '..tostring(ftype)..'"')
        send_response(w, 500, "ERROR", "text/plain",
            cjson.encode{
                setup_id = r.id,
                msg = 'Invalid query field "type", must be "sxr", "dbc", '..
                    '"sdbx" or "txt", got: '..tostring(ftype)..'"',
            })
        return
    elseif (origin ~= "database" and origin ~= "setup") then
        log:error('Invalid query field "origin", must be "database" or '..
            '"setup", got: "'..tostring(origin)..'"')
        send_response(w, 500, "ERROR", "text/plain",
            cjson.encode{
                setup_id = r.id,
                msg = 'Invalid query field "origin", must be "database" or '..
                    '"setup", got: "'..tostring(origin)..'"',
            })
        return
    end
    if origin == "setup" and not (ftype == "sxr" or ftype == 'xlsx') then
        log:warn("Invalid setup. File must be of type *.sxr or *.xlsx.")
        send_response(w, 200, "OK", "text/plain",
            cjson.encode{
                setup_id = r.id,
                msg = "Invalid setup. "..
                "File must be of type *.sxr or *.xlsx."
            })
        return
    end
    local filename
    local sdbx_file = (ftype=='sdbx')
    r.query = r.query or {}
    if r.post and w.multipart then
        for i = 1, #w.multipart do
            local mp = w.multipart[i]
            if mp.tmp_filename then
                filename = mp.filename

                if ftype == 'xlsx' then
                    local ok
                    local data
                    local pcall_ok, pcall_err = pcall(function()
                        local xlsx = require 'xlsx'
                        ok, data = xlsx.import(mp.tmp_filename)
                    end)
                    if not pcall_ok then
                        log:log(log.ERR, "Excel import failed: " .. pcall_err)
                        send_response(w, 500, "ERROR", "application/json",
                            cjson.encode({ setup_id = r.id, msg = pcall_err }))
                        return
                    elseif not ok then
                        log:log(log.ERR, "Excel import failed: " .. data)
                        send_response(w, 500, "ERROR", "application/json",
                            cjson.encode({ setup_id = r.id, msg = data }))
                        return
                    end
                    w.content = cjson.encode(data)
                else
                    local fh = io.open(mp.tmp_filename)
                    w.content = fh:read('*a')
                    fh:close()
                end

                if ftype == 'dbc' or ftype == 'txt' or ftype == 'sdbx' then
                    assert(shellf("mv %s '/tmp/%s'", mp.tmp_filename, filename))
                    shellf("rm /tmp/%s.sxr", extract_setup_name(filename))
                    if ftype == 'dbc' then
                        assert(shellf("dbc2sxr '/tmp/%s'%s",
                            filename, extended))
                    elseif ftype == 'txt' then
                        assert(shellf("can2sxr '/tmp/%s'%s",
                            filename, extended))
                    else
                        assert(shellf("mdb-translation-main '/tmp/%s'", filename))
                    end
                    local fh = io.open(string.format("/tmp/%s.sxr",
                        extract_setup_name(filename)))
                    if fh then
                        w.content = fh:read('*a')
                        fh:close()
                        shellf("rm /tmp/%s.sxr", extract_setup_name(filename))
                        if sdbx_file then
                            fh = io.open('/tmp/sensor_delta.sxr')
                            w.sdbx_actual = fh:read('*a')
                            fh:close()
                            sdbx_file = filename
                            delete_preexisting_sdbx( sdbx_file ,w)
                        end
                    end
                end
                break
            end
        end
    end
    filename = filename or r.query.filename or 'imported_file'
    local migration_result
    do
        r.id = r.id or uuid.generate()
        local ss = setup_spec:new()
        local ok, setup = pcall(ss.import, ss, w.content)
        if not ok or not setup then
            --log:warn('Unable to read SXR file content.\n' ..
            --            'File content may be corrupt')
            send_response(w, 500, "ERROR", "application/json",
                cjson.encode{
                    w.content
                --    setup_id = r.id,
                --    msg = '',
                })
            return
        end
        local ok, err = pcall(ss.validate_structure, ss)
        if not ok then
            log:warn('SXR file content validation failure.\n' ..
                        'File content may be corrupt')
            send_response(w, 500, "ERROR", "application/json",
                cjson.encode{
                    setup_id = r.id,
                    msg = 'SXR file content validation failure.\n' ..
                        'File content may be corrupt',
                })
            return
        end
        if ftype == 'sxr' then
            local ok, err = pcall(ss.user_order, ss)
            if not ok then
                --"Changing to text/plain fixes IE
                send_response(w, 200, "OK", "text/plain",
                    cjson.encode{
                        setup_id = r.id,
                        msg = err,
                    })
                return
            end
        end
        local summary
        local ids = {}
        for i = 1, #setup do
            local item = setup[i]
            -- KLUDGE this next bit is disconcerting
            if item.id == 'summary' or item.category == 'summary' then
                item.id = 'summary'
                item.category = 'summary'
                item.setup_id = r.id
                summary = item
            else
                local new_id = uuid.generate()
                if item.id then
                    ids[item.id] = new_id
                end
                item.id = new_id
            end
        end
        if not summary then
            summary = {
                setup_id = r.id,
                id = 'summary',
                category = 'summary',
            }
            for k, v in pairs{
                test_author = 'Default Author', -- KLUDGE
                test_title = 'Default Title', -- KLUDGE
                test_synopsis = 'Default Synopsis', -- KLUDGE
            } do
                summary[k] = r.query[k] or v
            end
            table.insert(setup, 1, summary)
        end

        -- if persisted plots exist, sync the ids here as well
        if summary.charts then
            summary.charts.setup_id = r.id
                if summary.charts.configurations and #summary.charts.configurations > 0 then
                local configs = summary.charts.configurations
                for i = 1, #configs do
                    local config = configs[i]
                    if config and config.chart_state and #config.chart_state > 0 then
                        local chart_states = config.chart_state
                        for j = 1, #chart_states do
                            local chart_state = chart_states[j]
                            if chart_state and chart_state.signals and #chart_state.signals > 0 then
                                local signals = chart_state.signals
                                for k = 1, #signals do
                                    local signal = signals[k]
                                    -- we have a signal, replace its id with the new one
                                    if signal then
                                        signal.id = ids[signal.id]
                                    end
                                end
                            end
                            if chart_state and chart_state.view then
                                for k, v in pairs(chart_state.view) do
                                    if ids[k] then
                                        chart_state.view[ids[k]] = v
                                        chart_state.view[k] = nil
                                    end
                                end
                            end
                        end
                    end
                end
            end
        end

        -- Migrate the setup from it's source version to the version needed on
        -- this machine.
        setup, migration_result =
            setup_migrator.migrate_setup(summary.setup_id, setup)

        local test_name = extract_setup_name(filename) or summary.test_name
        summary.test_name = get_unique_name(w.server.setup_db, test_name)

        local delta = {old={}, new={}}
        for i = 1, #setup do
            delta.old[setup[i].id] = cjson.null
            delta.new[setup[i].id] = setup[i]
        end
        w.content = cjson.encode(delta)
    end
    local frag = string.format('/setups/%s/_delta', r.id)
    if not r.query.description then
        r.query.description = string.format('Import "%s"', filename)
    end
    w.api_internal.import = true
    w.headers.REQUEST_METHOD = 'POST' -- KLUDGE to make tracked_db happy
    local got = handlers.tracked_db(loop, w, frag)
    local ok, result = got.ok, got.result
    if ok then
        result = cjson.encode{
            setup_id = r.id,
            migration_result = migration_result or 'No migration required\n',
        }
    end
    send_response(w, ok and 200 or 500, ok and "OK" or "ERROR",
        "text/plain", result)
    if sdbx_file then
        update_sdbx_content( w ,r ,sdbx_file)
    end
end

local function _get_sxr(loop, w, sie_id)

    local db = w.server.setup_db

    local r, ok = sql_exec(db, 'tsd_setups', [[
        SELECT content
        FROM "%s"
        WHERE db_name = ? AND id != '' AND deleted = '';
    ]], { sie_id })

    if ok and #r > 0 then

        local result
        local setup = {}
        for i = 1, #r do setup[i] = cjson.decode(r[i].content) end
        local ss = setup_spec:new(setup)
        local ok, err = pcall(function() result = ss:export('user_order') end)
        if not ok then return false, 'Failed to sort test in user order' end
        if not result or type(result) ~= 'string' then
            return false, 'Unexpected result from test export sorter'
        end
        return true, result

    else return false, 'Failed to look up setup record in database' end

end

function api.test.export(loop, w, r)

    local sie_name = r.id .. '.sxr'
    local disp = r.filename and r.filename or sie_name
    local file_type = r.filename:match('%.([a-zA-Z]+)$') or 'sxr'
    w.api_internal.export = true
    local frag = string.format('/setups/%s/_all_docs', r.id)
    table.insert(w.out_headers,
        'Content-Disposition: attachment; filename="' .. disp .. '"\n'
    )

    local ok, data = _get_sxr(loop, w, r.id)
    local content_type

    if not ok then
        send_response(w, 500, "ERROR", "application/json",
            cjson.encode({ ok = false, url = frag, msg = data }))
        return
    end

    if file_type == 'xlsx' then
        content_type = 'application/vnd.openxmlformats-' ..
            'officedocument.spreadsheetml.sheet'
        local pcall_ok, pcall_result = pcall(
            function()
                local xlsx = require 'xlsx'
                local tmp_filename = '/tmp/' ..
                    r.id:gsub('[^%w%d%_%-]', '_') .. '.xlsx'
                local ok, msg = xlsx.export(cjson.decode(data), tmp_filename)
                if not ok then
                    log:log(log.ERR, "Excel export failed: " .. msg)
                    return false, msg
                end
                local f = io.open(tmp_filename, 'rb')
                if not f then
                    return false, 'Failed to open xlsx conversion'
                end
                data = f:read('*a')
                f:close()
                os.remove(tmp_filename)
            end)
        if not pcall_ok then
            log:log(log.ERR, "Excel export failed: " .. pcall_result)
            send_response(w, 500, "ERROR", "application/json",
                cjson.encode({ ok = false, url = frag, msg = pcall_result }))
            return
        end
    elseif file_type == 'sxr' then
        content_type = 'application/json'
    end

    downloadCompleteNotification(w.session_key)
    send_response(w, 200, "OK", content_type, data)

end

function api.test._start(loop, w, r, pending_test_id)
    -- Dump the platform info for all of the MX devices to a file that can
    -- be used by the interrupted test restart mechanism to reboot the MX
    -- devices if they fail to "announce" after the reboot.
    if mx_pis and next(mx_pis) then
        local fh = io.open('/var/lib/armada/mx_pi.json', 'w')
        if fh then
            fh:write(cjson.encode(mx_pis))
            fh:close()
        end
    end

    -- Decode the JSON
    local s = cjson.decode(w.content)
    if not s then
        delete_pending_test_pidfile(pending_test_id)
        send_response(w, 200, "OK", "application/json", cjson.encode{
                msg = "Unable to parse JSON from SXR",
                severity = 'normal',
                type = 'error',
                id = uuid.generate(),
                title = 'Test Start Failure',
        })
        return
    end

    if not w.query.friendly_name then
        w.query.friendly_name = r.id
    end

    -- Make a UUID...E.G.  Ignore UUID from user
    r.id = pending_test_id

    local friendly_name = 'dummy_name'
    local sie_name = '/dev/null'

    local summary
    for _, stanza in ipairs(s) do
        if stanza.category and stanza.category == 'summary' then
            summary = stanza

            if not summary.run_description_1 then
                summary.run_description_1 =
                    w.query.run_description_1 or 'No Run Description'
            end

            -- If this is an SIE test then figure out the actual name to use
            -- for the SIE and write the restart test stuff
            if not summary.purpose then
                -- If the system is out of SIE storage space,
                -- don't allow test to start
                if fs_info.get_storage_available() < (2048*1024) then
                    send_response(w, 200, "OK", "application/json",
                        cjson.encode{
                            msg = "Can't start test because the system is "..
                                  "out of storage space.",
                            severity = 'normal',
                            type = 'error',
                            id = uuid.generate(),
                            title = 'Test Start Failure',
                    })
                    delete_pending_test_pidfile(pending_test_id)
                    return
                end

                friendly_name = w.query.friendly_name or r.id
                sie_name = string.format('%s/%s.sie', storage_path, r.id)
                if friendly_name and w.query.append then
                    sie_name =
                        sie_file_mgr.get_sie_path_on_disk(friendly_name) or sie_name
                else
                    --friendly_name = sie_file_mgr.assign_friendly_sie_name(
                    --    friendly_name, sie_name, summary.test_author,
                    --    summary.test_summary, summary.sie_filename_mode)
                end

                local fd = C.open(storage_path .. '/.temp.info',
                    C.O_RDWR+C.O_CREAT, ffi.new('int', 438))
                if fd < 0 then
                    local e = ffi.string(C.strerror(ffi.errno()))
                    send_response(w, 200, "OK", "application/json",
                        cjson.encode{
                            msg = string.format(
                                "Unable to create test restart file: %s", e),
                            severity = 'normal',
                            type = 'error',
                            id = uuid.generate(),
                            title = 'Test Start Failure',
                    })
                    delete_pending_test_pidfile(pending_test_id)
                    return
                end
                local restart_info = cjson.encode{
                    setup = cjson.decode(w.content),
                    duration = w.query.duration or summary.duration,
                    id = r.id,
                    friendly_sie_name = friendly_name
                }
                if C.write(fd, restart_info, #restart_info) ~=
                    #restart_info then
                    local e = ffi.string(C.strerror(ffi.errno()))
                    C.close(fd)
                    send_response(w, 200, "OK", "application/json",
                        cjson.encode{
                            msg = string.format(
                                "Unable to write test restart file: %s", e),
                            severity = 'normal',
                            type = 'error',
                            id = uuid.generate(),
                            title = 'Test Start Failure',
                    })
                    delete_pending_test_pidfile(pending_test_id)
                    return
                end
                if C.close(fd) ~= 0 then
                    local e = ffi.string(C.strerror(ffi.errno()))
                    send_response(w, 200, "OK", "application/json",
                        cjson.encode{
                            msg = string.format(
                                "Unable to close test restart file: %s", e),
                            severity = 'normal',
                            type = 'error',
                            id = uuid.generate(),
                            title = 'Test Start Failure',
                    })
                    delete_pending_test_pidfile(pending_test_id)
                    return
                end
                os.rename(storage_path .. '/.temp.info',
                    storage_path .. '/.test-to-restart.info')
            end
            break
        end
    end

    -- The test restart record exists from this point on.  Any failure detected
    -- below MUST remove the test restart record.

    local spec = setup_spec:new(s)
    
    local setup = nil
    local ok, err = pcall(function() setup = spec:export() end)
    if not ok then
        send_response(w, 200, "OK", "application/json",
            cjson.encode({ok=ok, error_message=err}))
        os.remove(storage_path .. '/.test-to-restart.info')
        delete_pending_test_pidfile(pending_test_id)
        return
    end

    -- KLUDGE error on w.server.running_tests_kludge[r.id]
    local setup_name, fd = w.server:temp_file(w, '/tmp/nginx/test_', '', true)
    if not fd then
        send_response(w, 200, "OK", "application/json",
            cjson.encode{
                msg = "Unable to create temporary test file",
                severity = 'normal',
                type = 'error',
                id = uuid.generate(),
                title = 'Test Start Failure',
        })
        os.remove(storage_path .. '/.test-to-restart.info')
        delete_pending_test_pidfile(pending_test_id)
        return
    end
    if C.write(fd, setup, #setup) ~= #setup then
        C.close(fd)
        send_response(w, 200, "OK", "application/json",
            cjson.encode{
                msg = "Unable to write temporary test file",
                severity = 'normal',
                type = 'error',
                id = uuid.generate(),
                title = 'Test Start Failure',
        })
        os.remove(storage_path .. '/.test-to-restart.info')
        delete_pending_test_pidfile(pending_test_id)
        return
    end
    if C.close(fd) ~= 0 then
        send_response(w, 200, "OK", "application/json",
            cjson.encode{
                msg = "Unable to close temporary test file",
                severity = 'normal',
                type = 'error',
                id = uuid.generate(),
                title = 'Test Start Failure',
        })
        os.remove(storage_path .. '/.test-to-restart.info')
        delete_pending_test_pidfile(pending_test_id)
        return
    end

    -- If this is NOT a test restart, reset ALL of the interactive triggers
    if not w.query.restart then
        -- initialize all triggers to false
        local owner
        for trig_i = 1, 8 do
            owner = '_interactive-trigger-' .. tostring(trig_i)
            w.server.network:RemoveLocalRROwner(owner)
            w.server.network:ModifyRRs({},
                {{owner, 'TXT', owner ..  ":" .. "false"}})
        end
    end

    local function add(t, a)
        if a then
            if type(a) == 'table' then
                for _, x in ipairs(a) do
                    add(t, x)
                end
            else
                t[#t + 1] = a
            end
        end
    end

    local cmd = {}
    add(cmd, w.query.strace and {
        '/usr/bin/strace',
        '-s1024',
        '-x',
        string.format('-o%s.strace', setup_name),
    })
    add(cmd, {
        '/usr/bin/run-test',
        setup_name,
        sie_name,
        w.query.duration or summary.duration or 'null',
        r.id,
        w.query.restart or 'null',
        w.query.suspend or 'null',
        friendly_name
    })

    -- OK spawn the test
    local test_pid = w.server.master:SpawnTest(cmd[1], r.id,
        summary.restart_count or 0, cmd)

    local pid_filename = '/var/run/pending-test-' .. pending_test_id .. '.pid'
    local f = io.open(pid_filename, 'w')
    if f then
        f:write(test_pid .. "\n")
        f:close()
    end

    -- Test is running, send success
    if test_pid > 0 then
        send_response(w, 200, "OK", "application/json", cjson.encode{
            ok = true,
            test_id = r.id,
            --test_sie_file = friendly_name,
        })
    else
        send_response(w, 200, "OK", "application/json",
            cjson.encode{
                msg = "Unable to run test, spawn failed",
                severity = 'normal',
                type = 'error',
                id = uuid.generate(),
                title = 'Test Start Failure',
        })
        os.remove(storage_path .. '/.test-to-restart.info')
        delete_pending_test_pidfile(pending_test_id)
    end
end

local function write_pending_test_pidfile(pending_test_id, pid)
    local filename = '/var/run/pending-test-' .. pending_test_id .. '.pid'
    local f = io.open(filename, 'w')
    if f then
        f:write(pid .. "\n")
        f:close()
    end
end

local function read_pending_test_pidfile(pending_test_id)
    local filename = '/var/run/pending-test-' .. pending_test_id .. '.pid'
    local f = io.open(filename, 'r')
    if not f then return nil end
    local pid = tonumber(f:read('*a'):match('(-?%d+)'))
    f:close()
    return pid
end

local function delete_pending_test_pidfile(pending_test_id)
    local filename = '/var/run/pending-test-' .. pending_test_id .. '.pid'
    os.execute('rm ' .. filename .. ' 2>/dev/null')
end

-- FIXME: Remove this when the GUI is modified to use start with
-- 'initialize_only' parameter
function api.test.initialize(loop, w, r)
    api.test.start(loop, w, r, {init_only = true})
end

function api.test.start(loop, w, r, add_to_summary)
    if w.query.initialize_only then
        add_to_summary = {init_only = true}
    end

    -- If start is locked due to interrupted test restart in progress, deny the
    -- start request
    if start_is_locked then
        send_response(w, 200, "OK", "application/json", cjson.encode({
                msg =
                    'Restart of test stopped by system shutdown is ' ..
                    'in progress.  Click "Abort" if you want to abort ' ..
                    'interrupted test restart and enable user initiated ' ..
                    'test start. Click "Close" if you want the interrupted ' ..
                    'test restart to continue.',
                severity = 'normal',
                type = 'warning',
                id = uuid.generate(),
                title = 'Interrupted Test Restart In Progress',
                buttons = {
                    {
                        text='Abort',
                        uri='/api/system/cancel_restart',
                    },
                    {
                        text='Close',
                    }
                },
            }))
        return
    end

    -- FIXME:  When this restriction is lifted from the eDAQXR, remove this
    if w.server.network.IsNetworkTopologyValid then
        local v = w.server.network:IsNetworkTopologyValid()
        if w.server.network:IsNetworkTopologyValid() == 0 then
            send_response(w, 200, "OK", "application/json", cjson.encode({
                    msg =
                        "Unable to start test.  " ..
                        "Connection of only ETH2 has been detected and " ..
                        "that is not valid use of ETH1/ETH2.  If ETH2 is " ..
                        "connected ETH1 must also be connected.  Please " ..
                        "correct the network connections and try again.",
                    severity = 'normal',
                    type = 'warning',
                    id = uuid.generate(),
                    title = 'Unsupported ETH1/ETH2 Network Connections',
                }))
            return
        end
    end

    -- If any XR node in the system has a different firmware version, deny
    -- test start
    local _, pi = next(xr_pis)
    for _, pi2 in pairs(xr_pis) do
        if pi.version ~= pi2.version then
            send_response(w, 200, "OK", "application/json", cjson.encode({
                    msg =
                        "Unable to start test.  " ..
                        "Two or more XR devices are connected together and " ..
                        "form a 'networked' system and not all of the " ..
                        "devices are running the same firmware version.  " ..
                        "All devices must be running the same firmware " ..
                        "version before a test may be run.  Please update " ..
                        "all XR nodes to the same firmware version as is " ..
                        "running on this XR device or remove offending " ..
                        "nodes from the network.",
                    severity = 'normal',
                    type = 'warning',
                    id = uuid.generate(),
                    title = 'Networked Devices Firmware Inconsistency',
                }))
            return
        end
    end

    -- If a test request is pending in this system, deny the start request.
    -- Tests are pending from the moment they're requested until either the
    -- test shows up in the network database, or the run-test process exits.
    if #pending_tests > 0 then
        send_response(w, 200, "OK", "application/json", cjson.encode({
                msg =
                    "A test request is pending in this system.  This " ..
                    "is sometimes a result of clicking 'start' in rapid " ..
                    "succession.  Please wait for the current test " ..
                    "request to finish.",
                severity = 'normal',
                type = 'warning',
                id = uuid.generate(),
                title = 'Test Engine Busy',
            }))
        return
    end

    -- If the test is being started by test id rather than sent with the
    -- request, get the test from the setup database.  If the test can't be
    -- found, send a warning back.
    if w.query and w.query.setup_id then
        local r, ok = sql_exec(w.server.setup_db, 'tsd_setups',
            [[SELECT content FROM "%s" WHERE db_name = ? AND id != ''
                AND deleted = '';]],
            {w.query.setup_id})
        if ok and r and #r > 1 then
            -- Make sure the test engine gets the "user order" sorted version
            -- of the SXR
            local cj = cjson.decode(json_concat(r))

            -- If there are things to add to the "summary" stanza, add them
            if add_to_summary then
                for i, v in ipairs(cj) do
                    if v.category and v.category == 'summary' then
                        for k, v in pairs(add_to_summary) do
                            cj[i][k] = v
                        end
                        break
                    end
                end
            end

            local s = setup_spec:new(cj)
            w.content = s:export('user_order')
        else
            send_response(w, 200, "OK", "application/json", cjson.encode({
                msg = "Unable to find test id '" .. w.query.setup_id ..
                    "' on this system.",
                severity = 'normal',
                type = 'warning',
                id = uuid.generate(),
                title = 'Test Not Found',
            }))
            return
        end
    elseif add_to_summary then
        -- If there are things to add to the "summary" stanza, add them
        local cj = cjson.decode(w.content)
        for i, v in ipairs(cj) do
            if v.category and v.category == 'summary' then
                for k, v in pairs(add_to_summary) do
                    cj[i][k] = v
                end
                break
            end
        end
        w.content = cjson.encode(cj)
    end

    start_pending_test_watchdog()
    local pending_test = {id=uuid.generate()}
    write_pending_test_pidfile(pending_test.id, -1)
    pending_tests[#pending_tests + 1] = pending_test

    local ok = pcall(function()

        -- If a test is already running in this system, deny the start request.
        -- FIXME:  This is system wide...so test start will be denied if a a test
        -- is running on ANY of the devices that are interconnected.
        if w.server.running_tests and w.server.running_tests > 0 then
            table.remove(pending_tests)
            delete_pending_test_pidfile(pending_test.id)
            send_response(w, 200, "OK", "application/json", cjson.encode({
                    msg =
                        "A test is already running in this system.  This " ..
                        "is sometimes a result of clicking 'start' in rapid " ..
                        "succession.  The currently running test must be " ..
                        "stopped before this operation can be performed.",
                    severity = 'normal',
                    type = 'warning',
                    id = uuid.generate(),
                    title = 'Test Engine Busy',
                }))
            return
        end

        -- Wait here for the start delay to expire
        while start_delay do
            ev_loop:loop('once')
        end

        -- This stuff should all be done in a fork but doing so is extremely
        -- complicated because stuff create in the (would be) child needs to end
        -- up back in the parent!!  UGH!!!
        local f = w.server:fork(loop, w, "Test Start", {}, function()
                pcall(api.test._start, loop, w, r, pending_test.id)
            end)
        if f.ok then
            write_pending_test_pidfile(pending_test.id, f.pid)
            send_response(w, 200, "OK", "application/json", '{"ok":true}\n')
        else
            table.remove(pending_tests)
            delete_pending_test_pidfile(pending_test.id)
            send_response(w, 200, "OK", "application/json", cjson.encode({
                    msg = 'Unable to start test due to fork failure.',
                    severity = 'normal',
                    type = 'warning',
                    id = uuid.generate(),
                    title = 'Test Start Failure',
                }))
        end
        end)

    if not ok then
        table.remove(pending_tests)
        delete_pending_test_pidfile(pending_test.id)
    end

end

function api.test.enable(loop, w , r)
    local test_pid = w.server.running_tests_kludge[r.id]
    local err = nil
    if not test_pid then
        send_response(w, 200, "OK", "application/json", cjson.encode{
            result = 'error',
            msg = 'No such test',
        })
        return
    end
    C.kill(ffi.new('pid_t', tonumber(test_pid)), u.SIGUSR1)
    send_response(w, 200, "OK", "application/json", '{"ok":true}\n')
end

function api.test.suspend(loop, w , r)
    local test_pid = w.server.running_tests_kludge[r.id]
    local err = nil
    if not test_pid then
        send_response(w, 200, "OK", "application/json", cjson.encode{
            result = 'error',
            msg = 'No such test',
        })
        return
    end
    C.kill(ffi.new('pid_t', tonumber(test_pid)), u.SIGUSR2)
    send_response(w, 200, "OK", "application/json", '{"ok":true}\n')
end

function api.test.init_start(loop, w , r)
    local test_pid = w.server.running_tests_kludge[r.id]
    local err = nil
    if not test_pid then
        send_response(w, 200, "OK", "application/json", cjson.encode{
            result = 'error',
            msg = 'No such test',
        })
        return
    end
    C.kill(ffi.new('pid_t', tonumber(test_pid)), u.SIGALRM)
    send_response(w, 200, "OK", "application/json", '{"ok":true}\n')
end

function api.test.restart(loop, w, r)

    start_pending_test_watchdog()
    local pending_test = {id=uuid.generate()}
    write_pending_test_pidfile(pending_test.id, -1)
    pending_tests[#pending_tests + 1] = pending_test

    local ok = pcall(function()

        local f = w.server:fork(loop, w, "Test Restart", {}, function()
                return api.test._start(loop, w, r, pending_test.id)
            end)
        if f.ok then
            write_pending_test_pidfile(pending_test.id, f.pid)
            send_response(w, 200, "OK", "application/json", '{"ok":true}\n')
        else
            table.remove(pending_tests)
            delete_pending_test_pidfile(pending_test.id)
            send_response(w, 200, "OK", "application/json", cjson.encode({
                    msg = 'Unable to restart test due to fork failure.',
                    severity = 'normal',
                    type = 'warning',
                    id = uuid.generate(),
                    title = 'Test Restart Failure',
                }))
        end

    end)

    if not ok then
        table.remove(pending_tests)
        delete_pending_test_pidfile(pending_test.id)
    end

end

function api.test.uninitialize(loop, w, r)
    os.remove(storage_path .. '/.remote-test.info')
    api.test.stop(loop,w,r,true)
end

local function stop_test(w, url)
    local ip, port = url:match('//%[(.+)%]:(.+)/stop')
    local client = w.server:connect(ip, port, function(loop, w) end)
    if not client then
        log:log(log.NOTICE, "connect fail")
        send_response(w, 200, "OK", "application/json",
            cjson.encode{
                result = 'error',
                msg = 'Unable to stop test on remote device',
            })
        return
    end
    local req = "GET /stop HTTP/1.0\r\n\r\n"
    ev.IO.new(function(l, i, e)
            i:stop(l)
            C.close(client.socket)
            -- A bit of time is needed before the response can be sent back
            ev.Timer.new(function(l, t, e)
                    send_response(w, 200, "OK", "application/json",
                        cjson.encode{
                            result = 'success',
                            msg = 'Remote device test stopped',
                        })
                end, 1, 0):start(ev.Loop.default)
        end, client.socket, ev.READ):start(ev.Loop.default)
    C.write(client.socket, req, #req)
end

function api.test.stop(loop, w, r, uninitialize)
    -- KLUDGE until we have stuff hooked in for real
    local test_pid = w.server.running_tests_kludge[r.id]
    w.server.running_tests_kludge[r.id] = nil
    local err = nil
    if not test_pid then
        -- Maybe not a locally running test
        local urls = w.server.net_db:get_stop_urls()
        if not urls or not urls[r.id] then
            -- No test running anywhere with the specified ID
            send_response(w, 200, "OK", "application/json", cjson.encode{
                result = 'error',
                msg = 'No such test',
            })
        else
            stop_test(w, urls[r.id])
        end
        return
    end

    -- Make sure this test can't be restarted
    os.remove(storage_path .. '/.test-to-restart.info')

    -- Clear cyclic test run status (see #5141 for feature specification)
    os.remove(storage_path .. '/.cyclic-test.info')

    local got = system_config.force_brutal_stop and 0 or
        C.kill(ffi.new('pid_t', tonumber(test_pid)), u.SIGTERM)
    if got ~= 0 then
        send_response(w, 200, "OK", "application/json", cjson.encode{
            result = 'error',
            msg = 'The test was not running when test stop was attempted',
        })
    else
        -- Defer the response until the test actually stops
        local timeout = 60
        ev.Timer.new(function(l, t, e)
                if C.kill(test_pid, 0) ~= 0 then
                    send_response(w, 200, "OK", "application/json",
                        cjson.encode{
                            result = 'success',
                            msg = 'Test stopped',
                        })
                    t:stop(l)
                    return
                end
                timeout = timeout - 1
                if timeout == 0 then
                    send_response(w, 200, "OK", "application/json",
                        cjson.encode{
                            result = 'error',
                            msg = 'Test would not stop, killed!',
                        })
                    t:stop(l)
                    -- Be brutal about getting rid of the test!
                    log:log(log.ERR, "Timeout waiting for test to exit, killed")
                    -- Why all of the conversion...to make sure the argument is
                    -- a string!
                    w.server.master:UnregisterCriticalProcess(
                        string.format("%.0f", tonumber(test_pid)))
                    os.execute('rm /storage/.test-to-restart.info 2>/dev/null')
                    C.kill(test_pid, u.SIGKILL)
                end
            end, 1, 1):start(loop)
        end
end

function api.test.data(loop, w, r)
    local disp = (r.filename or r.id)
    if disp and not disp:find('%.sie$') then
        disp = disp .. '.sie'
    end
    local sie_name = storage_path .. '/' .. disp
    local pod = sie_file_mgr.get_sie_path_on_disk(disp:gsub('%.sie$', ''))
    if pod then
        sie_name = pod
    end
    table.insert(w.out_headers,
        'Content-Disposition: attachment; filename="' .. disp .. '"\n' ..
        'X-Accel-Redirect: ' .. sie_name .. '\n'
    )
    send_response(w, 200, "OK", "application/sie", '')
    downloadCompleteNotification(w.session_key)
end

function api.test.rename(loop, w, r)
    local old_name = w.query.old_name
    local new_name = w.query.new_name

    if not old_name or not new_name then
        send_response(w, 200, "OK", "application/json",
            cjson.encode({
                result = 'fail',
                msg = 'Both old and new name parameters must be given',
            }))
        return
    end

    local name = sie_file_mgr.rename_sie(new_name, old_name)
    if name then
        send_response(w, 200, "OK", "application/json",
            cjson.encode({
                result = 'success',
                msg = 'Name changed',
                name = name,
            }))
    else
        send_response(w, 200, "OK", "application/json",
            cjson.encode({
                result = 'fail',
                msg = 'Unable to change name',
            }))
    end
end

function api.test.data_file_list(loop, w, r)
    send_response(w, 200, "OK", "application/json",
        cjson.encode(sie_file_mgr.get_sie_file_list()))
end

function api.system.get_cal_cert(loop, w, r)
    if w.query.module and w.query.module:match('^mx') then
        local connector = 'control.status@' .. w.query.module
        w.content = cjson.encode({
            connector = connector,
            request = 'get_cal_cert',
            params = {}
        })
        return api.system.control(loop, w, r)
    end

    local f = w.server:fork(loop, w, "System Get Cal Cert", 10, function()
        local filename
        local fp
        if not system_config.device_type or
            system_config.device_type == 'cx23r' then
            fp = io.popen('ls /var/etc/armada/CX23R*|sort -r', 'r')
        else
            fp = io.popen('ls /var/etc/armada/EDAQXR*|sort -r', 'r')
        end
        if fp then
            for line in fp:lines() do
                filename = line
                break
            end
            fp:close()
            if not filename or filename == '' then
                send_response(w, 200, "OK", "text/plain", '{"ok":false, "msg":"No certificate"}\n')
                return
            end
            local cert = io.open(filename, 'r')
            if cert then
                local file_size = cert:seek("end")
                cert:seek("set", 0)
                local header_lines = {}
                table.insert(header_lines, "HTTP/1.1 200 OK\n")
                table.insert(header_lines, "Cache-Control: no-cache\n")
                table.insert(header_lines, "Pragma: no-cache\n")
                table.insert(header_lines, "Content-Type: application/pdf\n")
                table.insert(header_lines, 'Content-Length: ' .. file_size .. "\n")
                table.insert(header_lines, 'Content-Disposition: attachment; ')
                table.insert(header_lines, 'filename="' ..
                                           filename:match('^.+/(.+%..+)') ..
                                           '"\n\n')
                u.write(w.socket, table.concat(header_lines))
                local remaining = file_size
                while remaining > 0 do
                    local amt = 16 * 1024
                    if amt > remaining then amt = remaining end
                    local data = cert:read(amt)
                    u.write(w.socket, data)
                    remaining = remaining - data:len()
                end
                cert:close()
                u.write(w.socket, '\n\n')
            end
        else
            send_response(w, 200, "OK", "text/plain", '{"ok":false}\n')
            return
        end
        downloadCompleteNotification(w.session_key)
    end)
    if f.ok then
        send_response(w, 200, "OK", "application/json", '{"ok":true}\n')
    else
        send_response(w, 500, "ERROR", "application/json", cjson.encode(f))
    end
end

function api.system.set_cal_cert(loop, w, r)
    local mp = multipart_file(w, r)
    if mp and mp.filename and mp.tmp_filename then
        if mp.filename:match('\\') then
            mp.filename = mp.filename:match('^.+\\(.+%..+)') or mp.filename
        end
        shellf("/bin/mv '%s' /var/etc/armada/%s", mp.tmp_filename, mp.filename)
        send_response(w, 200, "OK", "application/json", '{"ok":true}\n')
    else
        send_response(w, 500, "ERROR", "application/json",
                      '{"error_message":"Incomplete multipart file!"}\n')
    end
end

function api.system.get_oscillator_offset(loop, w, r)
    -- This should always return a value, since the calling function will write
    -- in a 0 if no calibration previously exists.  But will check for nil just
    -- in case.
    local offset = armada_mmio:oscillator_offset()
    if offset == nil then
        send_response(w, 500, "ERROR", "application/json",
                      '{"error_message":"no calibration offset returned"}\n')
    else
        send_response(w, 200, "OK", "application/json", '{"offset":' 
                                        .. offset .. ',"ok":true}\n')
    end
end

function api.system.get_ca(loop, w, r)
    local f = w.server:fork(loop, w, "System Get CA", 10, function()
        local fh = io.open('/etc/nginx/cx23.crt', 'r')
        if fh then
            local ca = fh:read("*a")
            fh:close()
            local response = 
                'HTTP/1.1 200 OK\n' ..
                'Cache-Control: no-cache\n' ..
                'Pragma: no-cache\n' ..
                'Content-Type: application/txt\n' ..
                'Content-Length: ' .. tostring(ca:len()) .. '\n' ..
                'Content-Disposition: attachment; filename="cx23.crt"\n\n'  ..
                ca
            while response:len() > 0 do
                local amt, estr, e = u.write(w.socket, response)
                if amt then
                    response = string.sub(response, amt + 1)
                elseif e ~= u.EAGAIN then
                    u.close(w.socket)
                    return
                end
            end
        else
            send_response(w, 200, "OK", "text/plain", '{"ok":false}\n')
            return
        end
    end)
    if f.ok then
        send_response(w, 200, "OK", "application/json", '{"ok":true}\n')
    else
        send_response(w, 500, "ERROR", "application/json", cjson.encode(f))
    end
end

function api.system.configure_channel(loop, w, r)
    local owner = '_config_channel.' .. uuid.generate()
    w.server.network:ModifyRRs({}, {{owner, 'CONFIGURE', w.content}})
    w.server.network:ModifyRRs({{owner, 'CONFIGURE', w.content}}, {})
    send_response(w, 200, "OK", "application/json", '{"ok":true}\n')
end

function api.system.echo_local_file(loop, w, r)
    r.query = r.query or {}
    if r.post and w.multipart then
        for i = 1, #w.multipart do
            local mp = w.multipart[i]
            if mp.tmp_filename then
                local fh = io.open(mp.tmp_filename)
                w.content = fh:read('*a')
                fh:close()
            end
        end
    end
    send_response(w, 200, "OK", "text/plain", w.content)
end

function api.system.set_inter_trig(loop, w, r)
    -- Make sure escape crap is removed! (quoting MikeL)
    w.content = string.gsub(w.content, '\\"', '"')
    w.content = string.gsub(w.content, '\\', '')

    -- Decode the JSON
    local cj = cjson.decode(w.content)
    if not cj then
        local f = {
            result = 'error',
            msg = 'Unable to decode JSON content',
        }
        send_response(w, 200, "OK", "application/json", cjson.encode(f))
    end
    if not cj.trigger_index then
        local f = {
            result = 'error',
            msg = 'JSON content does not have trigger id',
        }
        send_response(w, 200, "OK", "application/json", cjson.encode(f))
    end

    if not cj.value or (cj.value ~= 'true' and cj.value ~= 'false') then
        local f = {
            result = 'error',
            msg = 'JSON content does not have legal trigger value',
        }
        send_response(w, 200, "OK", "application/json", cjson.encode(f))
    end

    -- set interactive trigger in database to specified value
    local owner = '_interactive-trigger-' .. cj.trigger_index
    w.server.network:RemoveLocalRROwner(owner)
    w.server.network:ModifyRRs({}, {{owner, 'TXT', owner ..
        ":" .. cj.value}})
    -- Send success
    local f = {
        result = 'success',
        msg = 'OK',
    }
    send_response(w, 200, "OK", "application/json", cjson.encode(f))
end

function handlers.test(loop, w, frag)
    local method = w.headers.REQUEST_METHOD
    local tomatch = w.headers.REQUEST_METHOD .. ':' .. w.headers.DOCUMENT_URI
    local r = api_url_parser:match(tomatch)

    --[[
    dbg('handlers.test(): %s\n', tomatch)
    dbg('query_string = %s\n', tostring(w.headers.QUERY_STRING))
    dbg('content_type = %s\n', tostring(w.headers.CONTENT_TYPE))
    dbg("r=%s\n", (require 'data_dumper')(r))
    dbg("content=%s\n", tostring(w.content))
    dbg('==============================\n\n')
    --]]

    if not r then
        return send_response(w, 404, "Not Found", "text/plain",
            "No such URL: %s\n", w.headers.DOCUMENT_URI)
    end
    r.action = r.action or r.method:lower()
    assert(r.prefix, 'prefix required')
    api[r.prefix][r.action](loop, w, r)
end

handlers.system = handlers.test
handlers.sie = handlers.test
handlers.session = handlers.test

function m.handler(loop, w)
    local uri = w.headers.DOCUMENT_URI
    local unp
    for _, uri_unprefix_match in ipairs(uri_unprefix_matches) do
        unp = uri:match(uri_unprefix_match)
        if unp then
            break
        end
    end
    local req = nil
    local frag = ""
    local req_name = "" -- KLUDGE

    if unp then
        req = handlers[unp]
        if req then
            req_name = unp
        else
            -- Find the longest matching request
            for k, v in pairs(handlers) do
                if #k > #req_name and #unp >= #k and k == unp:sub(1, #k) then
                    req_name = k
                    req = v
                end
            end
            if req then
                frag = unp:sub(#req_name + 1)
                local f1 = string.sub(frag, 1, 1)
                if f1 ~= '' and f1 ~= '/' and f1 ~= '?' then
                    req = nil
                end
            end
        end
    end

    if not w.authorized then
        local addr = w.headers.REMOTE_ADDR
        if (addr == "::ffff:127.0.0.1" or addr == "127.0.0.1")
                and w.authorization_key == local_authorization_key then
            w.authorized = true -- Special case for initial database loads
        else
            w.authorization_key = nil
            w.session_key = nil
            w.authorized = always_authorized_handlers[req_name]
        end
    end

    w.out_headers = {}
    if req and w.authorized then
        w.coro = coroutine.create(function() req(loop, w, frag) end)
        local ok, err = coroutine.resume(w.coro)
        if not ok then
            pcall(function() w.server.setup_db:exec("ROLLBACK;") end)
            send_response(w, 500, "Server Error", "text/plain", err)
        end
    else
        send_response(w, 404, "Not Found", "text/plain",
            "No such URL: %s\n", w.headers.DOCUMENT_URI)
    end
end

return m
