local dbg = require 'dbg'
local cjson = require 'cjson.safe'
local config = require 'system_config'

local function db_exec(db, fmt, ...)
    fmt = fmt:gsub('^[ ]*%|', '')
    fmt = fmt:gsub('\n[ ]*%|%+[ ]*', ' ')
    fmt = fmt:gsub('\n[ ]*%|', '\n')
    fmt = fmt:gsub('[ ]+$', '')
    local sql = string.format(fmt, ...)
    dbg('db_exec, sql=%s\n', sql)
    return db:exec(sql)
end

local function reschema_table(db, table_name, field_map)
    local tmp_table = table_name .. '__reschema'
    local new_fields = {}
    local new_field_names = {}
    local convert_old = {}
    for i = 1, #field_map, 2 do
        local new = field_map[i]
        new_fields[#new_fields + 1] = new
        new_field_names[#new_field_names + 1] = new:gsub(' .*$', '')
        convert_old[#convert_old + 1] = field_map[i + 1]
    end
    db_exec(db, [[
        CREATE TEMPORARY TABLE %s AS SELECT * FROM %s;
    ]], tmp_table, table_name)
    db_exec(db, [[
        DROP TABLE %s;
    ]], table_name)
    db_exec(db, 'CREATE TABLE %s (\n    %s\n);',
        table_name, table.concat(new_fields, ',\n    '))
    db_exec(db, [[
        INSERT INTO %s (%s) SELECT %s FROM %s;
    ]], table_name, table.concat(new_field_names, ', '),
        table.concat(convert_old, ', '), tmp_table)
    db_exec(db, [[
        DROP TABLE %s;
    ]], tmp_table)

end

local m = {}

m['/var/lib/armada/log.db'] = {{
    from = 0, to = 1, updater =
    function(db)
        db_exec(db, [==[
        |CREATE TABLE log (
        |    id INTEGER PRIMARY KEY,
        |    time TEXT NOT NULL,
        |    component TEXT NOT NULL,
        |    level TEXT NOT NULL,
        |    message TEXT NOT NULL
        |);
        ]==])
        return db
    end
},{
    from = 1, to = 2, updater =
    function (db)
        db_exec(db, [==[
        |CREATE INDEX log_time ON log (time);
        |CREATE INDEX log_level ON log (level);
        ]==])
        return db
    end
}}

m['/var/lib/armada/network.db'] = {{
    from = 0, to = 1, updater =
    function (db)
        db_exec(db, [==[
        |CREATE TABLE master (
        |    owner TEXT NOT NULL,
        |    type TEXT NOT NULL,
        |    rdata TEXT NOT NULL,
        |    source TEXT NOT NULL
        |);
        |CREATE UNIQUE INDEX master_unique
        |    ON master (owner, type, rdata, source);
        |
        |CREATE TABLE local (
        |    owner TEXT NOT NULL,
        |    type TEXT NOT NULL,
        |    rdata TEXT NOT NULL,
        |    source TEXT NOT NULL
        |);
        |CREATE UNIQUE INDEX local_unique
        |    ON local (owner, type, rdata, source);
        ]==])
        return db
    end
},{
    from = 1, to = 2, updater =
    function (db)
        db_exec(db, [==[
        |DROP INDEX local_unique;
        |DROP TABLE local;
        |DROP INDEX master_unique;
        |DROP TABLE master;
        ]==])

        db_exec(db, [==[
        |CREATE TABLE unified (
        |    owner TEXT NOT NULL,
        |    type TEXT NOT NULL,
        |    rdata TEXT NOT NULL,
        |    source TEXT NOT NULL,
        |    local_source TEXT NOT NULL DEFAULT ''
        |);
        |
        |CREATE INDEX unified_owner ON unified (owner);
        |CREATE INDEX unified_type ON unified (type);
        |CREATE INDEX unified_rdata ON unified (rdata);
        |CREATE INDEX unified_source ON unified (source);
        |CREATE INDEX unified_local_source ON unified (local_source);
        |CREATE INDEX unified_otrl ON unified (owner, type, rdata, local_source);
        ]==])
        return db
    end
},{
    from = 2, to = 3, updater =
    function (db)
        db_exec(db, [==[
        |CREATE INDEX unified_otrl ON unified (owner, type, rdata, local_source);
        ]==])
        return db
    end
}}

if config.device_type ~= 'dx23r' then
    m['/var/lib/armada/setup.db'] = {{
        from = 0, to = 1, updater =
        function (db)
            -- Remove any old SIE filename mappings
            db_exec(db, "DELETE FROM tsd_users WHERE db_name = 'sie';")
            return db
        end
    },{
        from = 1, to = 2, updater =
        function (db)
            -- Vacuum the database
            db_exec(db, "VACUUM;")
            return db
        end
    },{
        from = 2, to = 3, updater =
        function (db)
            -- Vacuum the database
            db_exec(db,
                "DELETE FROM tsd_users WHERE db_name = 'sie_name';VACUUM;")
            require('sie_file_mgr').check_sie_files()
            return db
        end
    }}
end

return m
