--! \file

local common = require 'mx.device_common'
local underscore = require 'underscore'
local ffi = require 'ffi'
local C = ffi.C
local log = require 'log' 
local m = {}

-- XML used to do initial configuration of channels when proxy starts up
local default_xml = [[<?xml version="1.0" encoding="UTF-8"?>
                <Amplifier>
                    <Connector#conn_num#>
                        <Mode choice="AnalogInChannel">
                            <AnalogInChannel>
                                <Sensor>
                                    <TedsOptions>
                                        <UseTeds>2</UseTeds>
                                    </TedsOptions>
                                </Sensor>
                                <SyncSignal address="1">
                                    <OutputRate>10</OutputRate>
                                </SyncSignal>
                            </AnalogInChannel>
                        </Mode>
                    </Connector#conn_num#>
                </Amplifier>
]]

-- Function used to return the XML used to set the channel to it's default
-- configuration
function m:get_default_xml(channel_num)
    return default_xml:gsub('#conn_num#', tostring(channel_num))
end

m.json2xpath = {
    transducer_power = '/Amplifier/Connector#channel_num#/Mode[@choice="AnalogInChannel"]/AnalogInChannel/Sensor/Supply',
    excitation_range = '/Amplifier/Connector#channel_num#/Mode[@choice="AnalogInChannel"]/AnalogInChannel/Sensor/Sensortype/#sensor_input_mode#/ExciteAmpl',
    excitation_mode = '/Amplifier/Connector#channel_num#/Mode[@choice="AnalogInChannel"]/AnalogInChannel/Sensor/Sensortype/#sensor_input_mode#/ExciteFreq',
    _bridge_type = '/Amplifier/Connector#channel_num#/Mode[@choice="AnalogInChannel"]/AnalogInChannel/Sensor/Sensortype/#sensor_input_mode#/BridgeType',
    bridge_resistance = '/Amplifier/Connector#channel_num#/Mode[@choice="AnalogInChannel"]/AnalogInChannel/Sensor/Sensortype/#sensor_input_mode#/Impedance',
    _bridge_wiring = '/Amplifier/Connector#channel_num#/Mode[@choice="AnalogInChannel"]/AnalogInChannel/Sensor/Sensortype/#sensor_input_mode#/Wiring',
    _pt_wiring = '/Amplifier/Connector#channel_num#/Mode[@choice="AnalogInChannel"]/AnalogInChannel/Sensor/Sensortype[@choice="Pt"]/Pt/Wiring',
    _pt_type = '/Amplifier/Connector#channel_num#/Mode[@choice="AnalogInChannel"]/AnalogInChannel/Sensor/Sensortype/Pt/PtType',
    _s2enbl = '/Amplifier/Connector#channel_num#/Mode[@choice="AnalogInChannel"]/AnalogInChannel/Sensor/Sensortype/#sensor_input_mode#/S2Enbl',
    _f4enbl = '/Amplifier/Connector#channel_num#/Mode[@choice="AnalogInChannel"]/AnalogInChannel/Sensor/Sensortype/#sensor_input_mode#/F4Enbl',
    _indexenbl = '/Amplifier/Connector#channel_num#/Mode[@choice="AnalogInChannel"]/AnalogInChannel/Sensor/Sensortype/#sensor_input_mode#/IndexEnbl',
    _indexdivisor = '/Amplifier/Connector#channel_num#/Mode[@choice="AnalogInChannel"]/AnalogInChannel/Sensor/Sensortype/#sensor_input_mode#/IndexDivisor',
    _ssi_coding = '/Amplifier/Connector#channel_num#/Mode[@choice="AnalogInChannel"]/AnalogInChannel/Sensor/Sensortype/#sensor_input_mode#/SsiCoding',
    _ssi_resolution = '/Amplifier/Connector#channel_num#/Mode[@choice="AnalogInChannel"]/AnalogInChannel/Sensor/Sensortype/#sensor_input_mode#/SsiResolution',
    _ssi_clock = '/Amplifier/Connector#channel_num#/Mode[@choice="AnalogInChannel"]/AnalogInChannel/Sensor/Sensortype/#sensor_input_mode#/SsiClockFreq'
}

underscore.extend(m.json2xpath, common.json2xpath)

m.phase_delay = {}
m.phase_delay.fixed = 398
m.phase_delay.bessel_iir = {}
m.phase_delay.butterworth_iir = {}
m.phase_delay.linear_phase_fir = {}
m.phase_delay.butterworth_fir = {}

m.physical_unit_map = {
    Voltage = 'V',
    Current = 'mA',
    IEPE = 'C',
}

-- List of parameters in a channel setup that we do NOT want sent off to the MX
-- device
m.ignore_list = {}
underscore.extend(m.ignore_list, common.ignore_list)

-- List of parameters returned by MX device that we don't verify after setting
-- up a channel.
m.dont_verify_list = {}
underscore.extend(m.dont_verify_list, common.dont_verify_list)

function m:sensor_input_type(p, fromxml)
    return m.translate({
            ['1'] = 'quarter',
            ['2'] = 'half',
            ['4'] = 'full',
        }, p, fromxml)
end

function m:excitation_mode(p, fromxml)
    return m.translate({
            ['1'] = 'dc',
            ['5'] = 'ac_4800_hz_sine',
        }, p, fromxml)
end

function m:excitation_range(p, fromxml)
    if fromxml then
        return string.format("%.1fv", tonumber(p))
    end
    return p:match('(.+)[V|v]$')
end

m.translator = {
    sensor_input_type = m.sensor_input_type,
    excitation_mode = m.excitation_mode,
    excitation_range = m.excitation_range
}

function m:to_internal(spec)
    local spec = spec
    -- Do some processing based on sensor input mode (if it exists)
    if spec.sensor_input and spec.sensor_input.mode then

        -- Pt sensors require a specific set of parameter to be set or not set
        -- so we take care of that here!
        if spec.sensor_input.mode:find('^pt') then
            if spec.sensor_input.mode == 'pt1000' then
                spec._pt_type = 3
            else
                spec._pt_type = 1
            end
            spec.calibration.rezero.units = "°C"
            spec._pt_wiring = 4
            spec.sensor_input.mode = 'pt'
            spec.calibration.sensor_scaling.mode = 'internal'
            spec.calibration.sensor_scaling.units = "°C"
            spec.calibration.define_source_units.units = "°C"

            spec.drift_compensation = nil
            spec.teds = nil
        end

        -- If this is a bridge, set the bridge wiring
        if (spec.sensor_input.mode == 'bridge' or
            spec.sensor_input.mode == 'inductive_bridge') and
        spec.excitation_range then
            -- If this is NOT a TEDS device do this stuff
            --will always branch because bindata is not bin_data. Left to ease testing concerns
            if not spec.teds_sensor or not spec.teds_sensor.bindata then
                local bridge_max_levels = {
                    ['2.5'] = 5,
                    ['1.0'] = 12.5,
                }
                local er = string.format("%.1f",
                    tonumber(spec.excitation_range:match('(.*)[V|v]') or 0))
                if not bridge_max_levels[er] then
                    return nil, string.format("Invalid excitation range '%s'",
                        spec.excitation_range)
                end
                spec.excitation_range = er
                spec.sensor_input.max = spec.sensor_input.max or
                    bridge_max_levels[er]
                if spec.sensor_input.type == 'full' and
                    tonumber(spec.sensor_input.wires) == 4 then
                    spec.sensor_input.wires = '6'
                elseif spec.sensor_input.type == 'half' and
                    tonumber(spec.sensor_input.wires) == 3 then
                    spec.sensor_input.wires = '5'
                end
            end
            spec._bridge_type =
                m.sensor_input_type(self, spec.sensor_input.type)
            spec.sensor_input.min = spec.sensor_input.max * -1
            spec._bridge_wiring = spec.sensor_input.wires
            spec.sensor_input.wires = nil
            spec.sensor_input.type = nil
        end

        -- If the sensor is a resistance sensor, make sure max is set
        if spec.sensor_input.mode == 'resistor' then
            spec.sensor_input.max = spec.sensor_input.max or 5000
        elseif spec.sensor_input.mode == 'potentiometer' then
            spec.sensor_input.resistance = spec.sensor_input.resistance or 500
            spec.calibration = spec.calibration or {}
            spec.calibration.define_source_units =
                spec.calibration.define_source_units or {
                    units = "mV/V",
                    applied = true,
                }
            spec.calibration.units_conversion =
                spec.calibration.units_conversion or {
                    units = "mV/V",
                    applied = true,
                }
        end
        
        if spec.sensor_input.mode == 'pulse_counter' then
            spec._s2enbl = '0'
            spec._f4enbl = '0'
            spec._indexenbl = '0'
            spec._indexdivisor = '1'
        elseif spec.sensor_input.mode == 'encoder_counter' then
            spec._s2enbl = '1'
            spec._f4enbl = spec.sensor_input.quadrature and '1' or '0'
            spec._indexenbl = tostring(spec.sensor_input.encoder_index_reset or 'false') == 'true' and '1' or '0'
            spec._indexdivisor = tostring(spec.sensor_input.encoder_index_divisor or '1')
        elseif spec.sensor_input.mode == 'pulse_frequency' then
            spec._s2enbl = '0'
            spec._f4enbl = '0'
        elseif spec.sensor_input.mode == 'encoder_frequency' then
            spec._s2enbl = '1'
            spec._f4enbl = spec.sensor_input.quadrature and '1' or '0'
        elseif spec.sensor_input.mode == 'ssi_encoder_counter' then
            if spec.sensor_input.ssi_coding == "gray" then
                spec._ssi_coding = 0
            else 
                spec._ssi_coding = 1
            end
            if spec.sensor_input.ssi_clock == 100 then
                spec._ssi_clock = 100000
            elseif spec.sensor_input.ssi_clock == 200 then
                spec._ssi_clock = 200000
            elseif spec.sensor_input.ssi_clock == 500 then
                spec._ssi_clock = 500000
            elseif spec.sensor_input.ssi_clock == 1000 then
                spec._ssi_clock = 1000000
            end
            spec._ssi_resolution = spec.sensor_input.ssi_bits
        end 
        
        if spec.sensor_input.mode == 'thermocouple' then
            spec.calibration = spec.calibration or {}
                spec.calibration.define_source_units = {
                purpose = 'define_source_units',
                applied = true,
                units = '°C',
            }
            spec.calibration.sensor_scaling = {
                purpose = 'sensor_scaling',
                mode = 'internal',
                units = '°C',
            }
            spec._thermocouple_type = 3
        end   
    end

    return spec
end

local set_shunt_xml_template =[[
<?xml version="1.0"?>
<SrvParams>
<Control>SHUNTCALIBRATION</Control>
<ControlParams>&lt;?xml version="1.0"
encoding="UTF-8"?&gt;&lt;ControlParams&gt;&lt;Par1&gt;#channel&lt;/Par1&gt;&lt;Par2&gt;#state&lt;/Par2&gt;&lt;/ControlParams&gt;</ControlParams>
</SrvParams>
]]

local zero_xml_template =[[
<?xml version="1.0"?>
<SrvParams>
<Control>SETZERO</Control>
<ControlParams>&lt;?xml version="1.0"
encoding="UTF-8"?&gt;&lt;ControlParams&gt;&lt;Par1&gt;#channel&lt;/Par1&gt;&lt;Par2&gt;&lt;/Par2&gt;&lt;Par3&gt;#target&lt;/Par3&gt;&lt;/ControlParams&gt;</ControlParams>
</SrvParams>
]]

m.controls = {}
m.controls.channel = {}
m.controls.device = {}

local shunt_directions = { off = 0, up_scale = 1, down_scale = 2 }

local control_shunt = function(control, chan, params, cb)
    if params and type(params) == 'table' and params.direction and
            shunt_directions[params.direction] then
        local xml = set_shunt_xml_template
        xml = string.gsub(xml, '#channel', chan)
        xml = string.gsub(xml, '#state', shunt_directions[params.direction])
        control:_send_transaction_request(C.TR_SCHEMA, C.TR_XML,
            function(obj, ecode, data, len)
                cb { result = 'success', msg = 'ok' }
            end, xml)
    else
        cb {
            result = 'error',
            msg = 'unsupported or missing shunt direction arg'
        }
    end
end

-- Function used to control shunts on Quantum devices that have them.
-- Note, self in this function is the ChannelProxy object
function m.controls.channel:shunt(params, cb)
    control_shunt(self.mxdev.mx.control, self.index, params, cb)
end

-- Device level shunt control that enables/disables shunts on all channels
function m.controls.device:shunt(params, cb)
    control_shunt(self.mx.control, 0, params, cb)
end

-- Function used to control channel zeroing.  This is really only used for
-- interactive zeroing of counter sensor types.
-- Note, self in this function is the ChannelProxy object
function m.controls.channel:zero(params, cb)
    local chan = self.index
    local target = params and params[1]
    local xml = zero_xml_template
    xml = string.gsub(xml, '#channel', chan)
    xml = string.gsub(xml, '#target', target or 0.0)
    self.mxdev.mx.control:_send_transaction_request(C.TR_SCHEMA, C.TR_XML,
        function(obj, ecode, data, len)
            cb{result = 'success', msg = 'ok'}
        end, xml)
end

function m:to_external(spec)
    local spec = spec
    spec.calibration.units_conversion = nil
    -- Don't return this stuff to the external entities
    spec.wiring = nil

    if spec.sensor_input.mode and
    spec.sensor_input.mode ~= 'bridge' and
    spec.sensor_input.mode ~= 'inductive_bridge' then
        spec.bridge_bridge_factor = nil
        spec.bridge_gage_factor = nil
        spec.bridge_resistance = nil
        spec.excitation_mode = nil
        spec.excitation_range = nil
    else
        spec.sensor_input.type =
            m.sensor_input_type(self, spec._bridge_type, true)
        spec.sensor_input.wires = spec._bridge_wiring
        spec.sensor_input.regulation = true
        if spec.calibration.sensor_scaling.mode == "strain_gage" and
            spec.teds_sensor then
            spec.calibration.units_conversion = {}
            spec.calibration.units_conversion.mode = "defined_slope_intercept"
            spec.calibration.units_conversion.slope = 1
            spec.calibration.units_conversion.intercept = 0
        end
    end

    spec._bridge_wiring = nil
    spec._bridge_type = nil

    if spec.sensor_input.mode and spec.sensor_input.mode:find('^pt') then
        spec.calibration = spec.calibration or {}
        spec.calibration.define_source_units = {
            purpose = 'define_source_units',
            units = spec.calibration.sensor_scaling.units or "°C",
        }

        spec.calibration.sensor_scaling = spec.calibration.sensor_scaling or {}
        spec.calibration.sensor_scaling.purpose = 'sensor_scaling'
--[[
        spec.calibration.sensor_scaling.mode = 'defined_slope_intercept'
        spec.calibration.sensor_scaling.slope = '1'
        spec.calibration.sensor_scaling.intercept = '0'
]]
        spec.calibration.sensor_scaling.mode = 'internal'
        spec.calibration.sensor_scaling.slope = nil
        spec.calibration.sensor_scaling.intercept = nil
        spec.calibration.sensor_scaling.units =
            spec.calibration.define_source_units.units
        spec.calibration.sensor_scaling.electrical = nil
        spec.calibration.sensor_scaling.physical = nil

        spec.sensor_input.wires = spec._pt_wiring or 4
        if spec.teds_sensor then
            spec.calibration.units_conversion = {}
            spec.calibration.units_conversion.mode = "defined_slope_intercept"
            spec.calibration.units_conversion.slope = 1
            spec.calibration.units_conversion.intercept = 0
        end
        if spec._pt_type and tonumber(spec._pt_type) and
            tonumber(spec._pt_type) == 3 then
            spec.sensor_input.mode = 'pt1000'
        else
            spec.sensor_input.mode = 'pt100'
        end
    end

    if spec.sensor_input.mode and (spec.sensor_input.mode == 'pulse_counter' or
            spec.sensor_input.mode == 'encoder_counter') then
        if spec._s2enbl == '0' then
            spec.sensor_input.mode = 'pulse_counter'
        else
            spec.sensor_input.mode = 'encoder_counter'
            spec.sensor_input.encoder_index_reset = spec._indexenbl == '1' and true or false
            if spec.sensor_input.encoder_index_reset then
                spec.sensor_input.encoder_index_divisor = tonumber(spec._indexdivisor or 1)
            end
        end
        if spec.sensor_input.mode == 'encoder_counter' then
            spec.sensor_input.quadrature = spec._f4enbl == '1'
        end
        spec._s2enbl = nil
        spec._f4enbl = nil
        spec._indexenbl = nil
        spec._indexdivisor = nil
        spec.sensor_input.min = nil
        spec.sensor_input.max = nil
    end

    if spec.sensor_input.mode and (spec.sensor_input.mode == 'pulse_frequency' or
            spec.sensor_input.mode == 'encoder_frequency') then
        if spec._s2enbl == '0' then
            spec.sensor_input.mode = 'pulse_frequency'
        else
            spec.sensor_input.mode = 'encoder_frequency'
        end
        if spec.sensor_input.mode == 'encoder_frequency' then
            spec.sensor_input.quadrature = spec._f4enbl == '1'
        end
        spec._s2enbl = nil
        spec._f4enbl = nil
        spec._indexenbl = nil
        spec._indexdivisor = nil
        spec.sensor_input.min = nil
        spec.sensor_input.max = nil
    end
    
    if spec.sensor_input.mode and spec.sensor_input.mode == 'ssi_encoder_counter' then
        if tonumber(spec._ssi_coding) == 0 then
            spec.sensor_input.ssi_coding = "gray"
        else
            spec.sensor_input.ssi_coding = "binary"
        end
        if tonumber(spec._ssi_clock) == 100000 then
            spec.sensor_input.ssi_clock = 100
        elseif tonumber(spec._ssi_clock) == 200000 then
            spec.sensor_input.ssi_clock = 200
        elseif tonumber(spec._ssi_clock) == 500000 then
            spec.sensor_input.ssi_clock = 500
        elseif tonumber(spec._ssi_clock) == 1000000 then
            spec.sensor_input.ssi_clock = 1000
        end
        spec.sensor_input.ssi_bits = tonumber(spec._ssi_resolution)
    end

    -- Redmine #4632: LVDT sensors on the 840 use fixed excitation settings of
    -- 4800 Hz AC and 1 volt. This is not configurable. However, we want the
    -- information to appear in the SXR file, so we insert those values here.
    if spec.sensor_input.mode == 'lvdt' then
        spec.excitation_mode = 'ac_4800_hz_sine'
        spec.excitation_range = '1.0v'
    end

    -- Dont supply transducer power for potentiometer sensor modes
    if spec.sensor_input.mode and spec.sensor_input.mode == 'potentiometer' then
        spec.transducer_power = nil
    end

    if spec.sensor_input.mode and spec.sensor_input.mode == 'thermocouple' then
        spec.sensor_input.type = 'k'
        spec.sensor_input.ThermocoupleType = nil
        spec.sensor_input.Resistance = nil
        spec.calibration = spec.calibration or {}
        spec.calibration.define_source_units = {
            purpose = 'define_source_units',
            units = '°C',
        }
        spec.calibration.sensor_scaling = {
            purpose = 'sensor_scaling',
            mode = 'internal',
            units = '°C',
        }  
        if spec.teds_sensor then
            spec.calibration.units_conversion = {}
            spec.calibration.units_conversion.mode = "defined_slope_intercept"
            spec.calibration.units_conversion.slope = 1
            spec.calibration.units_conversion.intercept = 0
        end
        spec.digital_filter_type = 'bessel_iir'
        spec.digital_filter_attenuation_frequency = '1'
    end

    spec._pt_wiring = nil
    spec._pt_type = nil

    return spec
end

-- Specification of the default settings of this channel type
m.default_spec = {
    sample_rate = '10',
    sensor_input = {
        mode = 'volt',
        min = '-10',
        max = '10.000000',
    },
    calibration = {
        define_source_units = {
            purpose = 'define_source_units',
            units = 'V',
        },
        sensor_scaling = {
            purpose = 'sensor_scaling',
            mode = 'defined_slope_intercept',
            slope = '1',
            intercept = '0',
            units = 'V',
        },
        units_conversion = {
            purpose = 'units_conversion',
            mode = 'defined_slope_intercept',
            units = 'V',
            intercept = '0',
            slope = '1',
        },
        rezero = {
            purpose = 'rezero',
            mode = 'defined_slope_intercept',
            intercept = '0.00000',
        },
    },
    transducer_power = '0.000000',
    digital_filter_type = 'none',
    sensor = {
        use_teds = 2,
    },
}

m.virtual_channels = { }
m.virtual_channels[1] = { name = 'can1', channel_num = 1 }

return m
