--! \file

require 'middleclass'
local u = require 'unix'
local ev = require 'ev'
local struct = require 'struct'
local ffi = require 'ffi'
local C = ffi.C

ffi.cdef[[
typedef int ssize_t;

ssize_t send(int sockfd, const void *buf, size_t len, int flags);
ssize_t recv(int sockfd, void *buf, size_t len, int flags);

enum {
    EAGAIN = 11,
};
]]

local MSG_DONTWAIT = 0x40

--! Various network utilities.
--? namespace net
local net = { }

--! A very low-level socket class
--? class net.Socket
net.Socket = { }
net.Socket.__index = net.Socket
function net.socket(opts)
    opts = opts or { }
    return setmetatable(opts, net.Socket)
end

function net.Socket:gai(passive)
    if self.addrs then return end
    local flags = passive and u.AI_PASSIVE or 0
    local host = self.host
    local service = self.service
    local family = self.family
    local type = self.type or u.SOCK_STREAM
    if self.ip then
        host = self.ip
        flags = flags + u.AI_NUMERICHOST
    end
    if host and not family then family = u.AF_UNSPEC end
    if not service and not self.port then self.port = 0 end
    if self.port then
        service = self.port
        flags = flags + u.AI_NUMERICSERV
    end
    if not family then family = u.AF_INET6 end
    self.addrs = assert(u.getaddrinfo(host, service, flags, family, type))
    assert(self.addrs[1])
end

function net.Socket:connect()
    self:gai(false)
    -- KLUDGE handle optional non-blocking connect
    for i,a in ipairs(self.addrs) do
        local sock = assert(u.socket(a.family, a.socktype, a.protocol))
        -- print('new connect', sock)
        if u.connect(sock, a.addr) then
            self.addr = a
            self.sock = sock
            return self
        end
        -- Don't leak the socket away if the connect fails
        u.close(sock)
    end
    error("Couldn't connect")
end

function net.Socket:bind()
    self:gai(true)
    self.addr = self.addrs[1]
    local a = self.addr
    local sock = assert(u.socket(a.family, a.socktype, a.protocol))
    -- print('new bind', sock)
    assert(u.setsockopt_int(sock, u.SOL_SOCKET, u.SO_REUSEADDR, 1))
    assert(u.bind(sock, a.addr))
    self.sock = sock
    self.ip, self.port = u.getnameinfo(u.getsockname(sock),
                                       u.NI_NUMERICHOST + u.NI_NUMERICSERV)
    return self
end

function net.Socket:listen(opts)
    opts = opts or { }
    if not self.sock then self:bind() end
    assert(u.listen(self.sock, opts.backlog or 16))
    self.acceptor = opts.acceptor
    local function handler(loop, w)
        local s, addr = u.accept(self.sock)
        -- print('new accept', s)
        if s then
            self:acceptor(loop, s, addr)
        end
    end
    self.listen_watcher = ev.IO.new(handler, self.sock, ev.READ)
    if self.loop then
        sock.listen_watcher:start(self.loop)
    end
    return self
end

function net.Socket:close()
    u.close(self.sock)
end

function net.Socket:start(loop)
    self.loop = loop
    if self.listen_watcher then
        self.listen_watcher:start(loop)
    end
end

function net.Socket:stop(loop)
    assert(loop == self.loop)
    if self.listen_watcher then
        self.listen_watcher:stop(loop)
        self:close()
    end
    self.loop = nil
end

--! A buffer abstraction for stream sockets
--? class net.StreamBuffer
net.StreamBuffer = class('net.StreamBuffer')

function net.streambuf(socket)
    return net.StreamBuffer:new(socket)
end

function net.StreamBuffer:initialize(socket)
    assert(socket)
    self.socket = socket
    self.blocksize = self.blocksize or 8192
    self.readbuf = ''
    self.read_wm = 1
    self.read_closing = false
    self.read_closed = false
    local function reader(loop, w)
        local got, estr, e = u.recv(self.socket, self.blocksize, MSG_DONTWAIT)
        local read_will_close = false
        if got == '' then
            self.reader:stop(loop)
            read_will_close = true
        end
        if not got then
            if e == u.EAGAIN then
                return
            end
            self.reader:stop(loop)
            read_will_close = true
        else
            self.readbuf = self.readbuf .. got
        end
        while self.read_cb and self.readbuf:len() >= self.read_wm do
            self:read_cb(loop)
        end
        if read_will_close then
            self.read_closed = true
            if self.close_cb then
                self:close_cb(loop, 'read', 'remote')
            end
            if self.write_closed then
                self:close()
            end
        end
    end
    self.reader = ev.IO.new(reader, self.socket, ev.READ)

    self.write_wm = 0
    self.write_closing = false
    self.write_closed = false
    self.writebuf = ''
    local function writer(loop, w)
        local sent = C.send(self.socket, self.writebuf,
            self.writebuf:len(), MSG_DONTWAIT)
        if sent <= 0 then
            if sent < 0 and ffi.errno() ~= C.EAGAIN then
                self.writer:stop(loop)
                self.writebuf = ''
                self.write_closed = true
                if self.read_closed then
                    self:close()
                end
            end
            return
        elseif sent then
            self.writebuf = self.writebuf:sub(sent + 1)
        end
        if self.write_closing then
            if self.writebuf:len() == 0 then
                self.writer:stop(loop)
                u.shutdown(self.socket, u.SHUT_WR)
                self.write_closed = true
                if self.read_closed then
                    self:close()
                end
                return
            end
        else
            while self.write_cb and self.writebuf:len() < self.write_wm do
                self:write_cb(loop)
            end
        end
        if self.writebuf:len() == 0 then
            self.writer:stop(loop)
        end
    end
    self.writer = ev.IO.new(writer, self.socket, ev.WRITE)
end

function net.StreamBuffer:read(size)
    assert(not size or size > 0)
    local got = self.readbuf:sub(1, size or -1)
    self.readbuf = self.readbuf:sub(got:len() + 1)
    if got == '' and self.read_closed then
        return nil
    end
    return got
end

function net.StreamBuffer:write(data)
    if self.write_closed then
        return nil
    end
    self.writebuf = self.writebuf .. data
    if self.loop then
        self.writer:start(self.loop)
    end
end

function net.StreamBuffer:close()
    if self.socket then
        -- print('closing', self.socket)
        assert(u.close(self.socket))
        if self.loop then self:stop(self.loop) end
        self.socket = nil
    end
end

function net.StreamBuffer:shutdown_write()
    if not self.write_closed then
        self.write_closing = true
        if self.writebuf:len() == 0 then
            u.shutdown(self.socket, u.SHUT_WR)
            self.write_closed = true
            if self.loop then
                self.writer:stop(self.loop)
            end
        end
    end
    if self.write_closed and self.read_closed then
        self:close()
    end
end

function net.StreamBuffer:shutdown()
    self:shutdown_write()
end

function net.StreamBuffer:start(loop)
    self.loop = loop
    if not self.read_closed then
        self.reader:start(loop)
    end
    if self.writebuf:len() > 0 and not self.write_closed then
        self.writer:start(loop)
    end
end

function net.StreamBuffer:stop(loop)
    assert(loop == self.loop)
    self.reader:stop(loop)
    self.writer:stop(loop)
    self.loop = nil
end

--! A buffer abstraction for message sockets
--? class net.MessageBuffer
net.MessageBuffer = class('net.MessageBuffer')

function net.MessageBuffer:initialize(sbuf)
    -- print('msgbuf init')
    assert(instanceOf(net.StreamBuffer, sbuf))
    self.sbuf = sbuf
    self.msgs = { }
    self.msg_start = nil
    self.msg_end = nil
    function self.read_len4(sbuf, loop)
        -- print('read_len4')
        local len = struct.unpack('>I4', sbuf.readbuf)
        self.msg_start = 5
        self.msg_end = len + 4
        sbuf.read_cb = self.read_data
        sbuf.read_wm = 4 + len
    end
    function self.read_data(sbuf, loop)
        -- print('read_data')
        local msg = sbuf.readbuf:sub(self.msg_start, self.msg_end)
        sbuf.readbuf = sbuf.readbuf:sub(self.msg_end + 1)
        sbuf.read_cb = self.read_len4
        sbuf.read_wm = 4
        self.msg_start = nil
        self.msg_end = nil
        table.insert(self.msgs, msg)
        while self.msg_cb and #self.msgs > 0 do
            -- print('#self.msgs', #self.msgs)
            self:msg_cb(loop)
        end
    end
    sbuf.read_cb = self.read_len4
    sbuf.read_wm = 4
    self.msg_cb = nil
end

function net.MessageBuffer:recvmsg()
    return table.remove(self.msgs, 1)
end

function net.packmsg(msg)
    return struct.pack('>I4c0', msg:len(), msg)
end

function net.MessageBuffer:sendmsg(msg)
    return self.sbuf:write(net.packmsg(msg))
end

-- ---------------------------------------------------------------

--! Normalize an IPv6 address.
function net.normalize_ipv6(addr)
    local r =
        assert(u.getaddrinfo(addr, 0, u.AI_NUMERICHOST + u.AI_NUMERICSERV,
                             u.AF_INET6, u.SOCK_STREAM))[1]
    local ip =
        assert(u.getnameinfo(r.addr, u.NI_NUMERICHOST + u.NI_NUMERICSERV))
    return ip
end

-- local sb = net.streambuf(s)
-- function sb:read_size()
--     local got = self:read(4)
--     self.size = struct.unpack('>I4', got)
--     self:read_watermark(self.size)
--     self:read_callback(self.read_data)
-- end
-- function sb:read_data()
--     local got = self:read(self.size)
--     -- process data
--     self:write(output)
--     self:read_watermark(4)
--     self:read_callback(self.read_size)
-- end
-- sb:read_callback(sb.read_size)
-- sb:read_watermark(4)
-- sb:start(loop)

-- local s = net.socket{host = 'www.google.com', port = 80}:connect()
-- local sbuf = net.streambuf(s.sock)
-- function sbuf:read_cb(loop)
--     local got = self:read()
--     if got then
--         print('got', got:len())
--     else
--         self:shutdown()
--     end
-- end
-- sbuf:write('GET / HTTP/1.0\r\n')
-- sbuf:write('Host: www.google.com\r\n')
-- sbuf:write('\r\n')
-- sbuf:start(ev.Loop.default)

-- local ls = net.socket{port = 12345}:listen()
-- function ls:acceptor(loop, sock)
--     local sbuf = net.streambuf(sock)
--     function sbuf:read_cb(loop)
--         local got = self:read()
--         if got then
--             self:write(got)
--         else
--             self:shutdown()
--         end
--     end
--     sbuf:start(loop)
-- end
-- ls:start(ev.Loop.default)

-- local ls = net.socket{port = 12346}:listen()
-- function ls:acceptor(loop, sock)
--     local sbuf = net.streambuf(sock)
--     sbuf:start(loop)
--     sbuf:shutdown_read()
--     for i=1,1000 do
--         sbuf:write(string.format('i = %f, i * i = %f\n', i, i * i))
--     end
--     sbuf:shutdown()
-- end
-- ls:start(ev.Loop.default)

-- ev.Loop.default:loop()

return net
