-- ring_buffers package to interact with the FPGA soft processor
--
-- Uses the connctrl package
--
-- This package provides an API to set up and maintain buffer descriptor rings.
-- One API function, make_new_bd_ring, creates a ring buffer using a provided
-- C data structure type and the number of buffers needed.  This ring buffer
-- can then be used for input data (inbound functions) or output data (outbound
-- functions).  For inbound data, the FPGA soft processor writes data to the
-- buffer, closes out the buffer descriptor, and signals via an interrupt that
-- a new buffer is available.  The API function 'get_buffer' is used to get a
-- copy of the buffer as a C data structure.  The caller is then free to use
-- the C data structure and dispose of it when finished without effecting the
-- operation of the ring buffer.  On the outbound side the API function
-- 'put_buffer' is called with a C data structure as its parameter.  The data
-- is copied out of this data structure and into the ring buffer for eventual
-- processing by the FPGA soft processor when it notices the updated buffer
-- descriptor.
--
-- API Functions:
--   make_new_bd_ring(devname, buffer_size, buffer_count)
--   get_initial_pa()
--   dump()
--   get_buffer()
--   put_buffer(buffer, size_override)
--
-- Note that 'make_new_bd_ring' returns a bd_ring object (a table) that is
-- objectfied so that the various API functions can be called through the
-- bd_ring object, e.g. bd_ring:get_buffer().
--
-- Generic buffer descriptor (BD) format (16 bytes):
--   BD Status/Bytes Written
--   Next BD Physical Address
--   BD Size
--   Buffer Physical Address
--   Next BD Virtual Address
--   Buffer Virtual Address
--
-- The BD is read in as needed and when the BD is closed out the status
-- is written and the Next BD Address is read.  If the next BD is not free,
-- processing stops until the buffer is freed.  If data is lost, an interrupt
-- is generated.
--
-- Status Codes
--     0 - Buffer Free and Available
--     1 - Buffer in use, being written to
--     2 - Buffer full, good status
--     3 - Buffer full, closed due to heartbeat, good status
--


local u = require 'unix'
local cc = require 'connctrl'
local ffi = require 'ffi'
local config = require 'system_config'

local PAGE_SIZE = 4096

local ring_buffers = {}

ring_buffers.PAGE_SIZE = PAGE_SIZE

ffi.cdef[[
typedef struct bd {
    volatile uint16_t   status;
    volatile uint16_t   bytes_written;
    uint32_t            next_bd_physaddr;
    uint32_t            bufsize;
    uint32_t            buffer_physaddr;
    union {
        uint32_t        virtaddr;
        struct bd       *ptr;
    } next;
    union {
        uint32_t        virtaddr;
        void            *ptr;
    } buffer;
    uint32_t            padding[2];
} bd_t;
typedef bd_t * bd_ptr_t;
]]

function ring_buffers.make_new_bd_ring(devname, buffer_ct,
                                       element_ct, buffer_count, offset)
    local bd_ring = {}
    local buffer_size = ffi.sizeof(buffer_ct)

    -- Reduce size if buffer would span a page; DMA hardware can't handle this
    if config.device_type ~= 'edaqxr' and buffer_size > PAGE_SIZE then
        local els = ffi.sizeof(buffer_ct) / ffi.sizeof(element_ct)
        local total_els = els * buffer_count
        els = math.floor(PAGE_SIZE / ffi.sizeof(element_ct))
        buffer_ct = string.format('%s[%d]', element_ct, els)
        buffer_size = ffi.sizeof(buffer_ct)
        -- Increase buffer_count to make overall ring buffer size the same
        buffer_count = math.ceil(total_els / els)
    end

    bd_ring.devname = devname
    bd_ring.buffer_ct = buffer_ct
    bd_ring.buffer_ct_size = ffi.sizeof(buffer_ct)
    bd_ring.element_ct = element_ct
    bd_ring.element_ct_size = ffi.sizeof(element_ct)
    bd_ring.element_ptr_ct = ffi.typeof(element_ct .. '*')
    bd_ring.buffer_size = buffer_size
    bd_ring.buffer_count = buffer_count
    bd_ring.bufdata = ffi.new(bd_ring.buffer_ct)

    local bd_size = ffi.sizeof("bd_t")
    local bytes_for_bds = bd_size * buffer_count
    local bytes_for_buffers = buffer_size * buffer_count
    local bd_pad = PAGE_SIZE - (bytes_for_bds % PAGE_SIZE)
    if bd_pad == PAGE_SIZE then bd_pad = 0 end
    local buffers_after_bds = math.floor(bd_pad / buffer_size)
    local buffers_per_page = math.floor(PAGE_SIZE / buffer_size)
    local pages_needed = math.ceil(bytes_for_bds / PAGE_SIZE)
    if buffer_count > buffers_after_bds then
        pages_needed = pages_needed +
            math.ceil((buffer_count - buffers_after_bds) / buffers_per_page)
    end
    local total_bytes_needed = pages_needed * PAGE_SIZE
    local dma_map, pa, va =
        cc.get_dma_mapping(devname, total_bytes_needed, offset)
    if config.device_type ~= 'edaqxr' then
        pa = cc.get_dma_mapping_physical_address(dma_map)
        va = cc.get_dma_mapping_virtual_address(dma_map)
    end
    bd_ring.dma_map = dma_map
    bd_ring.dma_map_pa = pa
    bd_ring.dma_map_va = va
    bd_ring.total_size = total_bytes_needed

    -- Now create a ring of buffer descriptors
    local ptr = ffi.cast("bd_ptr_t", va)
    local buffer_offset = bytes_for_bds
    for i=0,buffer_count-1 do
        -- if buffer_offset crosses a page boundary, adjust it to the start of
        -- the next page
        local start_page = math.floor(buffer_offset / PAGE_SIZE)
        local end_page = math.floor((buffer_offset + buffer_size - 1) /
                                    PAGE_SIZE)
        if start_page ~= end_page then
            buffer_offset = end_page * PAGE_SIZE
        end

        ptr.status = 0   -- Empty
        ptr.bytes_written = 0
        if i ~= buffer_count - 1 then
            ptr.next.virtaddr = va + ((i + 1) * bd_size)
            ptr.next_bd_physaddr = pa + ((i + 1) * bd_size)
        else
            ptr.next.virtaddr = va
            ptr.next_bd_physaddr = pa
        end
        ptr.bufsize = buffer_size
        ptr.buffer.virtaddr = va + buffer_offset
        ptr.buffer_physaddr = pa + buffer_offset
        buffer_offset = buffer_offset + buffer_size
        ptr = ptr.next.ptr
    end

    -- Initialize the read/write pointer to first bd, i.e. ptr
    bd_ring.ptr = ptr

    cc.flush_entire_dma_region(bd_ring.dma_map)

    setmetatable(bd_ring, {__index = ring_buffers})

    return bd_ring
end

function ring_buffers:get_initial_pa()
    return self.dma_map_pa
end

function ring_buffers:get_buffer()
    cc.flush_dma_region_by_ptr(self.dma_map, self.ptr, ffi.sizeof("bd_t"))
    if self.ptr.status > 1 then
        local bytes_written = self.ptr.bytes_written
        if bytes_written > self.buffer_ct_size then
            bytes_written = self.buffer_ct_size
        end
        local element_count = bytes_written / self.element_ct_size
        local heartbeat = self.ptr.status == 3
        local bufdata = self.bufdata
        cc.flush_dma_region_by_ptr(self.dma_map,
                                   self.ptr.buffer.ptr,
                                   bytes_written)
        ffi.copy(bufdata, self.ptr.buffer.ptr, bytes_written)
        self.ptr.bytes_written = 0
        self.ptr.status = 0
        cc.flush_dma_region_by_ptr(self.dma_map, self.ptr, ffi.sizeof("bd_t"))
        self.ptr = self.ptr.next.ptr
        return bufdata, bytes_written, element_count, heartbeat
    else
        return nil  -- No buffer available
    end
end

function ring_buffers:get_buffer_ptr()
    cc.flush_dma_region_by_ptr(self.dma_map, self.ptr, ffi.sizeof("bd_t"))
    if self.ptr.status <= 1 then
        return nil  -- No buffer available
    end
    local bytes_written = self.ptr.bytes_written
    local element_count = self.ptr.bytes_written / self.element_ct_size
    local heartbeat = self.ptr.status == 3
    local bufdata = ffi.cast(self.element_ptr_ct, self.ptr.buffer.ptr)
    cc.flush_dma_region_by_ptr(self.dma_map, self.ptr.buffer.ptr, bytes_written)
    return bufdata, element_count, heartbeat, bytes_written
end

function ring_buffers:free_buffer_ptr()
    self.ptr.bytes_written = 0
    self.ptr.status = 0
    cc.flush_dma_region_by_ptr(self.dma_map, self.ptr, ffi.sizeof("bd_t"))
    self.ptr = self.ptr.next.ptr
end

function ring_buffers:put_buffer(buffer, size_override)
    cc.flush_dma_region_by_ptr(self.dma_map, self.ptr, ffi.sizeof("bd_t"))
    if self.ptr.status == 0 then
        self.ptr.status = 1
        self.ptr.bytes_written = size_override or ffi.sizeof(buffer)
        ffi.copy(self.ptr.buffer.ptr, buffer, self.ptr.bytes_written)
        cc.flush_dma_region_by_ptr(self.dma_map,
                                   self.ptr.buffer.ptr,
                                   self.ptr.bytes_written)
        self.ptr.status = 2
        cc.flush_dma_region_by_ptr(self.dma_map, self.ptr, ffi.sizeof("bd_t"))
        self.ptr = self.ptr.next.ptr
        return true
    end
    return false
end

function ring_buffers:destroy()
    cc.release_dma_mapping(self.dma_map)
end

function ring_buffers:dump(name)
    print(string.format("Ring: %s", name))
    print(string.format("  Device %s", self.devname))
    print(string.format("  Count %d", self.buffer_count))
    print(string.format("  Size %d", self.buffer_size))
    print(string.format("  Total Size %d", self.total_size))
    print(string.format("  Phys Addr %08X", self.dma_map_pa))
    print(string.format("  Virt Addr %08X", self.dma_map_va))
    print(string.format("  Buffers"))
    print(string.format("    State Bytes   Virtual    Physical"..
                        "  Next BD VA  Next BD PA"))
    local ptr = self.ptr
    for j=1,self.buffer_count do
        print(string.format("     %4d  %4d  %08X    %08X    %08X    %08X",
                            ptr.status, ptr.bytes_written, ptr.buffer.virtaddr,
                            ptr.buffer_physaddr, ptr.next.virtaddr,
                            ptr.next_bd_physaddr))
        ptr = ptr.next.ptr
    end
    print("")
end

return ring_buffers

