--! \file

local u = require "unix"
local ev = require "ev"
local net = require "net"
local class = require "class"
local struct = require "struct"

do
    local sec, usec = u.clock_gettime()
    math.randomseed(sec * 1e9 + usec)
end

--! A TCP server.
--? class server
local m = class()
--? alias m server

function m:init(settings)
    --self.super().init(self, settings)
    if not self.ev_loop then
        self.ev_loop = ev.Loop.default
    end
    self.watchers = {}
    self.add_to_watchers = self.add_to_watchers or {}
end

local function add_streambuf(self, w)
    w.streambuf = net.StreamBuffer:new(w.socket)
    local streambuf = w.streambuf
    streambuf.loop = self.ev_loop -- KLUDGE, enables unintended behavior
    w.write = function(...) return streambuf:write(...) end
    w.read = function(...) return streambuf:read(...) end
end

local function add_to_watcher(self, w)
    for k, v in pairs(self.add_to_watchers) do
        w[k] = v
    end
end

function m:loop()
    return self.ev_loop
end

function m:shutdown_server()
    for socket, w in pairs(self.watchers) do
        self.close(self.ev_loop, w)
    end
end

function m.close(loop, w)
    u.shutdown(w.socket, u.SHUT_RDWR)
    u.close(w.socket)
    w:stop(loop)
    -- KLUDGE This should be redone in a nicer way - w:cleanup(function(w)...)
    if w.cleanup then
        w.cleanup(w)
    end
    local cl = w._cleanups
    if cl and cl.temp_files then
        for name, _ in pairs(cl.temp_files) do
            u.unlink(name)
        end
        cl.temp_files = nil
    end
    w.server.watchers[w.socket] = nil
end

function m.reader(loop, w, revents)
    local got = u.read(w.socket, 8192)
    if not got or got == "" then
        w.server.close(loop, w)
        return
    end

    w.buf = w.buf .. got
    return got
end

function m.acceptor(loop, listen_w, revents)
    local s, addr = u.accept(listen_w.socket)
    if s then
        assert(u.setfl(s, u.O_NONBLOCK))
        local reader = listen_w.reader
        local w = ev.IO.new(reader, s, ev.READ)
        local ip, port = assert(u.getnameinfo(addr,
            u.NI_NUMERICHOST + u.NI_NUMERICSERV))
        -- print(string.format("Accept from [%s]:%s", ip, port))
        w.server = listen_w.server
        w.socket, w.ip, w.port = s, ip, port
        w.buf = ""
        w.reader = reader
        add_streambuf(w.server, w)
        add_to_watcher(w.server, w)
        w:start(loop)
        w.server.watchers[w.socket] = w
        return w
    end
end

function m:_listen_local(args)
    args.family = u.AF_UNIX
    args.acceptor = args.acceptor or self.acceptor
    args.reader = args.reader or self.reader

    local ls = assert(u.socket(u.AF_UNIX, u.SOCK_STREAM, 0))

    local r = u.new_sockaddr_un(args.path)

    -- Make sure the file is gone if it had been created already
    u.unlink(args.path)

    assert(u.bind(ls, r))
    assert(u.listen(ls, 10))

    local w = ev.IO.new(args.acceptor, ls, ev.READ)
    w.server = self
    w.socket = ls
    w.family = args.family
    w.acceptor = args.acceptor
    w.reader = args.reader
    w.path = args.path
    add_to_watcher(self, {server=self, socket=ls}) -- KLUDGE more?
    w:start(self.ev_loop)
    self.watchers[w.socket] = w
    return w
end

function m:listen(args)
    args = args or {}
    if args.path then return self:_listen_local(args) end

    args.address = args.address or "::1"
    args.port = args.port or 0
    args.family = args.family or u.AF_UNSPEC
    args.acceptor = args.acceptor or self.acceptor
    args.reader = args.reader or self.reader

    local res = assert(u.getaddrinfo(args.address, args.port, u.AI_PASSIVE,
        args.family, u.SOCK_STREAM))
    local r = assert(res[1])

    local ip = assert(u.getnameinfo(r.addr,
        u.NI_NUMERICHOST + u.NI_NUMERICSERV))
    -- print("Got", ip, r.family, r.socktype, r.protocol)
    local ls = assert(u.socket(r.family, r.socktype, r.protocol))
    assert(u.setsockopt_int(ls, u.SOL_SOCKET, u.SO_REUSEADDR, 1))
    assert(u.bind(ls, r.addr))
    assert(u.listen(ls, 10))

    local w = ev.IO.new(args.acceptor, ls, ev.READ)
    w.server = self
    w.socket = ls
    w.family = args.family
    w.acceptor = args.acceptor
    w.reader = args.reader
    w.ip, w.port
        = u.getnameinfo(u.getsockname(ls), u.NI_NUMERICHOST + u.NI_NUMERICSERV)
    add_to_watcher(self, {server=self, socket=ls}) -- KLUDGE more?
    w:start(self.ev_loop)
    self.watchers[w.socket] = w
    return w
end

function m:connect_nb(address, port, reader, connect_cb)
    local res = assert(u.getaddrinfo(address, port,
        u.AI_ADDRCONFIG, u.AF_UNSPEC, u.SOCK_STREAM))
    local r = assert(res[1])

    local s = assert(u.socket(r.family, r.socktype, r.protocol))
    assert(u.setfl(s, u.O_NONBLOCK))

    local ch = ev.IO.new(function(l, i, e)
            i:stop(l)
            local v, es, ec = u.getsockopt(s, u.SOL_SOCKET, u.SO_ERROR, 4)
            local w
            if v and struct.unpack('>i4', v) == 0 then
                w = ev.IO.new(reader or self.reader, s, ev.READ)
                w.ip, w.port = u.getnameinfo(
                    u.getsockname(s), u.NI_NUMERICHOST + u.NI_NUMERICSERV)
                w.server = self
                w.socket = s
                w.buf = ""
                add_streambuf(self, w)
                add_to_watcher(self, w)
                w:start(self.ev_loop)
                self.watchers[w.socket] = w
            else
                u.close(s)
            end
            connect_cb(self.ev_loop, w)
        end, s, ev.WRITE)

    local status, estr, ecode = u.connect(s, r.addr)
    if not status and ecode ~= u.EINPROGRESS then
        u.close(s)
        return
    end
    ch:start(self.ev_loop)
    return s
end

function m:connect(address, port, reader)
    local res = assert(u.getaddrinfo(address, port,
        u.AI_ADDRCONFIG, u.AF_UNSPEC, u.SOCK_STREAM))
    local r = assert(res[1])

    local s = assert(u.socket(r.family, r.socktype, r.protocol))
    u.connect(s, r.addr)

    -- After the connect completes, make the socket non-blocking
    assert(u.setfl(s, u.O_NONBLOCK))

    local w = ev.IO.new(reader or self.reader, s, ev.READ)
    w.ip, w.port =
        u.getnameinfo(u.getsockname(s), u.NI_NUMERICHOST + u.NI_NUMERICSERV)
    w.server = self
    w.socket = s
    w.buf = ""
    add_streambuf(self, w)
    add_to_watcher(self, w)
    w:start(self.ev_loop)
    self.watchers[w.socket] = w
    return w
end

--[[
function m:temp_file(w, temp_prefix, suffix, keep)
    temp_prefix = temp_prefix or '/tmp/srvtmp_'
    for i = 1, 100 do
        local fname = string.format('%s%.08x%s', temp_prefix,
            math.random(0, 0xFFFFFFFF), suffix or '')
        local fd = u.open(fname,
            u.O_CREAT + u.O_EXCL + u.O_LARGEFILE + u.O_RDWR, 0x01B0)
        if fd then
            if not keep then
                w._cleanups = w._cleanups or {}
                w._cleanups.temp_files = w._cleanups.temp_files or {}
                w._cleanups.temp_files[fname] = true
            end
            -- Returned fd MUST be closed by calling code
            return fname, fd
        end
    end
    return nil
end
]]

local ffi = require 'ffi'
local C = ffi.C

ffi.cdef[[
    int mkostemp(char *template, int flags);
    int mkostemps(char *template, int suffixlen, int flags);
    int asprintf(char **strp, const char *fmt, ...);
    int free(void *ptr);
]]

function m:temp_file(w, temp_prefix, suffix, keep)
    temp_prefix = temp_prefix or '/tmp/svrtmp_'
    temp_prefix = temp_prefix .. 'XXXXXX'

    local fname = ffi.new('char *[1]')
    C.asprintf(fname, "%s%s", temp_prefix, suffix or '')

    local fd
    if suffix and suffix:len() > 0 then
        fd = C.mkostemps(fname[0], suffix:len(),
            u.O_CREAT + u.O_EXCL + u.O_LARGEFILE + u.O_RDWR)
    else
        fd = C.mkostemp(fname[0],
            u.O_CREAT + u.O_EXCL + u.O_LARGEFILE + u.O_RDWR)
    end
    if fd < 0 then
        C.free(fname[0])
        return nil
    end
    local name = ffi.string(fname[0])
    C.free(fname[0])
    if not keep then
        w._cleanups = w._cleanups or {}
        w._cleanups.temp_files = w._cleanups.temp_files or {}
        w._cleanups.temp_files[name] = true
    end
    return name, fd
end

return m
