local cjson         = require 'cjson.safe'
local lfs           = require 'lfs'
local config        = require 'system_config'
local uuid          = require 'uuid'
local log           = require 'log'

local m = {}

local path = config.storage.path

local function get_file_info(name)
    local fh = io.open(path..'/'..name:gsub('%.sie$', '%.meta'), 'r')
    if not fh then return end
    local v = fh:read('*a')
    fh:close()
    return cjson.decode(v)
end

-- Get the list of SIE files.  This is basically a mapping between the SIE file
-- name as stored on the storage pack and the "friendly" name assigned by the
-- user.
function m.get_sie_file_list()
    local result = {}
    for name in lfs.dir(path) do
        if name:find("%.sie$") then
            local c = get_file_info(name)
            local stat = lfs.attributes(path .. '/' .. name)
            result[#result + 1] = {
                name = c.friendly_name or name:match('^(.+)%.sie'), 
                size = stat.size,
                access_time = os.date('!%FT%H:%M:%S', stat.access),
                change_time = os.date('!%FT%H:%M:%S', stat.change),
                modification_time = os.date('!%FT%H:%M:%S', stat.modification),
                author = c.author or "Unknown",
                summary = c.summary or "Unknown"
            }
        end
    end
    return result
end

-- This function returns the path to the actual SIE file associated with the
-- "friendly" name given.
function m.get_sie_path_on_disk(name)
    local fh, d, j
    for n in lfs.dir(path) do
        if n:find('%.meta') then
            fh = io.open(path .. '/' .. n, 'r')
            if not fh then break end
            d = fh:read('*a')
            fh:close()
            j = cjson.decode(d)
            if j and j.friendly_name and j.friendly_name == name then
                return path .. '/' .. n:gsub('%.meta', '%.sie')
            end
        end
    end
end

-- This function removes the association between a "friendly" name and an SIE
-- file on the disk.
function m.delete_sie_friendly_name(name)
    local path_on_disk = m.get_sie_path_on_disk(name)
    if not path_on_disk then return end
    local txt_file = path_on_disk:gsub('%.sie', '%.meta')
    os.remove(txt_file)
    return path_on_disk
end

-- This function removes the association between a "friendly" name and an SIE
-- file on the disk AND removes the SIE from the disk.
function m.delete_sie_file(name)
    local path_on_disk = m.delete_sie_friendly_name(name:gsub('%.sie', ''))
    if path_on_disk then
        os.remove(path_on_disk)
        return true
    end
end

-- This function removes a list of SIE files
function m.delete_sie_files(list)
    -- Get a list of friendly name to hash name mappings
    local mappings = {}
    for file in lfs.dir(path) do
        if file:find('%.meta$') then
            local fh = io.open(path .. '/' .. file, 'r')
            if fh then
                local d = fh:read('*a')
                fh:close()
                if d then
                    local j = cjson.decode(d)
                    if j and j.friendly_name then
                        mappings[j.friendly_name] = file:match('^(.+)%.meta$')
                    end
                end
            end
        end
    end

    -- Remove all of the request files
    local removed, nonremoved = {}, {}
    for _, f in ipairs(list) do
        f = f:gsub("%.sie", "")
        if mappings[f] then
            removed[#removed + 1] = f .. '.sie'
            f = path .. '/' .. mappings[f]
            os.remove(f .. '.sie')
            os.remove(f .. '.meta')
        else
            nonremoved[#nonremoved + 1] = f .. '.sie'
        end
    end

    -- Return list of what and wasn't removed
    return removed, nonremoved
end

local function get_friendly_name_list(filter)
    local list = {}
    local fh
    for file in lfs.dir(path) do
        if file:find('%.meta') then
            fh = io.open(path .. '/' .. file)
            if fh then
                local d = fh:read('*a')
                fh:close()
                if d then
                    local j = cjson.decode(d)
                    if j and j.friendly_name and (not filter or
                        j.friendly_name:find(filter, 1, true)) then
                        list[j.friendly_name] = j
                    end
                end
            end
        end
    end
    return list
end

-- This function associates a "friendly" name with an SIE file on the disk.
-- If the friendly name already exists, a new one will be generated.  The name
-- used will be returned.
function m.assign_friendly_sie_name(requested_name,
        path_on_disk, author, summary, naming_scheme)

    -- SIE file naming schemes can be selected by the user; see #5582 for
    -- initial implementation details.
    local name

    -- Is "exact" naming requested (throw an error if name already exists)?
    if naming_scheme == 'exact' then

        local results = get_friendly_name_list()

        if results[requested_name] then return nil end

        name = requested_name

    -- Is date/timestamp based naming selected?
    elseif naming_scheme == 'utc_date_time' then

        -- Generate stamped filename like "name-YYYY-mm-ddTHH-MM-SS"
        name = requested_name .. '-' ..  os.date('%Y-%m-%dT%H-%M-%S')

    -- The default naming scheme is 'numeric_increment'. Here we
    -- generate a name of the form "name-NNNN" where NNNN is one
    -- greater than any other names of the same form (or 0001 if
    -- there are no other names of the same form on the system).
    else

        local results = get_friendly_name_list(requested_name)

        local highest_used_number = 0

        for _, result in pairs(results) do
            -- Is this friendly name of the form <requested_name>-%d+?
            local prefix, suffix = result.friendly_name:match('^(.+)-(%d+)$')
            if prefix and suffix and prefix == requested_name then
                local n = tonumber(suffix)
                if n > highest_used_number then highest_used_number = n end
            end
        end

        name = string.format('%s-%03d',
            requested_name, highest_used_number + 1)

    end

    -- Add the name mapping
    local info = {
        friendly_name = name,
        author = author,
        summary = summary
    }
    local f = io.open(path_on_disk:gsub('%.sie$', '%.meta'), 'w')
    if f then
        f:write(cjson.encode(info))
        f:close()
    end

    log:log(log.NOTICE,
        string.format("Selected 'friendly' SIE name: '%s'", name))

    return name

end

-- This function changes the "friendly" name of an SIE file and returns the new
-- friendly name.
function m.rename_sie(requested_name, current_name)

    local path_on_disk = m.get_sie_path_on_disk(current_name)
    if not path_on_disk then
        log:log(log.ERR, "Can't find SIE '" .. current_name .. "' on disk")
        return nil
    end

    local params = m.get_sie_parameters_by_name(current_name)

    local new_path_on_disk =
        path_on_disk:match('^(/.+/)') .. uuid.generate() .. '.sie'
    if not new_path_on_disk then
        log:log(log.ERR, "Can't generate new disk name for rename")
        return nil
    end

    local new_name = m.assign_friendly_sie_name(requested_name,
        new_path_on_disk, params.author, params.summary, 'exact')
    if not new_name then return nil end

    m.delete_sie_friendly_name(current_name)
    os.rename(path_on_disk, new_path_on_disk)

    return requested_name
end

-- This function is used to clean out friendly name to actual file associations
-- from the database for friendly names whos actual file no longer exists.
function m.clean_sie()
    for file in lfs.dir(path) do
        if file:match('%.meta$') then
            local sie_path = path .. '/' ..  file:gsub('%.meta', '%.sie')
            if not lfs.attributes(sie_path) then
                os.remove(path .. '/' .. file)
            end
        end
    end
    return {}
end

-- This function is used to get the "friendly" name associated with a
-- particular SIE file.
function m.get_sie_parameters_by_hash(hash_name)
    hash_name = hash_name:gsub('%.sie$', '')
    local fh = io.open(path .. '/' .. hash_name .. '.meta', 'r')
    if not fh then return end
    local d = fh:read('*a')
    fh:close()
    return cjson.decode(d)
end

-- This function is used to get the friendly parameters for a
-- particular friendly named SIE file.
function m.get_sie_parameters_by_name(name)
    local fh, d, j
    for n in lfs.dir(path) do
        if n:find('%.meta') then
            fh = io.open(path .. '/' .. n, 'r')
            if not fh then break end
            d = fh:read('*a')
            fh:close()
            j = cjson.decode(d)
            if j and j.friendly_name and j.friendly_name == name then
                return j
            end
        end
    end
end

-- Function used to check the SIE hash name to friendly name mappings
function m.check_sie_files()
    -- Clean out all friendly name mapping files for which the associated SIE
    -- no longer exists
    m.clean_sie()

    -- Generate friendly name to hash name metadata file for any SIE files that
    -- do not have them.
    for fh in lfs.dir(path) do
        -- Is this an SIE file ?
        if fh:match('%.sie$') then
            -- Yes, get the existing hash to friendly mapping
            local params = m.get_sie_parameters_by_hash(fh)
            local fn = params and params.friendly_name

            if not fn then
                -- No mapping of friendly name to hash name, attempt
                -- to use the test name in the SIE file
                -- KLUDGE:  This should probably be done by actually reading
                -- and parsing the XML for the information.
                fn = fh:match('(.+)%.sie$')
                local author, summary, file
                local f = io.popen('grep -m 3 -e \'"test_name": "\' ' ..
                    '-e \'"test_author": "\' -e \'"test_summary": "\' ' ..
                    path .. '/' .. fh)
                if f then
                    for l in f:lines() do
                        file = file or l:match('"test_name": "(.+)",')
                        author = author or l:match('"test_author": "(.+)",')
                        summary = summary or l:match('"test_summary": "(.+)",')
                    end
                    f:close()
                end

                -- Add the mapping to the database
                m.assign_friendly_sie_name(file or fn, path .. '/' .. fh,
                    author, summary)
            end
        end
    end
end

return m
