require 'strict'

-- function_generator.lua - Implements the Function generator computed channel.

local dbg = require 'dbg'
local ffi = require 'ffi'
local rational = require 'rational'
local frame = require 'test_engine/frame'
local pipe = require 'test_engine/pipe'
local utils = require 'test_engine/test_eng_utils'

local ffi_cast = ffi.cast
local ffi_sizeof = ffi.sizeof

local Frame_create = frame.create
local Frame_clone = frame.clone
local Frame_delete = frame.delete
local Pipe_attach_pipe = pipe.attach_pipe
local Pipe_get_frame = pipe.get_frame
local Pipe_has_frames = pipe.has_frames
local Pipe_new_pipe_list = pipe.new_pipe_list

local mt = {}

function mt:dump()
    dbg('tot_count=%lld, scale_factor=%lld',
        self.tot_count, self.scale_factor)
end

function sine(x, period_samples, scale, offset, drift, decay)
    scale = (decay^(x/period_samples)) * scale 
    offset = drift * x/period_samples + offset
    local result = scale * math.sin(2*math.pi*x/period_samples) + offset
    return result
end

function square(x, period_samples, scale, offset, drift, decay, duty_cycle)
    local result
    scale = (decay^(x/period_samples)) * scale 
    offset = drift * x/period_samples + offset
    if x%period_samples <= period_samples*duty_cycle and x%period_samples ~= 0 or x == 0 then
        result = -1 * scale + offset
    else
        result = scale + offset
    end
    return result
end

function ramp(x, period_samples, scale, offset, drift, decay, duty_cycle)
    local result
    scale = (decay^(x/period_samples)) * scale 
    offset = drift * x/period_samples + offset
    local n = x
    x = x%period_samples
    if (x < period_samples*duty_cycle) then
        result = ((2*scale)/(duty_cycle*period_samples))*(x) - scale + offset
    else
        result = ((-2*scale)/((1-duty_cycle)*period_samples))*x +  (2*scale)/(1-duty_cycle) - scale + offset 
    end
    return result
end

-- Performs any needed post instantiation initialization.
-- Parameter:  SIE writer - NOT USED by this module!
-- Return values:  true (does nothing, for this module)
function mt:setup(sie_writer)
    return true
end

-- Processes incoming frames, produces outgoing frames.
function mt:run()
    if not Pipe_has_frames(self.in_pipe) then
        return
    end

    -- Process each frame in input pipe.  There is only one input pipe.
    local of        -- will reference created output frame
    local out_vals  -- will reference "pointer" to output frame's data
    local _         -- we don't need 'in_vals' "pointer"
    local f = Pipe_get_frame(self.in_pipe)
    while f ~= nil do
        -- create frame for output
        of, _, out_vals = utils.create_output_frame(f.maxsize, f,
            self.tot_count, self.out_frame_sample_size)
        -- set output values
        local count = f.count
        for i = 0, count - 1 do
            if self.tot_count + i <= (self.cycles * self.period_samples) then
                if self.waveform == "sine" then
                    self.last_value = sine(self.tot_count + i, self.period_samples, self.scale, self.offset, self.drift, self.decay)
                    out_vals[i] = self.last_value
                    
                elseif self.waveform == "square" then
                    self.last_value = square(self.tot_count + i, self.period_samples, self.scale, self.offset, self.drift, self.decay, self.duty_cycle)
                    out_vals[i] = self.last_value
                else
                    self.last_value = ramp(self.tot_count + i, self.period_samples, self.scale, self.offset, self.drift, self.decay, self.duty_cycle)
                    out_vals[i] = self.last_value
                end
            else out_vals[i] = self.last_value
            end
        end
        of.count = count
        self.tot_count = self.tot_count + of.count
        -- put output frame in output pipe(s)
        if not utils.put_frame(of, self.pipe_list) then
            Frame_delete(of)
        end
        -- delete input frame, get next
        Frame_delete(f)
        f = Pipe_get_frame(self.in_pipe)
    end
end
local m = {}

-- Creates a new instantiation of this "module".
-- Standard parameters are:
--   stanza --> decoded JSON structure containing the metadata
--       for an instance of this module
--   node_list --> list of all nodes (i.e., module instances, proxy
--       nodes) created to this point, indexed by channel name
--   first_sie_index --> the index of the first sie buffer used by
--       this instance - this module only uses it to set
--       'sie_collect_idx' which is used by run-test.lua
-- Return values:
--   the new instance object
--   the number of output pipe_lists available
function m:new(stanza, node_list, first_sie_index)
    assert(stanza, "Missing parameter 'stanza'")
    local function_generator = setmetatable({}, {__index = mt})
    function_generator.sie_collect_idx = first_sie_index
    utils.set_common_meta(function_generator, stanza)
    -- validate needed metadata, and create input/output pipe
    -- structures
    assert(node_list, "Missing parameter 'node_list'")
    local input_channels = stanza.input_channels or
        stanza.summary.input_channels or {}
    assert(#input_channels == 1, "function generator channel only" ..
        " accepts exactly one input channel")
    local input_chan = input_channels[1]
    assert(input_chan, "Input Channel not found")
    local input_node = node_list[input_chan]
    assert(input_node, "Node for Input Channel not found")
    function_generator.in_pipe = Pipe_attach_pipe(input_node.pipe_list)
    assert(function_generator.in_pipe and function_generator.in_pipe ~= nil,
        "failed to create input pipe")
    function_generator.pipe_list = Pipe_new_pipe_list()
    assert(function_generator.pipe_list and function_generator.pipe_list ~= nil,
        "failed to create pipe list")

    -- get remaining needed metadata
    local sample_rate =
        rational:new(stanza.sample_rate):tonumbers() or
        rational:new(stanza.summary.sample_rate):tonumbers()
    assert(sample_rate and sample_rate > 0,
        string.format("Invalid input sample rate: %s",
        tostring(sample_rate)))
    assert(stanza.data_type, "Missing data type")
    function_generator.out_frame_sample_size = ffi_sizeof(stanza.data_type)
    assert(function_generator.out_frame_sample_size and
           function_generator.out_frame_sample_size > 0,
        string.format("Invalid frame sample size:  %s",
        tostring(function_generator.out_frame_sample_size)))

    function_generator.scale_factor = stanza.timescale or (1 / sample_rate)
    function_generator.scale_factor = tonumber(function_generator.scale_factor)
    
    assert(stanza.waveform == "sine" or stanza.waveform == "ramp" or stanza.waveform == "square", "Invalid waveform type")
    function_generator.waveform = stanza.waveform

    assert(tonumber(stanza.period_samples), "Invalid period samples")
    function_generator.period_samples = stanza.period_samples

    assert(tonumber(stanza.duty_cycle) or stanza.waveform == "sine", "Invalid duty cycle")
    if tonumber(stanza.duty_cycle) then --in cases of sine waves
        function_generator.duty_cycle = stanza.duty_cycle/100 --percent to decimal if there 
    end

    assert(tonumber(stanza.scale), "Invalid scale")
    function_generator.scale = stanza.scale
    
    assert(tonumber(stanza.offset), "Invalid offset")
    function_generator.offset = stanza.offset

    assert(tonumber(stanza.drift), "Invalid drift")
    function_generator.drift = stanza.drift

    assert(tonumber(stanza.decay), "Invalid decay")
    function_generator.decay = stanza.decay

    assert(tonumber(stanza.cycles), "Invalid Cycles")
    function_generator.cycles = stanza.cycles

    -- used for frames' time stamps (just a sample count for
    -- this module)
    function_generator.tot_count = 0

    return function_generator, 1
end

return m
