require 'strict'

-- statisical_analysis.lua - Implements the Statistical Analysis computed channel.

local ffi = require 'ffi'
local frame = require 'test_engine/frame'
local pipe = require 'test_engine/pipe'
local utils = require 'test_engine/test_eng_utils'
local rational = require 'rational'
local log = require 'log'

local ffi_cast = ffi.cast
local ffi_sizeof = ffi.sizeof

local math_min = math.min
local Frame_create = frame.create
local Frame_clone = frame.clone
local Frame_delete = frame.delete
local Pipe_attach_pipe = pipe.attach_pipe
local Pipe_get_frame = pipe.get_frame
local Pipe_has_frames = pipe.has_frames
local Pipe_new_pipe_list = pipe.new_pipe_list
local Utils_frame_data_type = utils.frame_data_type
local Utils_put_frame = utils.put_frame

local mt = {}

-- Performs any needed post instantiation initialization.
-- Parameter:  SIE writer
-- Returns:  true
--           NOTE that for this module this function does nothing!
function mt:setup(sie_writer)
    return true
end

function mt:flush_output_frame()
    self.of.count = self.bufsize
    self.of.hdr.ts = self.tot_count
    self.of.ts = math.floor(self.tot_count)
    if not Utils_put_frame(self.of, self.pipe_list) then
        Frame_delete(self.of)
        self.of = nil
    end
    self.of = Frame_create(self.bufsize, self.out_frame_sample_size)
    self.out_vals = ffi_cast(Utils_frame_data_type(
        self.out_frame_sample_size), self.of.buf)
    self.tot_count = self.tot_count + self.bufsize
    self.out_index = 0
end

function mt:new_window_analysis_check()
    if self.w_res == self.factor then
        if self.need_sum1 then
            self.sum1 = 0
        end
        if self.need_sum2 then
            self.sum2 = 0
        end
    end
end

-- Iterate through all of input buffer, unless the size needed
-- to complete a window analysis is less than the input buffer size
function mt:analysis_size_and_offset(f, inp_index)
    local size = 0
    if f.count - inp_index > self.w_res then
        size = self.w_res
    else
        size = f.count - inp_index
    end
    local fbuf_offset = self.factor - self.w_res
    assert(size > 0, "Size is "..size)
    return size, fbuf_offset
end

-- Track window based params
function mt:track_params(size, inp_index, in_vals, fbuf_offset)
    for w = 0, size-1 do
        if self.need_sum1 then
            self.sum1 = self.sum1 + in_vals[inp_index+w]
        end
        if self.need_sum2 then
            self.sum2 = self.sum2 + in_vals[inp_index+w]*in_vals[inp_index+w]
        end
        if self.fbuf then
            self.fbuf[fbuf_offset+w+1] = in_vals[inp_index+w]
        end
    end
end

function mt:calc_standard_deviation(mean)
    local diff, m2, sum_m2 = nil,nil,0
    for i = 0, self.factor-1 do
        diff = self.fbuf[i+1] - mean
        m2 = diff * diff
        sum_m2 = sum_m2 + m2
    end
    if sum_m2 > 0 then
        self.out_vals[self.out_index] =
            math.sqrt(sum_m2 / (self.factor-1))
    end
end

function mt:calc_skewness(mean)
    local diff, m2, sum_m2, sum_m3 = nil,nil,0,0
    for i = 0, self.factor-1 do
        diff = self.fbuf[i+1] - mean
        m2 = diff * diff
        sum_m2 = sum_m2 + m2
        sum_m3 = sum_m3 + (m2 * diff)
    end
    if sum_m2 > 0 then
        --print(diff, m2, sum_m2, sum_m3, sum_m3 / math.sqrt(sum_m2 * sum_m2 * sum_m2 / self.factor))
        self.out_vals[self.out_index] =
            sum_m3 / math.sqrt(sum_m2 * sum_m2 * sum_m2 / self.factor)
    end
end

function mt:calc_kurtosis(mean)
    local diff, m2 = nil, nil
    local sum_m2, sum_m4 = 0, 0
    for i = 0, self.factor-1 do
        diff = self.fbuf[i+1] - mean
        m2 = diff * diff
        sum_m2 = sum_m2 + m2
        sum_m4 = sum_m4 + m2*m2
    end
    if sum_m2 > 0 then
        self.out_vals[self.out_index] = (self.factor * sum_m4)/(sum_m2 * sum_m2)
    end
end

function mt:calc_percentile()
    local d_idx, fraction, delta, idx = nil,nil,nil,nil
    table.sort(self.fbuf)

    if self.percentile == 0 then
        self.out_vals[self.out_index] = self.fbuf[1]
    elseif self.percentile == 100 then
        self.out_vals[self.out_index] = self.fbuf[self.factor]
    else
        d_idx = (self.factor-1) * (self.percentile / 100.0)
        idx = math.floor(d_idx)
        fraction = d_idx%1
        delta = self.fbuf[idx+2] - self.fbuf[idx+1]
        self.out_vals[self.out_index] = self.fbuf[idx+1] + fraction * delta
    end
end

function mt:do_statistical_analysis()
    local mean = 0
    self.out_vals[self.out_index] = 0
    if self.need_sum1 then
        mean = self.sum1 / self.factor
    end
    if self.mode == "mean" then
        self.out_vals[self.out_index] = mean
    elseif self.mode == "standard_deviation" then
        self:calc_standard_deviation(mean)
    elseif self.mode == "rms" then
        local fwork = self.sum2 / self.factor
        if fwork > 0 then
            self.out_vals[self.out_index] = math.sqrt(fwork)
        end
    elseif self.mode == "skewness" then
        self:calc_skewness(mean)
    elseif self.mode == "kurtosis" then
        self:calc_kurtosis(mean)
    elseif self.mode == "percentile" then
        self:calc_percentile()
    end
end

--Finished reading input frame
function mt:check_window_end()
    if self.w_res == 0 then
        self.w_res = self.factor
        self:do_statistical_analysis()
        self.out_index = self.out_index + 1
        if self.out_index == self.bufsize then
            self:flush_output_frame()
        end
    end
end

-- Processes incoming frames, produces outgoing frames.
function mt:process_in_frame(f)
    local in_vals = ffi_cast(Utils_frame_data_type(f.element_size), f.buf)--data
    local inp_index = 0                    -- Index to input frame data
    while inp_index < f.count do
        self:new_window_analysis_check()
        local size, fbuf_offset = self:analysis_size_and_offset(f, inp_index)
        self:track_params(size, inp_index, in_vals, fbuf_offset)
        inp_index = inp_index + size
        self.w_res = self.w_res - size
        self:check_window_end()
    end
end

-- Processes incoming frames, produces outgoing frames.
function mt:run()
    while Pipe_has_frames(self.in_pipe) do
        local f = Pipe_get_frame(self.in_pipe) -- input frame
        self:process_in_frame(f)
        Frame_delete(f)
        f = nil
    end
end

local m = {}

-- Creates a new instantiation of this "module".
-- Standard parameters are:
--   stanza --> decoded JSON structure containing the metadata
--       for an instance of this module
--   node_list --> list of all nodes (i.e., module instances, proxy
--       nodes) created to this point, indexed by channel name
--   first_sie_index --> the index of the first sie buffer used by
--       this instance - this module only uses it to set
--       'sie_collect_idx' which is used by run-test.lua
-- Return values:
--   the new instance object
--   the number of output pipe_lists available
function m:new(stanza, node_list, first_sie_index)
    assert(stanza, "Missing parameter 'stanza'")
    local stat_a = setmetatable({}, {__index = mt})
    stat_a.sie_collect_idx = first_sie_index
    utils.set_common_meta(stat_a, stanza)
    -- validate needed metadata
    assert(node_list, "Missing parameter 'node_list'")
    local input_channels = stanza.input_channels or {}
    assert(#input_channels == 1, "Statistical Analysis computed channel " ..
        "only accepts exactly one input channel")
    local input_chan = input_channels[1]
    assert(input_chan, "Input Channel not found")
    local input_node = node_list[input_chan]
    assert(input_node, "Node for Input Channel not found")
    assert(input_node.is_logical ~= true, "Input channel must not be a logical "
        .."channel")
        --   pipe (frame) rate
    assert(stanza.pipe_rate, "Missing frame rate")
    stat_a.pipe_rate = rational:new(stanza.pipe_rate):tonumber()
    assert(stat_a.pipe_rate and stat_a.pipe_rate > 0,
        "Invalid frame rate: " .. tostring(stat_a.pipe_rate))
        --   sample rate
    assert(stanza.sample_rate, "Missing sample rate")
    stat_a.sample_rate = rational:new(stanza.sample_rate):tonumber()
    assert(stat_a.sample_rate and stat_a.sample_rate > 0,
        "Invalid sample rate: " .. tostring(stat_a.sample_rate))
    stat_a.out_frame_sample_size = ffi_sizeof(stanza.data_type)
    assert(stat_a.out_frame_sample_size and stat_a.out_frame_sample_size > 0,
        "Invalid frame sample size:  "..tostring(stat_a.out_frame_sample_size))

    -- create input/output pipe structures
    stat_a.in_pipe = Pipe_attach_pipe(input_node.pipe_list)
    assert(stat_a.in_pipe and stat_a.in_pipe ~= nil,
        "failed to create input pipe")
    stat_a.pipe_list = Pipe_new_pipe_list()
    assert(stat_a.pipe_list and stat_a.pipe_list ~= nil,
        "failed to create (output) pipe list")

    --local variables needed not set from stanza.
    stat_a.tot_count = 0
    stat_a.bufsize = frame.bufsize(stat_a.sample_rate, stat_a.pipe_rate)
    -- output frame
    stat_a.of = Frame_create(stat_a.bufsize, stat_a.out_frame_sample_size)
    -- "pointer" to output frame data
    stat_a.out_vals = ffi_cast(Utils_frame_data_type(
        stat_a.out_frame_sample_size), stat_a.of.buf)
    stat_a.out_index = 0

    --Module specific parameters
    stat_a.factor = tonumber(stanza.factor)
    assert(stat_a.factor and stat_a.factor%1 == 0 and stat_a.factor > 1,
        "Invalid factor: "..tostring(stanza.factor)..
        ", must be an integer greater than 1.")
    --IFDEF mean, standard_deviation, rms, kurtosis, skewness, xth_percentile
    stat_a.mode = stanza.mode
    assert(stat_a.mode == "mean" or
           stat_a.mode == "standard_deviation" or
           stat_a.mode == "rms" or
           stat_a.mode == "kurtosis" or
           stat_a.mode == "skewness" or
           stat_a.mode == "percentile",
        "Invalid mode: "..tostring(stanza.mode))
    if stat_a.mode == "percentile" then
        stat_a.percentile = tonumber(stanza.percentile)
        assert(stat_a.percentile and stat_a.percentile >= 0 and
            stat_a.percentile <= 100, "Invalid percentile: "..
            tostring(stanza.percentile).." must be a number in the range 0-100")
    end
    if stat_a.mode == "standard_deviation" or
       stat_a.mode == "skewness" or
       stat_a.mode == "kurtosis" or
       stat_a.mode == "percentile" then
        stat_a.fbuf = {}
    end
    stat_a.need_sum1 = true
    if stat_a.mode == "rms" or stat_a.mode == "percentile" then
        stat_a.need_sum1 = false
    end
    if stat_a.mode == "rms" then
        stat_a.need_sum2 = true
    end

    stat_a.w_res = stat_a.factor

    return stat_a, 1
end

return m
