require 'strict'

-- triggered_latch.lua - Implements the Triggered Latch computed channel.

local ffi = require 'ffi'
local frame = require 'test_engine/frame'
local pipe = require 'test_engine/pipe'
local utils = require 'test_engine/test_eng_utils'
local rational = require 'rational'

local ffi_cast = ffi.cast
local ffi_sizeof = ffi.sizeof
local Frame_create = frame.create
local Frame_delete = frame.delete
local Pipe_attach_pipe = pipe.attach_pipe
local Pipe_get_frame = pipe.get_frame
local Pipe_has_frames = pipe.has_frames
local Pipe_new_pipe_list = pipe.new_pipe_list
local Utils_frame_data_type = utils.frame_data_type
local Utils_put_frame = utils.put_frame

local mt = {}

-- Performs any needed post instantiation initialization.
-- Parameter:  SIE writer
-- Returns:  true
--           NOTE that for this module this function does nothing!
function mt:setup(sie_writer)
    return true
end

-- Processes incoming frames, produces outgoing frames.
function mt:run()
    while Pipe_has_frames(self.in_pipe) and Pipe_has_frames(self.trig_pipe) do
        local f  = Pipe_get_frame(self.in_pipe)   -- input frame
        local tf = Pipe_get_frame(self.trig_pipe) -- trigger frame
        local of, in_vals, out_vals = utils.create_output_frame(f.maxsize, f,
            self.tot_count, self.out_frame_sample_size)
        of.hdr.ts = self.tot_count
        local trig_vals = ffi_cast(Utils_frame_data_type(tf.element_size), tf.buf)
        
        for i = 0, f.count-1 do
            self.triggered = (trig_vals[i] ~= 0)
            if not self.invert_trigger then
                out_vals[i]    = self.triggered and self.latch_val or in_vals[i]
                if not self.triggered then
                    self.latch_val = out_vals[i]
                end
            else
                out_vals[i]    = self.triggered and in_vals[i] or self.latch_val
                if self.triggered then
                    self.latch_val = out_vals[i]
                end
            end
        end
        of.count = self.bufsize
        if not Utils_put_frame(of, self.pipe_list) then
            Frame_delete(of)
            of = nil
        end
        self.tot_count = self.tot_count + self.bufsize
        --delete used frames
        Frame_delete(f)
        Frame_delete(tf)
        f = nil
        tf = nil
    end
end


local m = {}

-- Creates a new instantiation of this "module".
-- Standard parameters are:
--   stanza --> decoded JSON structure containing the metadata
--       for an instance of this module
--   node_list --> list of all nodes (i.e., module instances, proxy
--       nodes) created to this point, indexed by channel name
--   first_sie_index --> the index of the first sie buffer used by
--       this instance - this module only uses it to set
--       'sie_collect_idx' which is used by run-test.lua
-- Return values:
--   the new instance object
--   the number of output pipe_lists available
function m:new(stanza, node_list, first_sie_index)
    assert(stanza, "Missing parameter 'stanza'")
    local trig_l = setmetatable({}, {__index = mt})
    trig_l.sie_collect_idx = first_sie_index
    utils.set_common_meta(trig_l, stanza)
    -- validate needed metadata
    assert(node_list, "Missing parameter 'node_list'")
    local input_channels = stanza.input_channels or {}
    assert(#input_channels == 1, "Triggered Latch computed channel " ..
        "only accepts exactly one input channel")
    local input_chan = input_channels[1]
    assert(input_chan, "Input Channel not found")
    local input_node = node_list[input_chan]
    assert(input_node, "Node for Input Channel not found")
        --   pipe (frame) rate
    assert(stanza.pipe_rate, "Missing frame rate")
    trig_l.pipe_rate = rational:new(stanza.pipe_rate):tonumber()
    assert(trig_l.pipe_rate and trig_l.pipe_rate > 0,
        "Invalid frame rate: " .. tostring(trig_l.pipe_rate))
        --   sample rate
    assert(stanza.sample_rate, "Missing sample rate")
    trig_l.sample_rate = rational:new(stanza.sample_rate):tonumber()
    assert(trig_l.sample_rate and trig_l.sample_rate > 0,
        "Invalid sample rate: " .. tostring(trig_l.sample_rate))
    trig_l.out_frame_sample_size = ffi_sizeof(stanza.data_type)
    assert(trig_l.out_frame_sample_size and trig_l.out_frame_sample_size > 0,
        "Invalid frame sample size:  "..tostring(trig_l.out_frame_sample_size))
        
    --Module specific parameters
    local trig_channel = stanza.trigger_channel
    assert(trig_channel, "Missing parameter 'trigger_channel'")
    local trig_node = node_list[trig_channel]
    assert(trig_node, "Node for Trigger Channel not found")
    assert(trig_node.is_logical, "Trigger channel must be a logical channel")
    trig_l.trig_pipe = Pipe_attach_pipe(trig_node.pipe_list)
    assert(trig_l.trig_pipe and trig_l.trig_pipe ~= nil,
        "failed to create trigger pipe")
    trig_l.invert_trigger = utils.xlate_meta_boolean(stanza.invert_trigger)
    assert(trig_l.invert_trigger ~= nil, "Missing parameter 'invert_trigger'")

    -- create input/output pipe structures
    trig_l.in_pipe = Pipe_attach_pipe(input_node.pipe_list)
    assert(trig_l.in_pipe and trig_l.in_pipe ~= nil,
        "failed to create input pipe")
    trig_l.pipe_list = Pipe_new_pipe_list()
    assert(trig_l.pipe_list and trig_l.pipe_list ~= nil,
        "failed to create (output) pipe list")
        
    --local variables needed not set from stanza.
    trig_l.tot_count = 0
    trig_l.bufsize = frame.bufsize(trig_l.sample_rate, trig_l.pipe_rate)
    trig_l.triggered = nil
    trig_l.latch_val = 0

    return trig_l, 1
end

return m
