require 'strict'

-- burst_message_logger.lua - Implements the Burst Message Logger data mode.
--
-- Implementation notes:
--     This data mode uses its own particular trigger mechanism,
--     not the "generic" datamode triggering mechanism.  The mechanism
--     is functionally the same as the Burst History trigger mechanism
--     but is quite different in implementation due to the differences
--     between message channels and sequential channels.
--
--     There is no "Storage mode" option as Burst History has.  The
--     storage operates essentially the same as Burst History's
--     "Enhanced" storage mode.  All triggers are recognized whenever
--     they occur and it is guaranteed that at least a 'burst_post_trigger_time'
--     amount of burst history is recorded after each trigger (with the
--     possible exception of the end of the test, of course).

local cjson = require 'cjson'
local ffi = require 'ffi'
local frame = require 'test_engine/frame'
local log = require 'log'
local pipe = require 'test_engine/pipe'
local rational = require 'rational'
local sie_schema = require 'test_engine/sie_schema'
local sie_utils = require 'test_engine/sie_utils'
local utils = require 'test_engine/test_eng_utils'

local C = ffi.C
local ffi_cast = ffi.cast
local ffi_new = ffi.new
local ffi_sizeof = ffi.sizeof
local math_max = math.max
local math_min = math.min
local math_floor = math.floor

local Frame_create = frame.create
local Frame_delete = frame.delete
local Frame_tail   = frame.tail
local Pipe_attach_pipe   = pipe.attach_pipe
local Pipe_has_frames    = pipe.has_frames
local Pipe_get_frame     = pipe.get_frame
local Pipe_new_pipe_list = pipe.new_pipe_list

-- "constant" values for the possible runtime states
local ST_LOOK_FOR_TRIG           = 1
local ST_FIND_PRE_HISTORY_START  = 2
local ST_WRITE_PRE_HISTORY_START = 3
local ST_FINISH_BURST            = 4
-- "constant" values to deal with Lua not having integer values
local EPSILON = 0.000001  -- much less than 1/100,000 (much < min clock incr.)
local HI_ZERO = 0.0 + EPSILON
local LO_ZERO = 0.0 - EPSILON

local TICKS_PER_SEC = 40000000

local VERY_SMALL = -math.huge

local mt = {}

-- Performs any needed post instantiation initialization.
-- Parameter:
--     'sie_writer'  SIE writer
-- Returns:  true
function mt:setup(sie_writer)
    self.sw = sie_writer
    self:sie_xml()
    return true
end

-- Creates the SIE buffers (pipes and XML) required.
-- There will be one channel description per input channel.
function mt:sie_xml()
    for i = 1, #self.chan_info do
        local info = self.chan_info[i]
        -- write XML for this SIE channel
        info.stanza.description = self.stanza.description or ""
        info.stanza.datamode_name = self.name
        info.stanza.sie_name = self.name .. '@' .. info.name
        info.stanza.somat_input_channel = info.stanza.name
        local schema_name =
            sie_schema.get_collect_schema_name(info.stanza)
        local schema_info = sie_schema.get_schema_info(schema_name, self.stanza)
        schema_info.dims[1]['core:label'] = "Time"
        schema_info.dims[1]['core:units'] = "Seconds"
        schema_info.dims[2]['core:label'] = info.name
        schema_info.dims[2]['core:units'] = info.stanza.output_units
        schema_info['core:uuid'] = self.stanza.id .. '/' .. info.stanza.id
        local md = sie_utils.write_sie_channel(self.sw, schema_info,
            info.stanza)
        -- Attach one end of pipe to this module for this input node
        local p = Pipe_attach_pipe(info.pipe_list)
        -- Create the SIE buffer and attach the pipe to it
        local not_fixed_size
        if info.stanza.data_type:find("message_raw") and
            (info.stanza.category:find("msg_gps$") or
             info.stanza.category:find("msg_camera$")) then
            not_fixed_size = true
        end
        local size = utils.sie_buf_size(self, info.stanza)
        self.sw:new_buffer(self.sie_collect_idx + i - 1,
            md._sie_group, size, nil, 8, p, true, nil, not_fixed_size)
    end
end

-- Implements "gate" trigger, returning true whenever the input value
-- is non-zero.
-- Parameter:
--     'trig_data'  sample from trigger channel
-- Returns:  true if 'trig_data' is non-zero, false otherwise
function mt:gate_trigger(trig_data)
    return not (trig_data >= LO_ZERO and trig_data <= HI_ZERO)
end

-- Implements "edge" trigger, returning true whenever the input value
-- is non-zero and the previous trigger channel value was zero.
-- Parameter: 'trig_data'  sample from trigger channel
-- Returns:  true if 'trig_data' is transitions from zero to non-zero,
--     false otherwise
function mt:edge_trigger(trig_data)
    local nbr = tonumber(trig_data)
    if nbr >= LO_ZERO and nbr <= HI_ZERO then
        self.prev_trig_on = false
        return false
    else  -- trig_data ~= 0  
        if self.prev_trig_on then
            return false
        else
            self.prev_trig_on = true
            return true
        end
    end
end

-- Collects any triggers from all currently available trigger frames.
-- The collected triggers are put in each channel's trigger_history.
-- Returns:  the timestamp of the last trigger frame processed,
--           or nil if no frames were found
function mt:collect_trigs()
    local t_his
    local trig_ts
    local last_frame_ts = nil
    local trig_frame = Pipe_get_frame(self.trig_pipe)
    while trig_frame and (trig_frame ~= nil) do
        last_frame_ts = tonumber(trig_frame.ts) * self.ticks_per_sample
        local trig_vals =
            ffi_cast(utils.frame_data_type(trig_frame.element_size),
                     trig_frame.buf)
        for i = 0, trig_frame.count - 1 do
            if self.trigger(self, trig_vals[i]) then
                trig_ts = (tonumber(trig_frame.ts) + i) * self.ticks_per_sample
                for i = 1, #self.chan_info do
                    t_his = {trig = trig_ts,
                             b_start = trig_ts - self.pre_trig_time,
                             b_end = trig_ts + self.post_trig_time}
                    table.insert(self.chan_info[i].trigger_history, t_his)
                end
            end
        end
        Frame_delete(trig_frame)
        trig_frame = Pipe_get_frame(self.trig_pipe)
    end

    return last_frame_ts
end

-- Determines whether the given frame contains a sample
-- which is within an active burst.  Also keeps track of
-- the current burst.
-- Parameter:  ch - input channel information
--             iv - pointer to the input frame's data buffer
-- Returns:  true if the input frame contains a sample within
--     an acive burst, false otherwise
function mt:frame_overlaps_burst(ch, iv)
    local have_burst = true
    local in_burst
    -- are we done with the current burst?
    local first_ts = tonumber(iv[0].timestamp[0])
    local last_ts = tonumber(iv[ch.in_frame.count - 1].timestamp[0])
    if first_ts > ch.curr_burst_end + EPSILON then
        -- "reset" burst start and end to indicate we're
        -- done with the burst
        ch.curr_burst_start = VERY_SMALL
        ch.curr_burst_end = VERY_SMALL
        have_burst = false
    end

    -- check for unprocessed triggers
    while #ch.trigger_history > HI_ZERO do
        local trig = ch.trigger_history[1]
        -- check for messed up trigger input
        if (type(trig) ~= "table") or
           (trig.trig < ch.curr_burst_start - EPSILON) then
            log:error(string.format("Invalid or out of order trigger, " ..
                "possibly at time %s (with burst start at %s)",
                tostring(trig.trig), tostring(ch.curr_burst_start)))
        else
            if have_burst then
                -- check for burst continuation
                if trig.b_start <= ch.curr_burst_end + 1 + EPSILON then
                    -- next burst overlaps or adjoins current burst
                    -- so extend the current burst
                    ch.curr_burst_end = trig.b_end
                else
                    -- there will be a disconnected burst
                    -- but we'll deal with it later
                    break
                end
            else
                -- start next burst
                if first_ts > trig.b_end + EPSILON then
                    -- but first remove stale triggers
                    have_burst = false
                else
                    ch.curr_burst_start = trig.b_start
                    ch.curr_burst_end = trig.b_end
                    have_burst = true
                end
            end
        end
        table.remove(ch.trigger_history, 1)
    end

    if have_burst then
        if (first_ts > ch.curr_burst_end + EPSILON) or
           (last_ts < ch.curr_burst_start - EPSILON) then
            in_burst = false
        else
            in_burst = true
        end
    else
        in_burst = false
    end

    return in_burst
end

-- Processes incoming frames, producing outgoing frames.
-- Parameters:
--     force - set (or not) by run-test.lua, ignored
--     do_not_wait_on_trigs - normally nil, set to true
--          when called by mt:finalize(); causes all existing
--          input frames to be processed regardless of how
--          much trigger history exists
function mt:run(force, do_not_wait_on_trigs)
    -- get all available triggers
    local trig_frame_ts = self:collect_trigs()
    if trig_frame_ts then
        self.trig_frame_time = trig_frame_ts 
    end

    -- Each input channel is processed to exhaustion before
    -- proceeding to the next one.
    for i = 1, #self.chan_info do
        local ch = self.chan_info[i]
        if ch.in_frame == nil and
           Pipe_has_frames(ch.in_pipe) then
            ch.in_frame = Pipe_get_frame(ch.in_pipe)
        end
        while ch.in_frame ~= nil do
            -- don't want empty frames
            if ch.in_frame.count <= HI_ZERO then
                log:error(string.format("Internal error - empty frame " ..
                    "with ts %s", tostring(ch.ts)))
            end
            local sample_type = ((ch.data_type == "message_can") and
                "message_can *") or "message_raw_buf *"
            local iv = ffi_cast(sample_type, ch.in_frame.buf)
            if not do_not_wait_on_trigs then
                -- check that we have enough trigger history
                if tonumber(iv[ch.in_frame.count - 1].timestamp[0]) >=
                   (self.trig_frame_time - self.pre_trig_time - EPSILON) then
                    break  -- need more trigger data before processing input
                end
            end
            -- process frames
            if self:frame_overlaps_burst(ch, iv) then
                if not utils.put_frame(ch.in_frame, ch.pipe_list) then
                    Frame_delete(ch.in_frame)
                end
            else
                Frame_delete(ch.in_frame)
            end
            ch.in_frame = nil

            -- any more input frames?
            if Pipe_has_frames(ch.in_pipe) then
                ch.in_frame = Pipe_get_frame(ch.in_pipe)
            end
        end
    end
end

-- Cleans up at test conclusion.
-- Parameter:
--     'sie_writer' is SIE writer object - NOT USED HERE
function mt:finalize(sie_writer)
    -- There are probably input frames waiting for enough
    -- trigger history to be processed.  Get any remaining
    -- trigger history and process all remaining input frames.
    self:run(false, true) -- force=false, do_not_wait_on_trigs=true
end


local m = {}

-- Creates a new instantiation of this "module".
-- Standard parameters are:
--   stanza --> decoded JSON structure containing the metadata
--       for an instance of this module
--   node_list --> list of all nodes (i.e., module instances, proxy
--       nodes) created to this point, indexed by channel name
--   first_sie_index --> the index of the first sie buffer for
--       this instance - this module only uses it to set
--       'sie_collect_idx' which is used by run-test.lua
--   meta_list --> list of all created-to-this-point nodes' stanzas,
--       indexed by channel name
-- Return values:
--   the new instance object
function m:new(stanza, node_list, first_sie_index, meta_list)
    assert(stanza, "Missing parameter 'stanza'")
    assert(node_list, "Missing parameter 'node_list'")
    assert(first_sie_index, "Missing parameter 'first_sie_index'")
    assert(meta_list, "Missing parameter 'meta_list'")
    local msg_burst_history = setmetatable({}, {__index = mt})
    msg_burst_history.stanza = stanza
    -- don't need to call set_common_meta() since there are
    -- no modules downstream
    msg_burst_history.sie_collect_idx = first_sie_index
    msg_burst_history.chan_info = {}
    -- validate and process needed metadata
    local input_channels = stanza.input_channels
    assert(#input_channels >= 1 - EPSILON, "Burst Message Logger data mode " ..
        "must have at least one input channel")
    --     node name
    msg_burst_history.name = stanza.name
    assert(msg_burst_history.name, "Missing node name")
    --     only message channels allowed
    for i = 1, #input_channels do
        local node_meta = meta_list[input_channels[i]]
        assert(node_meta.sample_rate == nil, 
               string.format("'%s' is not a message channel",
                             tostring(node_meta.name)))
    end
    --     pipe rate
    assert(stanza.pipe_rate, "Missing pipe_rate")
    msg_burst_history.pipe_rate = tonumber(stanza.pipe_rate)
    assert(msg_burst_history.pipe_rate and msg_burst_history.pipe_rate > 0,
           string.format("Invalid pipe_rate: %s", tostring(stanza.pipe_rate)))
    msg_burst_history.frame_time_incr = 1 / msg_burst_history.pipe_rate
    --     trigger sample rate
    local trig_chan = stanza.trigger_channel
    assert(trig_chan, "Burst Message Logger requires a trigger channel")
    local trig_node = node_list[trig_chan]
    assert(trig_node, "Node for trigger channel " ..
           tostring(trig_chan) .. " not found")
    local trig_meta = meta_list[trig_chan]
    assert(trig_meta, string.format("Internal error: Burst Message Logger " ..
           "can't find metadata for trigger channel '%s'", tostring(trig_chan)))
    msg_burst_history.trig_rate = tonumber(trig_meta.sample_rate)
    msg_burst_history.ticks_per_sample =
        TICKS_PER_SEC / msg_burst_history.trig_rate
    assert(msg_burst_history.trig_rate, string.format("Invalid trigger " ..
           "channel sample rate: '%s'", tostring(trig_meta.sample_rate)))
    --     pre- and post-trigger counts
    assert(stanza.burst_pre_trigger_time, "Missing pre-trigger time")
    local pre_trig = tonumber(stanza.burst_pre_trigger_time)
    msg_burst_history.pre_trig_time = pre_trig * TICKS_PER_SEC
    assert(msg_burst_history.pre_trig_time >= LO_ZERO, "Invalid pre-trigger " ..
        "time: \"" .. tostring(stanza.burst_pre_trigger_time) .. "\"")
    assert(stanza.burst_post_trigger_time, "Missing post-trigger time")
    local post_trig = tonumber(stanza.burst_post_trigger_time)
    msg_burst_history.post_trig_time = post_trig * TICKS_PER_SEC
    assert(msg_burst_history.post_trig_time >= LO_ZERO, "Invalid " ..
        "post-trigger time: \"" ..  tostring(stanza.burst_post_trigger_time) ..
        "\"")
    --     trigger
    local trigger_mode = stanza.trigger_mode
    assert(trigger_mode and
        (trigger_mode == "when_true" or
         trigger_mode == "on_false_to_true_edge"),
        "Missing or invalid trigger mode: \"" ..
        tostring(trigger_mode) .. "\"")
    msg_burst_history.trig_pipe = Pipe_attach_pipe(trig_node.pipe_list)
    assert(msg_burst_history.trig_pipe and msg_burst_history.trig_pipe ~= nil,
        "failed to create trigger pipe")
    msg_burst_history.trigger = trigger_mode == "when_true" and
        msg_burst_history.gate_trigger or msg_burst_history.edge_trigger
    --     create input pipe structures & other per input channel values
    local input_node
    local node_meta
    for i = 1, #input_channels do
        input_node = node_list[input_channels[i]]
        assert(input_node, "Node for Input Channel \"" ..
            tostring(input_channels[i]) .. "\" not found")
        node_meta = meta_list[input_channels[i]]
        assert(node_meta, "metadata stanza for Input Channel \"" ..
            tostring(input_channels[i]) .. "\" not found")
        --     data type
        local data_type = node_meta.data_type
        assert(data_type, "Missing data_type for \"" ..
            tostring(input_channels[i]) .. "\"")
        msg_burst_history.stanza.nbr_data_bits = ''
        -- channel info array
        msg_burst_history.chan_info[i] =
            {name = input_channels[i],
             -- in_node = input_node,
             in_pipe = Pipe_attach_pipe(input_node.pipe_list),
             pipe_list = Pipe_new_pipe_list(),
             stanza = node_meta,
             data_type = data_type,
             in_frame = nil,
             out_frame = nil,
             curr_burst_start = VERY_SMALL,
             curr_burst_end = VERY_SMALL,
             trigger_history = {}
         }
        assert(msg_burst_history.chan_info[i].in_pipe and
            msg_burst_history.chan_info[i].in_pipe ~= nil,
            "failed to create input pipe")
        assert(msg_burst_history.chan_info[i].pipe_list and
            msg_burst_history.chan_info[i].pipe_list ~= nil,
            "failed to create pipe list")
    end

    -- initialize volatile values
    msg_burst_history.prev_trig_on = false
    msg_burst_history.trig_frame_time = VERY_SMALL

    return msg_burst_history, #input_channels  -- 1 sie buf per input chan
end

return m
