require 'strict'

-- sie_templates.lua - Provides strings and string templates useful when
-- creating XML for a SIE file.

local m = {}


--------------------------------------------
-- Definition for the XML header for a file.
--------------------------------------------

m.sie_header = [==[
<?xml version="1.0" encoding="UTF-8"?>
<sie version="1.0" xmlns="http://www.somat.com/SIE">
<!-- SIE format standard definitions: -->
 <!-- SIE stream decoder: -->
 <decoder id="0">
  <loop>
   <read var="size" bits="32" type="uint" endian="big"/>
   <read var="group" bits="32" type="uint" endian="big"/>
   <read var="syncword" bits="32" type="uint" endian="big" value="0x51EDA7A0"/>
   <read var="payload" octets="{$size - 20}" type="raw"/>
   <read var="checksum" bits="32" type="uint" endian="big"/>
   <read var="size2" bits="32" type="uint" endian="big" value="{$size}"/>
  </loop>
 </decoder>
 <tag id="sie:xml_metadata" group="0" format="text/xml"/>

 <!-- SIE index block decoder:  v0=offset, v1=group -->
 <decoder id="1">
  <loop>
   <read var="v0" bits="64" type="uint" endian="big"/>
   <read var="v1" bits="32" type="uint" endian="big"/>
   <sample/>
  </loop>
 </decoder>
 <tag id="sie:block_index" group="1" decoder="1"/>

<!-- Stream-specific definitions begin here: -->

<!--
 Element Reference:
     ch = channel
    dim = dimension
  xform = transform
-->
 <tag id="core:version">1.0</tag>
 <tag id="somat:version">1.0</tag>
 <tag id="somat:tce_setup">There is no TCE setup file in a SomatXR generated SIE file. Use the tag "sie_file/somat:sxr_setup" to get the SXR configuration file.</tag>
 <tag id="somat:log">Log file is not provided at this time.</tag>
]==]

----------------------------------------------------
-- Definitions for a test contained in the SIE file.
----------------------------------------------------

-- defines the XML for a test's header
m.test_uuid = [[
 <test id="(=test_id=)">
  <tag id="core:uuid">(=uuid=)</tag>
 </test>
]]
m.test_header = [[
 <test id="(=test_id=)">
  <tag id="core:test_count">(=run_number=)</tag>
  <tag id="core:description">(=run_description=)</tag>
  <tag id="core:start_time">(=start_time=)</tag>
 </test>
]]

-------------------------------
-- Definitions for a group tag.
-------------------------------
m.grp_tag = '(=i=)<tag id="(=id=)" group="(=grp=)" format="(=fmt=)"/>\n'

-------------------------------------------------
-- Definitions for a channel contained in a test.
-------------------------------------------------

--=== Rainflow-specific definitions
m.unclosed_cycles = '<tag test="(=_test_id=)" ch="(=_ch_id=)" id="somat:rainflow_unclosed_cycles">(=_unclosed_cycles=)</tag>'

--=== Histogram-specific definitions
-- defines the XML for the start of a histogram bin bounds lookup table channel
m.generic_bin_ch = '<ch id="(=_sie_ch=)" private="bin">'
-- defines the XML for a histogram's bin bounds channel
m.priv_bin_ch = '<ch id="(=_sie_ch=)" base="(=_base_id=)" group="(=_sie_group=)" name="(=_dm=)@(=_name=)#bin_(=_histdim=)" private="bin"/>'
-- defines the XML for the start of a private test channel
--     (e.g., for a histogram)
m.priv_test_ch = '<ch id="(=_sie_ch=)" name="(=_sie_name=)" private="base">'

--=== general channel definitions
-- defines the XML for the start of a base channel
m.base_ch = '<ch id="(=_sie_ch=)" name="(=_sie_name=)" private="(=_priv_type=)">'
-- defines the XML for the start of a channel
m.test_ch = '<ch test="(=_sie_test=)" id="(=_sie_ch=)" name="(=_sie_name=)" group="(=_sie_group=)">'
-- defines the XML for the end of a channel specification
m.ch_end = '</ch>'
-- defines the XML for a channel tag
m.ch_tag = '<tag id="(=k=)">(=v=)</tag>'

--=== The following lines are used to generate the XML for a
--=== decoder definition:
-- decoder element start
m.dec_start = '<decoder id="(=_sie_decoder=)">'
-- assigns the next "bits" from the current data position
--     to the specified variable
m.dec_read = '<read var="(=varname=)" bits="(=bits=)" type="(=sie_data_type=)" endian="(=endianess=)"/>'
m.dec_read_octets = '<read var="(=varname=)" octets="(=octets=)" type="(=sie_data_type=)"/>'
m.dec_seek = '<seek from="(=from=)" offset="(=offset=)"/>'
-- sets the specified variable to the specified value
m.dec_set = '<set var="(=varname=)" value="(=value=)"/>'
-- simplest of loop headers
m.dec_bare_loop = '<loop>'
-- loop header with loop variable (incremented each iteration)
m.dec_short_loop = '<loop var="(=lvarname=)">'
-- loop header with loop variable which has specified
--     start and end values; note that the loop terminates if
--     loop variable equals end value (e.g., loop does not execute
--     if variable == end)
m.dec_long_loop = '<loop var="(=lvarname=)" start="(=start=)" end="(=endvar=)">'
-- Causes a data sample vector to be "added to the decoder's
--     output queue".  This will include all variables of the form
--     v0, v1, v2, etc., which are used in the decoder.
--     (See section 2.4.3.6 in sie-format document.)
m.dec_sample = '<sample/>'
--     loop end
m.dec_loop_end = '</loop>'
--     decoder element end
m.dec_end = '</decoder>'

--=== The following lines are used to generate the XML for SIE dims
-- defines the XML for the start of a dim (dimension) specification
m.dim_hdr = '<dim index="(=_dim_index=)">'

-- defines the XML for a decoder assignment
m.dec_assign = '<data decoder="(=_sie_decoder=)" v="(=_decoder_v=)"/>'

-- defines the XML for the end of a dim (dimension)
-- specification
m.dim_end = '</dim>'

-- defines the XML for a decoder dim's scaling
m.dec_scale = '<xform scale="(=_scale=)" offset="(=_offset=)"/>'

-- defines the XML for a decoder dim's index transform
m.dec_xindex = '<xform index_ch="(=_index_ch=)" index_dim="(=_index_dim=)"/>'

return m

